/*--------------------------------------------------------------------------
     "rxnmacro.cc" - reaction rate macros
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2010 by A. Miyoshi, University of Tokyo
                                       created: Oct.  31, 2003
                                        edited: Apr.  13, 2005
                                        edited: July   7, 2010
                                   last edited: Nov.  19, 2010
                            - added 'sum-of-rates' expression
--------------------------------------------------------------------------*/

#include "rxnmacro.h"

// ***** rateMacro and related classes ***********************************

// ===== rxnRate class ===================================================

rxnRate::rxnRate() { clear(); }                        // --- constructors

rxnRate::rxnRate(const rxnRate &rate) { (*this) = rate; }

void rxnRate::clear() {                                       // --- clear
  A = 0.; b = 0.; Ea = 0.;
  isTroe = false;
  AL = 0.; bL = 0.; EaL = 0.; aFc = 0.; T3Fc = 0.; T1Fc = 0.; T2Fc = 0.;
}

int rxnRate::restoreFrom(ifstream &ifs) {               // --- restoreFrom
  vector<string> vs;
  if (getTokensFrom(ifs, 3, vs) == ERRE) { return ERRE; }
  A = atof(vs[0].c_str()); b = atof(vs[1].c_str());
  Ea = atof(vs[2].c_str());
  return NORM;
}

int rxnRate::restoreLowTroeFrom(ifstream &ifs) { // --- restoreLowTroeFrom
  vector<string> vs;
  if (getTokensFrom(ifs, 3, vs) == ERRE) { return ERRE; }
  AL = atof(vs[0].c_str()); bL = atof(vs[1].c_str());
  EaL = atof(vs[2].c_str());
  if (getTokensFrom(ifs, 4, vs) == ERRE) { return ERRE; }
  aFc = atof(vs[0].c_str()); T3Fc = atof(vs[1].c_str());
  T1Fc = atof(vs[2].c_str()); T2Fc = atof(vs[3].c_str());
  return NORM;
}

string rxnRate::format() {                                   // --- format
  return formatE(A, 11, 3) + formatF(b, 8, 3) + formatF(Ea, 10, 1);
}

string rxnRate::formatLow() {                             // --- formatLow
  return formatE(AL, 11, 3) + formatF(bL, 8, 3) + formatF(EaL, 10, 1);
}

string rxnRate::formatTroe() {                           // --- formatTroe
  return ' ' + formatFautoLen(aFc, 4) + ' ' + formatFautoLen(T3Fc, 1)
       + ' ' + formatFautoLen(T1Fc, 1) + formatE(T2Fc, 11, 3);
}

// ===== rateMacro class =================================================

rateMacro::rateMacro() {                                // --- constructor
  libPath libp;
  clear();
  if (loadFrom(FL_RXNMCR, libp) == ERRE)
    { cout << "Error reading reaction macro table.\n"; }
}

// ---------------------------------------------------------- [loadFrom]
int rateMacro::loadFrom(string flname, libPath &libp) {
  string s;
  rxnRate rate;
  if (!empty()) { clear(); }
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  while (getTokenFrom(ifs, s) == NORM) {
    rate.clear();
    if (s == "") { break; }
    if (rate.restoreFrom(ifs) == ERRE) { ifs.close(); return ERRE; }
    if (s.substr(0, 5) == "Troe/") {
      if (rate.restoreLowTroeFrom(ifs) == ERRE) { ifs.close(); return ERRE; }
      rate.isTroe = true;
    }
    insert(make_pair(s, rate));
  }
  ifs.close();
  cout << size() << " rate macros loaded from [" << flname << "]\n";
  return NORM;
}

// ***** thermNASA and related classes ***********************************

// ===== thermNASA class =================================================

// ------------------------------------------------------- [thermNASA()]
thermNASA::thermNASA() { clear(); }

// ------------------------------------------------------- [thermNASA()]
thermNASA::thermNASA(const thermNASA &tm) { (*this) = tm; }

// ------------------------------------------------------------- [clear]
void thermNASA::clear() {
  int i, j;
  isUsed = false; atmCmp.clear();
  lowT = 300.; highT = 5000.; comT = 1500.;
  for (i = 0; i < 2; i++) { for (j = 0; j < 7; j ++) { c[i][j] = 0.; } }
}

// ------------------------------------------------------- [restoreFrom]
int thermNASA::restoreFrom(vector<string> &vs) {
  int ia, ic, pos, atmcNbr, numCnt;
  string aSym;

  clear();
  for (ia = 0; ia < 4; ia++) {
    pos = ia * 5 + 24;
    aSym = strToUpper(strRemoveSpc(vs[0].substr(pos, 2)));
    atmcNbr = elmt.atmcNbrBySymb(aSym);
    numCnt = atoi(strRemoveSpc(vs[0].substr(pos + 2, 3)).c_str());
    if ((atmcNbr > 0) && (numCnt > 0))
      { atmCmp.insert(make_pair(atmcNbr, numCnt)); }
  }
  lowT = atof(strRemoveSpc(vs[0].substr(45, 10)).c_str());
  highT = atof(strRemoveSpc(vs[0].substr(55, 10)).c_str());
  comT = atof(strRemoveSpc(vs[0].substr(65, 8)).c_str());
  for (ic = 0; ic < 5; ++ic)
    { pos = ic * 15; c[0][ic] = getValCoeff(vs[1].substr(pos, 15)); }
  for (ic = 0; ic < 2; ++ic)
    { pos = ic * 15; c[0][ic + 5] = getValCoeff(vs[2].substr(pos, 15)); }
  for (ic = 2; ic < 5; ++ic)
    { pos = ic * 15; c[1][ic - 2] = getValCoeff(vs[2].substr(pos, 15)); }
  for (ic = 0; ic < 4; ++ic)
    { pos = ic * 15; c[1][ic + 3] = getValCoeff(vs[3].substr(pos, 15)); }
  return NORM;
}

// ----------------------------------------------------- [getThermFuncs]
void thermNASA::getThermFuncs(double T, double &HRT, double &SR,
 double &CpR) {
  int i, idxTr;
  double a;
  if (T > comT) { idxTr = 0; } else { idxTr = 1; }
  HRT = 0.; SR = 0.; CpR = 0.;
  for (i = 4; i >= 0; i--) {
    a = c[idxTr][i];
    CpR = CpR * T + a;
    HRT = HRT * T + a / (i + 1);
    if (i == 0) { SR = SR * T + a * log(T); }
    else { SR = SR * T + a / i; }
  }
  HRT += c[idxTr][5] / T;
  SR += c[idxTr][6];
}

// ------------------------------------------------------- [getValCoeff]
double thermNASA::getValCoeff(const string &s)
  { string ws = fixCoefField(s); return atof(ws.c_str()); }

// ------------------------------------------------------ [fixCoefField]
string thermNASA::fixCoefField(const string &s) {
  static const string sfil = " 0.00000000e+00";
  string rets = s;
  for (int ic = 1; ic < 15; ic ++)
    { if (rets[ic] == ' ') { rets[ic] = sfil[ic]; } }
  return rets;
}

// -------------------------------------------------------- [fix1stLine]
string thermNASA::fix1stLine(const string &s, int asLower) {
  int lens, pos, ia, ic, atmcNbr, numCnt;
  string rets, aSym, aSymOut;

  if ((lens = s.length()) > 80) { rets = s.substr(0, 80); }
  else if (lens < 80)
    { rets = s; while (rets.length() < 80) { rets += ' '; } }
  else { rets = s; }
  for (ia = 0; ia < 4; ia++) {
    pos = ia * 5 + 24;
    aSym = strToUpper(strRemoveSpc(rets.substr(pos, 2)));
    atmcNbr = elmt.atmcNbrBySymb(aSym);
    numCnt = atoi(strRemoveSpc(rets.substr(pos + 2, 3)).c_str());
    if ((atmcNbr > 0) && (numCnt > 0)) {
      if (asLower) { aSymOut = strToLower(aSym); }
      else { aSymOut = aSym; }
      rets.replace(pos, 5, formatLeft(aSymOut, 2) + formatD(numCnt, 3));
    } else { rets.replace(pos, 5, "     "); }
  }
  rets.replace(73, 7, "      1");
  return rets;
}

// ------------------------------------------------------- [fixDataLine]
string thermNASA::fixDataLine(const string &s, int lineNbr) {
  int lens, ndat, id, pos;
  string rets, sfix;

  if ((lens = s.length()) > 80) { rets = s.substr(0, 80); }
  else if (lens < 80)
    { rets = s; while (rets.length() < 80) { rets += ' '; } }
  else { rets = s; }
  if (lineNbr > 3) { ndat = 4; } else { ndat = 5; }
  for (id = 0; id < ndat; id++) {
    pos = 15 * id; sfix = fixCoefField(rets.substr(pos, 15));
    rets.replace(pos, 15, sfix);
  }
  if (ndat == 4) { rets.replace(60, 15, "               "); }
  rets.replace(75, 5, formatD(lineNbr, 5));
  return rets;
}

// ===== thermTable class ================================================

// -------------------------------------------------------------- [setM]
void thermTable::setM() {
  string nam;
  thermNASA tm;
  clear();
  tm.clear();
  nam = "m"; insert(make_pair(nam, tm));
  nam = "M"; insert(make_pair(nam, tm));
}

// ------------------------------------------------ [appendFrom(string)]
int thermTable::appendFrom(string flname) {
  int nbef, naft;
  nbef = size();
  if (append(flname) == ERRE) { return ERRE; }
  naft = size();
  cout << naft - nbef << " thermodata loaded from [" << flname << "]\n";
  return NORM;
}


// ------------------------------------- [appendFrom(string, libPath &)]
int thermTable::appendFrom(string flname, libPath &libp) {
  int nbef, naft;
  nbef = size();
  if (append(libp.fullPath(flname)) == ERRE) { return ERRE; }
  naft = size();
  cout << naft - nbef << " thermodata loaded from [" << flname << "]\n";
  return NORM;
}

// ------------------------------------------------------ [setAllIsUsed]
void thermTable::setAllIsUsed() {
  for (iterator p = begin(); p != end(); p++) { p->second.isUsed = true; }
}

// ------------------------------------------------------ [getUsedAtoms]
set<int> thermTable::getUsedAtoms() {
  atmcComp ac;
  atmcComp::iterator pa;
  set<int> aUsd;
  ac.clear();
  for (iterator p = begin(); p != end(); p++)
    { if (p->second.isUsed) { ac += p->second.atmCmp; } }
  aUsd.clear();
  for (pa = ac.begin(); pa != ac.end(); pa++) { aUsd.insert(pa->first); }
  return aUsd;
}

// ------------------------------------------------------------ [append]
int thermTable::append(string flname) {
  int skip, slen, i;
  string s, nameSpc;
  vector<string> vs;
  thermNASA tnasa;
  ifstream ifs(flname.c_str());
  if (!ifs) { return ERRE; }
  while (!ifs.eof()) {
    skip = false;
    if (getOneLine(ifs, s) == ERRE) { break; }
    slen = s.length();
    if (slen < 3) { skip = true; }
    else if (slen < 6) {
      if (strToUpper(s.substr(0, 3)) == "END") { break; }
      skip = true;
    } else {
      if (strToUpper(s.substr(0, 3)) == "END") { break; }
      if (strToUpper(s.substr(0, 6)) == "THERMO") { skip = true; }
      else if ((s[0] == '!') || (s[0] == ' ')) { skip = true; }
    }
    if (!skip) {
      nameSpc = strTermBy(s, ' ');
      if (nameSpc.length() > 18) { nameSpc = nameSpc.substr(0, 18); }
      vs.clear(); vs.push_back(s);
      for (i = 0; i < 3; i++) {
        if (getOneLine(ifs, s) == ERRE) { ifs.close(); return ERRE; }
        vs.push_back(s);
      }
      if (tnasa.restoreFrom(vs) == ERRE) { ifs.close(); return ERRE; }
      insert(make_pair(nameSpc, tnasa));
    }
  }
  ifs.close();
  return NORM;
}

// ***** rxnInfo and related classes *************************************

// ===== rxnInfo class ===================================================

// --------------------------------------------------------- [rxnInfo()]
rxnInfo::rxnInfo() { clear(); }

// ------------------------------------------------------------- [clear]
void rxnInfo::clear() {
  int it;
  reac.clear(); prod.clear(); rate = "";
  isSumRates = false; factRate.clear(); rateStr.clear();
  shiftE = 0.; factA = 1.; shiftB = 0.;
  isRevrs = false; givenRev = false; modTroe = false;
  modAL = 0.; modbL = 0.; modEaL = 0.;
  modaFc = 0.; modT3Fc = 0.; modT1Fc = 0.; modT2Fc = 0.;
  delH = 0.; delS = 0.;
  kf.clear(); kr.clear();
  for (it = 0; it < NUMTLST; it++)
    { kfTbl[it] = 0.; krTbl[it] = 0.; KeTbl[it] = 0.; }
}

// ------------------------------------------------------- [restoreFrom]
int rxnInfo::restoreFrom(vector<string> &vs, thermTable &t,
 rateMacro &m) {
  int ir, ip;
  vector<string>::iterator p;
  atmcComp acReac, acProd;

  clear(); p = vs.begin();
  while (p != vs.end()) {
    if (*p == "=") { p++; break; }
    else { reac.push_back(*p); }
    p++;
  }
  if (p == vs.end()) { cout << " Parse error.\n"; return ERRE; }
  if ((reac.size() < 1) || (reac.size() > 3)) 
    { cout << " Wrong number of reactants.\n"; return ERRE; }
  acReac.clear();
  for (ir = 0; ir < reac.size(); ir++) {
    if (t.find(reac[ir]) == t.end()) {
      cout << " Unknown reactant [" << reac[ir] << "].\n";
      return ERRE;
    } else { acReac += t[reac[ir]].atmCmp; }
  }
  while (p != vs.end()) {
    if (*p == "$") { p++; break; }
    else { prod.push_back(*p); }
    p++;
  }
  if (p == vs.end()) { cout << " Parse error.\n"; return ERRE; }
  if ((prod.size() < 1) || (prod.size() > 3))
    { cout << " Wrong number of produnts.\n"; return ERRE; }
  acProd.clear();
  for (ip = 0; ip < prod.size(); ip++) {
    if (t.find(prod[ip]) == t.end()) {
      cout << " Unknown product [" << prod[ip] << "].\n";
      return ERRE;
    } else { acProd += t[prod[ip]].atmCmp; }
  }
  if (acReac != acProd)
    { cout << " Wrong Stoichiometry.\n"; return ERRE; }
  rate = *p; p++;
  if (p == vs.end()) { cout << " Parse error.\n"; return ERRE; }
  if (rate[rate.size() - 1] == ']') {
    // --- sum-of-rates (Nov. 19, 2010)
    if (parseSumRates(rate, m) == ERRE)
      { cout << " Wrong sum-of-rates.\n"; return ERRE; }
  } else {
    if (m.find(rate) == m.end())
      { cout << " Unknown macro.\n"; return ERRE; }
  }
  if (isSumRates) { rate = "sum-of-rates"; }
  shiftE = atof((*p).c_str()); p++;
  if (p == vs.end()) { cout << " Parse error.\n"; return ERRE; }
  factA = atof((*p).c_str()); p++;
  if (p == vs.end()) { cout << " Parse error.\n"; return ERRE; }
  isRevrs = valBoolStr(*p); p++;
  if (rate.substr(0, 3) == "Rv/") { givenRev = true; }
  else { givenRev = false; }
  // --- additional input (Apr. 14, 2010)
  if (p == vs.end()) { return NORM; }
  if (((*p)[0] == '!') || ((*p)[0] == '#')) { return NORM; }
  shiftB = atof((*p).c_str()); p++;
  // --- additional input (Jul. 7, 2010)
  if (rate.substr(0, 5) == "Troe/") {
    if (p == vs.end()) { return NORM; }
    if (((*p)[0] == '!') || ((*p)[0] == '#')) { return NORM; }
     // seven parameters are expected
    modTroe = true;
    modAL = atof((*p).c_str()); p++;
    modbL = atof((*p).c_str()); p++;
    modEaL = atof((*p).c_str()); p++;
    modaFc = atof((*p).c_str()); p++;
    modT3Fc = atof((*p).c_str()); p++;
    modT1Fc = atof((*p).c_str()); p++;
    modT2Fc = atof((*p).c_str()); p++;
  }
  // ---
  return NORM;
}

// ------------------------------------------------------ [setRateConst]
int rxnInfo::setRateConst(thermTable &t, rateMacro &m, double Ts[]) {
  int it;
  if (isSumRates) { return setSumOfRates(t, m, Ts); }
                           // ... temporarily swap reac and prod
  if (givenRev) { swap(reac, prod); }
  setTherm298(t);
  kf = m[rate]; kf.A *= factA; kf.Ea += shiftE; kf.b += shiftB;
  if (modTroe) {
    kf.AL = modAL; kf.bL = modbL; kf.EaL = modEaL; kf.aFc = modaFc;
    kf.T3Fc = modT3Fc; kf.T1Fc = modT1Fc; kf.T2Fc = modT2Fc;
  }
  if ((!isRevrs) && (!givenRev))
    { calRatesFowerdOnly(Ts); return NORM; }
  if (calRatesEquil(t, Ts) == ERRE) { return ERRE; }
  if (givenRev) {          // ... re-swap reac and prod
    swap(kf, kr); swap(reac, prod); delH = -delH; delS = -delS;
    for (it = 0; it < NUMTLST; it++)
      { swap(kfTbl[it], krTbl[it]); KeTbl[it] = 1. / KeTbl[it]; }
  }
  return NORM;
}

// ------------------------------------------------------ [outputToChem]
void rxnInfo::outputToChem(ofstream &ofs, double Ts[]) {
  int i, ir, ip;
  string s = " ";
  for (ir = 0; ir < reac.size(); ir++) {
    s += reac[ir];
    if (ir == (reac.size() - 1)) {
      if (kf.isTroe) { s += "(+M)"; }
      if (isRevrs) { s += "="; } else { s += "=>"; }
    }
    else { s += "+"; }
  }
  for (ip = 0; ip < prod.size(); ip++) {
    s += prod[ip];
    if (ip < (prod.size() - 1)) { s += "+"; }
  }
  if (kf.isTroe) { s += "(+M)"; }
  ofs << formatLeft(s, 48) << kf.format() << endl;
  cout << formatLeft(s, 48) << formatF(delH, 10, 3)
       << formatF(delS, 10, 3) << endl;
  if (kf.isTroe) {
    ofs << formatRight("low /", 48) << kf.formatLow() << " /\n";
    ofs << formatRight("troe /", 28) << kf.formatTroe() << " /\n";
  } else if (isRevrs) {
    ofs << formatRight("rev /", 48) << kr.format() << " /\n";
  }
  cout << "---------- rate calculations ----------\n";
  cout << " rev /" << kr.format() << " /\n";
  cout << "  T/K        kf         kr         Kc \n";
  for (i = 0; i < NUMTLST; i++) {
    cout << formatFnoDot(Ts[i], 5) << formatE(kfTbl[i], 11, 3);
    if (isRevrs)
      { cout << formatE(krTbl[i], 11, 3) << formatE(KeTbl[i], 11, 3); }
    cout << endl;
  }
  cout << "---------------------------------------\n";
}

// ----------------------------------------------------- [parseSumRates]
int rxnInfo::parseSumRates(string &s, rateMacro &m) {
  int pos = 0, pfound;
  double fact;
  string sfac, srat;
  while (pos < s.length()) {
    if (s[pos] == '+') { pos++; }
    pfound = s.find('[', pos);
    if (pfound == string::npos) { break; }
    sfac = s.substr(pos, pfound - pos);
    if (sfac.length() > 0) { fact = atof(sfac.c_str()); }
    else { fact = 1.; }
    pos = pfound + 1;
    pfound = s.find(']', pos);
    if (pfound == string::npos) { return ERRE; }
    srat = s.substr(pos, pfound - pos);
    if (m.find(srat) == m.end()) { return ERRE; }
    factRate.push_back(fact);
    rateStr.push_back(srat);
    pos = pfound + 1;
  }
  if (rateStr.empty()) { return ERRE; }
  isSumRates = true;
  return NORM;
}

// ----------------------------------------------------- [setSumOfRates]
int rxnInfo::setSumOfRates(thermTable &t, rateMacro &m, double Ts[]) {
  int isum, nsum = rateStr.size(), it;
  rxnRate k1;
   // --- fit forward rate (temporary in krTbl)
  for (it = 0; it < NUMTLST; it++) { krTbl[it] = 0.; }
  for (isum = 0; isum < nsum; isum++) {
    k1 = m[rateStr[isum]];
    for (it = 0; it < NUMTLST; it++) {
      krTbl[it] += factRate[isum] * k1.A * pow(Ts[it], k1.b) * exp(-k1.Ea / RGASCAL / Ts[it]);
    }
  }
  if (fitRevRate(Ts) == ERRE) { return ERRE; }
  kf = kr; kr.clear();
  for (it = 0; it < NUMTLST; it++) { kfTbl[it] = krTbl[it]; krTbl[it] = 0.; }
  setTherm298(t);
  if (!isRevrs) { calRatesFowerdOnly(Ts); return NORM; }
   // --- equilibrium const. and reverse rates
  if (calRatesEquil(t, Ts) == ERRE) { return ERRE; }
  return NORM;
}

// ------------------------------------------------------- [setTherm298]
void rxnInfo::setTherm298(thermTable &t) {
  int ir, ip;
  double T = 298.15, dHRT = 0., dSR = 0., HRT, SR, CpR;
  for (ir = 0; ir < reac.size(); ir++) {
    t[reac[ir]].getThermFuncs(T, HRT, SR, CpR); t[reac[ir]].isUsed = true;
    dHRT -= HRT; dSR -= SR;
  }
  for (ip = 0; ip < prod.size(); ip++) {
    t[prod[ip]].getThermFuncs(T, HRT, SR, CpR); t[prod[ip]].isUsed = true;
    dHRT += HRT; dSR += SR;
  }
  delH = dHRT * RGASCAL * T / 1000.;
  delS = dSR * RGASCAL;
}

// ----------------------------------------------------- [calRatesEquil]
int rxnInfo::calRatesEquil(thermTable &t, double Ts[]) {
  int it, ir, ip, dn;
  double dHRT, dSR, HRT, SR, CpR, dGRT, Kp, Kc;
  dn = prod.size() - reac.size();
  for (it = 0; it < NUMTLST; it++) {
    dHRT = 0.; dSR = 0.;
    for (ir = 0; ir < reac.size(); ir++) {
      t[reac[ir]].getThermFuncs(Ts[it], HRT, SR, CpR);
      dHRT -= HRT; dSR -= SR;
    }
    for (ip = 0; ip < prod.size(); ip++) {
      t[prod[ip]].getThermFuncs(Ts[it], HRT, SR, CpR);
      dHRT += HRT; dSR += SR;
    }
    dGRT = dHRT - dSR; Kp = exp(-dGRT);
    Kc = pow((1. / (RATMMOL * Ts[it])), dn) * Kp;
    KeTbl[it] = Kc;
    kfTbl[it] = kf.A * pow(Ts[it], kf.b) * exp(-kf.Ea / RGASCAL / Ts[it]);
    krTbl[it] = kfTbl[it] / KeTbl[it];
  }
  if (fitRevRate(Ts) == ERRE) { return ERRE; }
  return NORM;
}

// ------------------------------------------------ [calRatesFowerdOnly]
void rxnInfo::calRatesFowerdOnly(double Ts[]) {
  int it;
  for (it = 0; it < NUMTLST; it++) {
    kfTbl[it] = kf.A * pow(Ts[it], kf.b) * exp(-kf.Ea / RGASCAL / Ts[it]);
  }
}

// -------------------------------------------------------- [fitRevRate]
int rxnInfo::fitRevRate(double Ts[]) {
  double Qvec[3], Bmat[3][6], Avec[3];

  fitRRsetMat(Qvec, Bmat, Ts);
  if (fitRRinvB(Bmat) == ERRE)
    { cout << " Numerical error in fitting kr.\n"; return ERRE; }
  fitRRcalA(Qvec, Bmat, Avec);
  kr.A = exp(Avec[0]); kr.b = Avec[1]; kr.Ea = Avec[2] * RGASCAL;
  return NORM;
}

// -------------------------------------------------- [fitRRsetMat]
void rxnInfo::fitRRsetMat(double q[], double B[][6], double Ts[]) {
  int iRow, iCol, iDat;
  double x, y, f[3];

  for (iRow = 0; iRow < 3; ++iRow) {
    q[iRow] = 0.;
    for (iCol = 0; iCol < 6; ++iCol) B[iRow][iCol] = 0.;
  }
  for (iDat = 0; iDat < NUMTLST; ++iDat) {
    x = 1. / Ts[iDat]; y = log(krTbl[iDat]);
    f[0] = 1.; f[1] = -log(x); f[2] = -x;
    for (iRow = 0; iRow < 3; ++iRow) {
      q[iRow] += y * f[iRow];
      for (iCol = iRow; iCol < 3; ++iCol)
        { B[iRow][iCol] += f[iRow] * f[iCol]; }
    }
  }
  for (iRow = 1; iRow < 3; ++iRow) {
    for (iCol = 0; iCol < iRow; ++iCol) B[iRow][iCol] = B[iCol][iRow];
  }
  for (iRow = 0; iRow < 3; ++iRow) B[iRow][iRow + 3] = 1.;
}

// ---------------------------------------------------- [fitRRinvB]
int rxnInfo::fitRRinvB(double B[][6]) {
  int iRow, jRow, iPiv, iCol;
  double pivot, pres;

  for (iRow = 0; iRow < 3; ++iRow) {
    pivot = 0.;
    for (jRow = iRow; jRow < 3; ++jRow) {
      if (fabs(B[jRow][iRow]) > pivot)
        { pivot = fabs(B[jRow][iRow]); iPiv = jRow; }
    }
    if (fabs(pivot) < 1e-80) return ERRE;
    pivot = B[iPiv][iRow];
    for (iCol = iRow; iCol < 6; ++iCol) {
      pres = B[iPiv][iCol] / pivot;
      B[iPiv][iCol] = B[iRow][iCol]; B[iRow][iCol] = pres;
    }
    for (jRow = 0; jRow < 3; ++jRow) {
      if (jRow != iRow) {
        pres = B[jRow][iRow];
        for (iCol = iRow; iCol < 6; ++iCol) {
          B[jRow][iCol] -= pres * B[iRow][iCol];
        }
      }
    }
  }
  for (iRow = 0; iRow < 3; ++iRow) {
    for (iCol = 0; iCol < 3; ++iCol) B[iRow][iCol] = B[iRow][iCol + 3];
  }
  return NORM;
}

// ---------------------------------------------------- [fitRRcalA]
void rxnInfo::fitRRcalA(double q[], double B[][6], double a[]) {
  int iRow, iCol;

  for (iRow = 0; iRow < 3; ++iRow) {
    a[iRow] = 0.;
    for (iCol = 0; iCol < 3; ++iCol) a[iRow] += B[iRow][iCol] * q[iCol];
  }
}

