/* -------------------------------------------------------------------------
     "trangen.cc" - transport data generator
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2015 by A. Miyoshi, University of Tokyo
                                       created: Jan.  25, 2011
                                        edited: Feb.   4, 2011
                                   last edited: Jul.   3, 2015
                        - '--append_nox' option was added
 --------------------------------------------------------------------------
 usage: trangen inThgFile [--no_append|--append_nox]
------------------------------------------------------------------------- */

#include "joback.h"
using namespace std;

#define FL_BASTRN "bas_tran.dat"        // base transport data file
#define FL_NOXTRN "nox_tran.dat"        // NOx transport data file
#define FN_OUTTRN "_trn.dat"            // transport data output

int parseOneThgLine(ifstream &ifs, string &mol, jobackGrpList &jl);
void appendTrnData(ofstream &ofs, string apndfln, libPath &libp);

/* ===== main ========================================================= */

int main(int argc, char *argv[]) {
  int no_append = false, append_nox = false;
  string fnThgIn, fnBas, fnTrnOut, s, molnam, clopt = "";
  double epskB, sigma;
  libPath libp;
  jobackGrpList jbkGL;
  jobackLibrary jbkLib;

  showRevision();
  jbkLib.loadFrom(FL_TRNGCM, libp);
  if ((argc < 2) || (argc > 3))
    { cout << "usage: trangen inThgFile [--no_append|--append_nox]\n";
     return 1; }
  if (argc == 3) {
    clopt = argv[2];
    if (clopt == "--no_append") { no_append = true; }
    else if (clopt == "--append_nox") { append_nox = true; }
    else { cout << "usage: trangen inThgFile [--no_append|--append_nox]\n";
     return 1; }
  }
  fnThgIn = argv[1];
  fnBas = getBasePartEndWith(fnThgIn, "_thg.inp");
  fnTrnOut = fnBas + FN_OUTTRN;

  ifstream iThg(fnThgIn.c_str());
  if (!iThg) { cout << "failed to open [" << fnThgIn << "].\n"; return 1; }
  if (getOneLine(iThg, s) == ERRE) {
    cout << "error reading [" << fnThgIn << "].\n"; iThg.close(); return 1;
  }
  ofstream oTrn(fnTrnOut.c_str());
  if (!oTrn) {
    cout << "failed to open [" << fnTrnOut << "].\n";
    iThg.close(); return 1;
  }
  printHeadingInfo(oTrn);
  oTrn << "! Transport data generated by trangen (" << strCurDate() << ")\n";
  while (parseOneThgLine(iThg, molnam, jbkGL) == NORM) {
    if (jbkLib.estimate(jbkGL, epskB, sigma) == ERRE) { break; }
    oTrn << formatLeft(molnam, 18) << " 2" << formatF(epskB, 8, 1)
         << " " << formatF(sigma, 9, 3) << formatF(0., 8, 1) << "  "
         << formatF(0., 8, 1) << " " << formatF(1., 8, 1) << endl;
  }
  iThg.close();
  if (!no_append) {
    oTrn << "! Transport data for base mechanism species\n";
    appendTrnData(oTrn, FL_BASTRN, libp);
  }
  if (append_nox) {
    oTrn << "! Transport data for NOx mechanism species\n";
    appendTrnData(oTrn, FL_NOXTRN, libp);
  }
  oTrn.close();
  return 0;
}

/* ----- parse a thermgen input line ---------------------------------- */

int parseOneThgLine(ifstream &ifs, string &mol, jobackGrpList &jl) {
  int ntkn, itkn, numg;
  string s, gnam;
  vector<string> vs;
  if (getOneLine(ifs, s) == ERRE) { return ERRE; }
  jl.clear();
  ntkn = splitToken(s, " ,\t", vs);
  mol = vs[0]; itkn = 1;
  while (vs[itkn] != "$") { itkn++; }
  do {
    itkn++; gnam = vs[itkn];
    if (gnam != "$") {
      itkn++; numg = atoi(vs[itkn].c_str());
      jl[gnam] += numg;
    }
  } while (gnam != "$");
  return NORM;
}

/* ----- append transport data in libPath ----------------------------- */

void appendTrnData(ofstream &ofs, string apndfln, libPath &libp) {
  int skip, slen;
  string s;
  ifstream ifs(libp.fullPath(apndfln).c_str());
  if (!ifs) { return; }
  while (!ifs.eof()) {
    if (getOneLine(ifs, s) == ERRE) { break; }
    skip = false; slen = s.length();
    if (slen < 3) { skip = true; }
    else if (s[0] == '!') { skip = true; }
    if (!skip) { ofs << s << endl; }
  }
  ifs.close();
  return;
}

