/*--------------------------------------------------------------------------
     "molecule.cc" - molecular identification and property classes
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2015 by A. Miyoshi, University of Tokyo
                                       created: Aug.   4, 2002
                                        edited: Apr.  13, 2005
                                        edited: Jan.   7, 2011
                                   last edited: Dec.   4, 2015
--------------------------------------------------------------------------*/

#include "molecule.h"

/***************************************************************************
     'alkaneChain' and related classes
***************************************************************************/

// ===== alkaneSubChain ====================================================

alkaneSubChain::alkaneSubChain() { clear(); }          // --- constructors

alkaneSubChain::alkaneSubChain(int lS, int pCI, int pS, int lC)
  { levSubst = lS; parentChIdx = pCI; posSubInPCh = pS; lenChain = lC; }

void alkaneSubChain::clear()
  { levSubst = 0; parentChIdx = -1; posSubInPCh = 0; lenChain = 0; }

// ===== alkaneChain =======================================================

alkaneChain::alkaneChain() {                            // --- constructor
  libPath libp;
  clear();
  mcr.init(ALLALP, "-");
  if (mcr.loadFrom(FL_ALKMCR, libp) == ERRE)
    { cout << "Failed to load alkane macros.\n"; }
}

void alkaneChain::clear()                                     // --- clear
  { alkaneName = ""; isCyclic = false; numChains = 0; chain.clear(); }

void alkaneChain::addChain(alkaneSubChain sc)              // --- addChain
  { chain.push_back(sc); numChains++; }

int alkaneChain::setFrom(string &strInp) {                  // --- setFrom
  int retc;
  string s;
  clear();
  // cout << "fuel input : " << strInp << endl;
  if (extrHead(strInp) == ERRE) { return ERRE; }
  if (mcr.substMacros(strInp) == ERRE) { return ERRE; }
  // cout << "macro subs : " << strInp << endl;
  if (strInp[0] == '-')
    { s = "0-" + strInp.substr(1); isCyclic = true; }
  else
    { s = "0-" + strInp; isCyclic = false; }
  retc = parseStr(0, -1, s);
  return retc;
}

void alkaneChain::print(ostream *os) {                        // --- print
  *os << "----- alkaneChain print -----\n";
  *os << "name : " << alkaneName << endl;
  for (int i = 0; i < numChains; i++) {
    *os << chain[i].levSubst << ", " << chain[i].parentChIdx << ", "
        << chain[i].posSubInPCh << ", " << chain[i].lenChain << endl;
  }
  *os << "-----------------------------\n";
}

int alkaneChain::extrHead(string &s) {                     // --- extrHead
  // extract header from fuel input string
  int pos;
  if (s.find('[') == 0) {
    if ((pos = s.find(']')) == string::npos) { return ERRE; }
    alkaneName.assign(s, 1, pos - 1);
    s.erase(0, pos + 1);
  } else {
    alkaneName = "";
  }
  return NORM;
}

int alkaneChain::parseStr(int lev, int pIx, string &s) {   // --- parseStr
  int posHyph, posBra, posCket;
  int posSub, lenCh;
  int levNext, pIxNext;
  string partStr, sNext;
  if ((posHyph = s.find_first_of("-")) == string::npos) { return ERRE; }
  if (posHyph == 0) { return ERRE; }
  partStr = s.substr(0, posHyph);
  posSub = atoi(partStr.c_str());
  if ((posBra = s.find_first_of("(")) == string::npos)
    { posBra = s.length(); }
  if (posBra <= (posHyph + 1)) { return ERRE; }
  partStr = s.substr(posHyph + 1, posBra - posHyph - 1);
  lenCh = atoi(partStr.c_str());
  addChain(alkaneSubChain(lev, pIx, posSub, lenCh));
  levNext = lev + 1; pIxNext = numChains - 1;
  while (posBra < s.length()) {
    if ((posCket = nextCorCket(s, posBra + 1)) < 0) { return ERRE; }
    if (posCket <= (posBra + 1)) { return ERRE; }
    sNext = s.substr(posBra + 1, posCket - posBra - 1);
    if (parseStr(levNext, pIxNext, sNext) == ERRE) { return ERRE; }
    if ((posBra = s.find_first_of("(", posCket + 1)) == string::npos)
      { posBra = s.length(); }
  }
  return NORM;
}

ostream &operator<<(ostream &os, alkaneChain &ac)        // --- operator<<
  { ac.print(&os); return os; }


/*****************************************************************************
     'molecByChains' and related classes
*****************************************************************************/


/*============================================================================
     'atomChain' class
============================================================================*/


atomChain::atomChain() { numAtoms = 0; }

atomChain::atomChain(int ic, int isc, int n, vector<int> &ai, molecInfo &m)
  { set(ic, isc, n, ai, m); }

int atomChain::identLen(atomChain &ref) {
  int nmax, ilen;
  if (numAtoms < ref.numAtoms) { nmax = numAtoms; }
  else { nmax = ref.numAtoms; }
  for (ilen = 0; ilen < nmax; ilen++) {
    if (atomIndex[ilen] != ref.atomIndex[ilen]) { break; }
  }
  return ilen;
}

void atomChain::print(ostream *os) {
  if (isCyclic) { *os << "Cyclic   : "; }
  else { *os << "Straight : "; }
  *os << numAtoms << " : " << viewStr << " : " << sortKey << " : ";
  for (int i = 0; i < numAtoms; i++) {
    *os << format(atomIndex[i], 2);
    if (i == (numAtoms - 1)) { *os << endl; }
    else { *os << ","; }
  }
}

int atomChain::storeTo(ofstream &ofs) {
  ofs << "    atomChain{ " << isCyclic << " " << isSubChOfCyc << " "
      << numAtoms << " atomIndex[ ";
  for (int i = 0; i < numAtoms; i++) { ofs << atomIndex[i] << " "; }
  ofs << "] }\n";
  return NORM;
}

// --------------------------------------------------------- [restoreFrom]
int atomChain::restoreFrom(ifstream &ifs, molecInfo &m) {
  int ic, isc, n;
  string s;
  vector<int> ai;

  ai.clear();
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "atomChain{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  ic = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  isc = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  n = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "atomIndex[") { return ERRE; }
  for (int i = 0; i < n; i++) {
    if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
    ai.push_back(atoi(s.c_str()));
  }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "]") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  set(ic, isc, n, ai, m);
  return NORM;
}

int atomChain::isAromatic() {   // -------------------------- [isAromatic]
  static string aromaStr[2] = {"121212", "212121"};
  int ia, istr;
  string ringStr = "";
  if (!isCyclic) { return false; }
  for (ia = 0; ia < numAtoms; ia++)
    { ringStr += sortKey.substr(ia * 5 + 7, 1); }
  for (istr = 0; istr < 2; istr++) {
    if (ringStr == aromaStr[istr]) { return true; }
  }
  return false;
}

int atomChain::isIdxInChain(int idx) {   // --------------- [isIdxInChain]
  for (int i = 0; i < numAtoms; i++) {
    if (atomIndex[i] == idx) { return true; }
  }
  return false;
}

// ---------------------------- [set(int,int,int,vector<int>&,molecInfo&)]
void atomChain::set(int ic, int isc, int n, vector<int> &ai,
 molecInfo &m) {
  int i, idxa, idxn, aNum, aAtt, nBon, nSub, BOtoNext;
  isCyclic = ic; isSubChOfCyc = isc; numAtoms = n;
  sortKey = format(numAtoms, 3);
  viewStr = "";
  atomIndex.resize(numAtoms);
  for (i = 0; i < numAtoms; i++) { atomIndex[i] = ai[i]; }
  for (i = 0; i < numAtoms; i++) {
    idxa = atomIndex[i]; aNum = m.atom[idxa].atomicNumber;
    aAtt = m.atom[idxa].atomAttrib; nBon = m.atom[idxa].numBonds;
    nSub = nBon - 2;
    if ((!isCyclic) && (i == numAtoms - 1)) { nSub++; }
    if (((!isCyclic) && (!isSubChOfCyc)) && (i == 0)) { nSub++; }
    sortKey += format(aNum, 2) + format(aAtt, 1) + format(nSub, 1);
    viewStr += format(elmt[aNum].atomicSymbol, 2) + rpsAtmAttr(aAtt);
    if (isCyclic || (i < (numAtoms - 1))) {
      if (i == (numAtoms - 1)) { idxn = atomIndex[0]; }
      else { idxn = atomIndex[i + 1]; }
      for (int j = 0; j < nBon; j++) {
        if (m.atom[idxa].bond[j].toAtomIndex == idxn)
          { BOtoNext = m.atom[idxa].bond[j].order; break; }
      }
      sortKey += format(BOtoNext, 1);
      viewStr += rpsBndOrdr(BOtoNext);
    }
  }
}

ostream &operator<<(ostream &os, atomChain &ac)
  { ac.print(&os); return os; }


/*============================================================================
     'treeInfoChains' class
============================================================================*/

// ------------------------------------------------------------ [addILset]
void treeInfoChains::addILset(int nC, int iL, int nS) {
  vector<int> numChs;
  numChs.clear(); numChs.resize(MAXN_BONDS, 0);
  numChsInILs.push_back(nC); identLOfILs.push_back(iL);
  numSubsInILs.push_back(nS); numChsInSub.push_back(numChs);
  numILsets++;
}


/*============================================================================
     'molecByChains' class
============================================================================*/


molecByChains::molecByChains() { numChains = 0; }

molecByChains::molecByChains(int start, int num, molecByChains &mc)
  { copyPartial(start, num, mc); }

void molecByChains::clear()   // --------------------------------- [clear]
  { numChains = 0; chain.clear(); }

void molecByChains::addChain(atomChain ac)
  { chain.push_back(ac); numChains++; }

void molecByChains::beUnique() {   // ------------------------- [beUnique]
  int ic, numIdCh;
  vector<molecByChains> chcase;
  molecByChains mx;

  chcase.clear();
  if (numChains < 2) { return; }
  numIdCh = sortByKey();
  if ((numIdCh < 2) && (numChains < 3)) { return; }
  for (ic = 0; ic < numIdCh; ic++) {
    chcase.push_back(*this);
    if (ic > 1) { chcase[ic].swapChains(1, ic); }
    chcase[ic].sortTree();
  }
  mx = chcase[0];
  for (ic = 1; ic < numIdCh; ic++)
    { if (chcase[ic] > mx) { mx = chcase[ic]; } }
  *this = mx;
}

// ------------------------------------------------------------ [sortTree]
void molecByChains::sortTree() {
  int iIL, iSb, jSb, i, j;
  int idLeni, idLenj, ixAtomi, ixAtomj;
  if (numChains < 3) { return; }
  atomChain r = chain[0];
  for (i = 1; i < (numChains - 1); i++) {
    for (j = i + 1; j < numChains; j++) {
      idLeni = chain[i].identLen(r); idLenj = chain[j].identLen(r);
      if (idLenj < idLeni) {
        swapChains(i, j);
      } else if (idLenj == idLeni) {
        ixAtomi = chain[i].atomIndex[idLeni];
        ixAtomj = chain[j].atomIndex[idLenj];
        if (ixAtomj < ixAtomi) {
          swapChains(i, j);
        } else if (ixAtomj == ixAtomi) {
          if (chain[j].sortKey > chain[i].sortKey)
            { swapChains(i, j); }
        }
      }
    }
  }
  treeInfoChains ti;
  getTreeInfo(ti);
  int ofsChILs = 1, nChILs, ofsChSub, nChSub, start;
  vector<molecByChains> subc;
  molecByChains mbc;
  for (iIL = 0; iIL < ti.numILsets; iIL++) {
    nChILs = ti.numChsInILs[iIL];
    ofsChSub = 0;
    subc.clear();
    for (iSb = 0; iSb < ti.numSubsInILs[iIL]; iSb++) {
      nChSub = ti.numChsInSub[iIL][iSb];
      start = ofsChILs + ofsChSub;
      mbc.clear(); mbc.copyPartial(start, nChSub, *this);
      subc.push_back(mbc);
      if (nChSub > 1) { subc[iSb].beUnique(); }
      ofsChSub += nChSub;
    }
    for (iSb = 0; iSb < (ti.numSubsInILs[iIL] - 1); iSb++) {
      for (jSb = iSb + 1; jSb < ti.numSubsInILs[iIL]; jSb++) {
        if (subc[jSb] > subc[iSb])
          { mbc = subc[iSb]; subc[iSb] = subc[jSb]; subc[jSb] = mbc; }
      }
    }
    ofsChSub = 0;
    for (iSb = 0; iSb < ti.numSubsInILs[iIL]; iSb++) {
      nChSub = subc[iSb].numChains;
      start = ofsChILs + ofsChSub;
      replace(start, nChSub, subc[iSb]);
      ofsChSub += nChSub;
    }
    ofsChILs += nChILs;
  }
}

// --------------------------------------------------------- [getRenumTbl]
void molecByChains::getRenumTbl(int n, vector<int> &sortTbl,
 vector<int> &trnsTbl) {
  int cNewNbr = 0, i, j, idx;
  sortTbl.clear(); trnsTbl.clear();
  trnsTbl.assign(n, -1);
  for (i = 0; i < numChains; i++) {
    for (j = 0; j < chain[i].numAtoms; j++) {
      idx = chain[i].atomIndex[j];
      if (trnsTbl[idx] < 0) {
        sortTbl.push_back(idx); trnsTbl[idx] = cNewNbr; cNewNbr++;
      }
    }
  }
}

// ---------------------------------------------------------- [renumAtoms]
void molecByChains::renumAtoms(vector<int> &trnsTbl) {
  int i, j, oldIdx, newIdx;
  for (i = 0; i < numChains; i++) {
    for (j = 0; j < chain[i].numAtoms; j++) {
      oldIdx = chain[i].atomIndex[j];
      newIdx = trnsTbl[oldIdx];
      chain[i].atomIndex[j] = newIdx;
    }
  }
}

void molecByChains::print(ostream *os) {                  // ----- [print]
  *os << "----- molecByChains (" << numChains << ") -----\n";
  for (int i = 0; i < numChains; i++) { *os << chain[i]; }
}

void molecByChains::getStrucView(vector<string> &vs) { // - [getStrucView]
  if (!vs.empty()) { vs.clear(); }
  addChainView(vs, 0);
}

                                                   // ----- [addChainView]
void molecByChains::addChainView(vector<string> &vs, int clrLen) {
  int top = vs.size(), numSpc, posVerBar;
  string s;
  s = chain[0].viewStr;
  if (clrLen > 0) {
    numSpc = clrLen * 4 - 1; posVerBar = numSpc - 2;
    s.replace(0, numSpc, numSpc, ' ');
    s[posVerBar] = '|'; s[posVerBar + 1] = '_';
  }
  vs.push_back(s);
  treeInfoChains ti;
  getTreeInfo(ti);
  int ofsChILs, ofsChSub, nChSub, start, idLen;
  molecByChains *subc;
  for (int iIL = ti.numILsets - 1; iIL >= 0; iIL--) {
    ofsChILs = 1;
    for (int i = 0; i < iIL; i++)
      { ofsChILs += ti.numChsInILs[i]; }
    idLen = ti.identLOfILs[iIL];
    ofsChSub = 0;
    for (int iSb = 0; iSb < ti.numSubsInILs[iIL]; iSb++) {
      if (iSb == (ti.numSubsInILs[iIL] - 1)) {
        posVerBar = idLen * 4 - 3;
        for (int iStr = top + 1; iStr < vs.size(); iStr++)
          { (vs[iStr])[posVerBar] = '|'; }
      }
      nChSub = ti.numChsInSub[iIL][iSb];
      start = ofsChILs + ofsChSub;
      subc = new molecByChains(start, nChSub, *this);
      subc->addChainView(vs, idLen);
      delete subc;
      ofsChSub += nChSub;
    }
  }
}

void molecByChains::swapChains(int a, int b) {           // --- swapChains
  atomChain p;
  if ((a < 0) || (a >= numChains)) { return; }
  if ((b < 0) || (b >= numChains)) { return; }
  p = chain[a]; chain[a] = chain [b]; chain[b] = p;
}

int molecByChains::getCommonLen(int iCa, int iCb) {    // --- getCommonLen
  // get length of common part of chain iCa and iCb
  int lencom = 0, na, ia;
  if ((iCa < 0) || (iCa >= numChains) || (iCb < 0) || (iCb >= numChains))
    { return lencom; }
  na = chain[iCa].numAtoms;
  if (chain[iCb].numAtoms < na) { na = chain[iCb].numAtoms; }
  for (ia = 0; ia < na; ia++) {
    if (chain[iCa].atomIndex[ia] == chain[iCb].atomIndex[ia]) { lencom++; }
    else { break; }
  }
  return lencom;
}

                                                     // --- getChainOrigin
int molecByChains::getChainOrigin(int iTch, int &pos) {
  int och = -1, ic, lcom, mxlcom, mxlch;
  pos = -1;
  if ((iTch < 1) || (iTch >= numChains)) { return och; }
  mxlch = 0; mxlcom = getCommonLen(mxlch, iTch);
  for (ic = 1; ic < iTch; ic++) {
    lcom = getCommonLen(ic, iTch);
    if (lcom > mxlcom) { mxlcom = lcom; mxlch = ic; }
  }
  och = mxlch; pos = mxlcom - 1;
  return och;
}

int molecByChains::operator>(molecByChains &rm) {   // ------ [operator >]
  int numc, i;
  string skl, skr;
  if (numChains < rm.numChains) { numc = numChains; }
  else { numc = rm.numChains; }
  for (i = 0; i < numc; i++) {
    skl = chain[i].sortKey; skr = rm.chain[i].sortKey;
    if (skl > skr) { return true; }
    else if (skl < skr) { return false; }
  }
  if (numChains > numc) { return true; }
  return false;
}

int molecByChains::operator<(molecByChains &rm) {   // ------ [operator <]
  int numc, i;
  string skl, skr;
  if (numChains < rm.numChains) { numc = numChains; }
  else { numc = rm.numChains; }
  for (i = 0; i < numc; i++) {
    skl = chain[i].sortKey; skr = rm.chain[i].sortKey;
    if (skl < skr) { return true; }
    else if (skl > skr) { return false; }
  }
  if (rm.numChains > numc) { return true; }
  return false;
}

int molecByChains::operator==(molecByChains &rm) {   // ---- [operator ==]
  int numc, i;
  string skl, skr;
  if (numChains != rm.numChains) { return false; }
  numc = numChains;
  for (i = 0; i < numc; i++) {
    skl = chain[i].sortKey; skr = rm.chain[i].sortKey;
    if (skl != skr) { return false; }
  }
  return true;
}

int molecByChains::operator!=(molecByChains &rm) {   // ---- [operator !=]
  int numc, i;
  string skl, skr;
  if (numChains != rm.numChains) { return true; }
  numc = numChains;
  for (i = 0; i < numc; i++) {
    skl = chain[i].sortKey; skr = rm.chain[i].sortKey;
    if (skl != skr) { return true; }
  }
  return false;
}

int molecByChains::storeTo(ofstream &ofs) {   // --------------- [storeTo]
  ofs << "  molecByChains{ " << numChains << endl;
  for (int i = 0; i < numChains; i++) {
    if(chain[i].storeTo(ofs) == ERRE) { return ERRE; }
  }
  ofs << "  }\n";
  return NORM;
}

                   // -------------------------------------- [restoreFrom]
int molecByChains::restoreFrom(ifstream &ifs, molecInfo &m) {
  int numCh;
  string s;
  atomChain ac;
  clear();
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "molecByChains{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numCh = atoi(s.c_str());
  for (int i = 0; i < numCh; i++) {
    if (ac.restoreFrom(ifs, m) == ERRE) { return ERRE; }
    addChain(ac);
  }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}

                    // ------------------------------------- [copyPartial]
void molecByChains::copyPartial(int start, int num, molecByChains &mc) {
  int numc, ic;
  clear();
  numc = num;
  if ((start + numc) > mc.numChains) { numc = mc.numChains - start; }
  for (ic = 0; ic < numc; ic++) { addChain(mc.chain[start + ic]); }
}

                    // ----------------------------------------- [replace]
void molecByChains::replace(int start, int num, molecByChains &mc) {
  int numr = num, i;
  if (numr > mc.numChains) { numr = mc.numChains; }
  for (i = 0; i < numr; i++) { chain[start + i] = mc.chain[i]; }
}

int molecByChains::sortByKey() {   // ------------------------ [sortByKey]
  int numIdentChains = 1, i, j;
  if (numChains < 2) { return numChains; }
  for (i = 0; i < (numChains - 1); i++) {
    for (j = i + 1; j < numChains; j++) {
      if (chain[j].sortKey > chain[i].sortKey) { swapChains(i, j); }
    }
  }
  for (i = 1; i < numChains; i++) {
    if (chain[i].sortKey == chain[0].sortKey) { numIdentChains++; }
  }
  return numIdentChains;
}

// --------------------------------------------------------- [getTreeInfo]
void molecByChains::getTreeInfo(treeInfoChains &ti) {
  atomChain r = chain[0];
  int lastIL = r.numAtoms, idLen;
  ti.numILsets = 0;
  for (int ich = 1; ich < numChains; ich++) {
    idLen = chain[ich].identLen(r);
    if (idLen == lastIL) { ti.numChsInILs[ti.numILsets - 1]++; }
    else { ti.addILset(1, idLen, 0); }
    lastIL = idLen;
  }
  int ofsChIdx = 1, nCh, lastIDatom, iDatom;
  for (int iIL = 0; iIL < ti.numILsets; iIL++) {
    nCh = ti.numChsInILs[iIL];
    idLen = ti.identLOfILs[iIL];
    ti.numSubsInILs[iIL] = 0; lastIDatom = -1;
    for (int ich = 0; ich < nCh; ich++) {
      iDatom = chain[ich + ofsChIdx].atomIndex[idLen];
      if (iDatom == lastIDatom) {
        ti.numChsInSub[iIL][ti.numSubsInILs[iIL] - 1]++;
      } else {
        ti.numSubsInILs[iIL]++;
        ti.numChsInSub[iIL][ti.numSubsInILs[iIL] - 1] = 1;
      }
      lastIDatom = iDatom;
    }
    ofsChIdx += nCh;
  }
}

ostream &operator<<(ostream &os, molecByChains &mc)   // --- [operator <<]
  { mc.print(&os); return os; }


/*****************************************************************************
     group additivity classes
*****************************************************************************/


/*============================================================================
     'groupNeighbor' class
============================================================================*/

groupNeighbor::groupNeighbor() { clear(); }             // --- constructor

void groupNeighbor::clear()                                   // --- clear
  { bondOrder = 0; toAtomicNbr = 0; UnsBnd2nd = 0; UB2ndAtomicNbr = 0; }

int groupNeighbor::operator<(groupNeighbor &robj)        // --- [operator<
  { return (getIdxStr() < robj.getIdxStr()); }

string groupNeighbor::getIdxStr() {                       // --- getIdxStr
  string idxs;
  idxs = format(bondOrder, 1) + format(toAtomicNbr, 2)
       + format(UnsBnd2nd, 1);
  if (UnsBnd2nd > 0) { idxs += format(UB2ndAtomicNbr, 2); }
  else { idxs += format(0, 2); }
  return idxs;
}

int groupNeighbor::restoreFrom(ifstream &ifs) {         // --- restoreFrom
  string s;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "groupNeighbor{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  bondOrder = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  toAtomicNbr = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  UnsBnd2nd = atoi(s.c_str());
  if (UnsBnd2nd > 0) {
    if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
    UB2ndAtomicNbr = atoi(s.c_str());
  }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}


/*============================================================================
     'groupAtom' class
============================================================================*/

groupAtom::groupAtom() { clear(); }                     // --- constructor

void groupAtom::clear() {                                     // --- clear
  int ib;
  atomicNumber = 0; atomAttrib = 0; numNeighbors = 0;
  for (ib = 0; ib < MAXN_BONDS; ib++) { neigh[ib].clear(); }
}

string groupAtom::getIdxStr() {                           // --- getIdxStr
  string idxs;
  idxs = format(atomicNumber, 2) + format(atomAttrib, 1)
       + format(numNeighbors, 1);
  for (int i = 0; i < numNeighbors; i++) { idxs += neigh[i].getIdxStr(); }
  return idxs;
}

void groupAtom::sortNeighbors() {                     // --- sortNeighbors
  int i, j;
  groupNeighbor p;
  for (i = 0; i < (numNeighbors - 1); i++) {
    for (j = i + 1; j < numNeighbors; j++) {
      if (neigh[j] < neigh[i])
        { p = neigh[i]; neigh[i] = neigh[j]; neigh[j] = p; }
    }
  }
}

int groupAtom::restoreFrom(ifstream &ifs) {             // --- restoreFrom
  string s;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "groupAtom{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  atomicNumber = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  atomAttrib = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numNeighbors = atoi(s.c_str());
  for (int i = 0; i < numNeighbors; i++) { neigh[i].restoreFrom(ifs); }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}

int groupAtom::isRadical()                                // --- isRadical
  { return (atomAttrib > 0); }

void groupAtom::beNonRadical()                         // --- beNonRadical
  { atomAttrib = 0; }


/*============================================================================
     'groupMap' class
============================================================================*/


// -------------------- [groupAtomCFunc::operator()(groupAtom, groupAtom)]
int groupAtomCFunc::operator()(groupAtom lobj, groupAtom robj) const
  { return (lobj.getIdxStr() < robj.getIdxStr()); }

groupMap::groupMap() { clear(); }                       // --- constructor

// ------------------------------------------------------------ [loadFrom]
int groupMap::loadFrom(string flname, libPath &libp) {
  int l;
  string s, grpName;
  groupAtom ga;

  if (!empty()) { clear(); }
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  while (getTokenFrom(ifs, s) == NORM) {
    if ((l = s.length()) < 2) { return ERRE; }
    if ((s[0] != '[') || (s[l - 1] != ']')) { return ERRE; }
    grpName = s.substr(1, l - 2);
    if (ga.restoreFrom(ifs) == ERRE) { return ERRE; }
    ga.sortNeighbors();
    insert(make_pair(ga, grpName));
  }
  ifs.close();
  cout << size() << " groups loaded from [" << flname << "]\n";
  return NORM;
}

// ------------------------------------------------------------ [findName]
string groupMap::findName(groupAtom &ga) {
  iterator p = find(ga);
  if (p != end()) { return p->second; }
  return "?";
}


/*============================================================================
     'groupCycAtom' class
============================================================================*/

groupCycAtom::groupCycAtom() { clear(); }               // --- constructor

void groupCycAtom::clear()                                    // --- clear
  { atomicNumber = 0; atomAttrib = 0; bondOrdToNext = 0; }


string groupCycAtom::getIdxStrAtom() {                // --- getIdxStrAtom
  string idxs;
  idxs = format(atomicNumber, 2) + format(atomAttrib, 1);
  return idxs;
}

string groupCycAtom::getIdxStrBond() {                // --- getIdxStrBond
  string idxs;
  idxs = format(bondOrdToNext, 1);
  return idxs;
}

int groupCycAtom::restoreFrom(ifstream &ifs) {          // --- restoreFrom
  string s;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "groupCycAtom{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  atomicNumber = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  atomAttrib = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  bondOrdToNext = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}


/*============================================================================
     'groupCyclic' class
============================================================================*/

groupCyclic::groupCyclic() { clear(); }                 // --- constructor

void groupCyclic::clear() { numAtoms = 0; atom.clear(); }     // --- clear

string groupCyclic::getIdxStr() {   // ----------------------- [getIdxStr]
  string idxs;
  idxs = format(numAtoms, 3);
  for (int i = 0; i < numAtoms; i++) {
    idxs += atom[i].getIdxStrAtom() + atom[i].getIdxStrBond();
  }
  return idxs;
}

void groupCyclic::beUnique() {   // --------------------------- [beUnique]
  int iStart, iDirec, diff, ic, idxAtom, idxPrev, MxStart = 0, MxDirec = 0;
  string s = "";
  vector<string> vs0(2, "");
  vector<vector<string> > ids(numAtoms, vs0);
  vector<groupCycAtom> mxa;
  groupCycAtom atm;

  for (iStart = 0; iStart < numAtoms; iStart++) {
    for (iDirec = 0; iDirec < 2; iDirec++) {
      diff = iDirec * 2 - 1; idxAtom = iStart;
      for (ic = 0; ic < numAtoms; ic++) {
        ids[iStart][iDirec] += atom[idxAtom].getIdxStrAtom();
        if (diff < 0) {
          idxPrev = idxAtom - 1;
          if (idxPrev < 0) { idxPrev += numAtoms; }
          ids[iStart][iDirec] += atom[idxPrev].getIdxStrBond();
        } else {
          ids[iStart][iDirec] += atom[idxAtom].getIdxStrBond();
        }
        idxAtom += diff;
        if (idxAtom < 0) { idxAtom += numAtoms; }
        else if (idxAtom >= numAtoms) { idxAtom -= numAtoms; }
      }
    }
  }
  for (iStart = 0; iStart < numAtoms; iStart++) {
    for (iDirec = 0; iDirec < 2; iDirec++) {
      if (ids[iStart][iDirec] > s) {
        MxStart = iStart; MxDirec = iDirec; s = ids[iStart][iDirec];
      }
    }
  }
  diff = MxDirec * 2 - 1; idxAtom = MxStart;
  mxa.clear();
  for (ic = 0; ic < numAtoms; ic++) {
    atm = atom[idxAtom];
    if (diff < 0) {
      idxPrev = idxAtom - 1;
      if (idxPrev < 0) { idxPrev += numAtoms; }
      atm.bondOrdToNext = atom[idxPrev].bondOrdToNext;
    }
    mxa.push_back(atm);
    idxAtom += diff;
    if (idxAtom < 0) { idxAtom += numAtoms; }
    else if (idxAtom >= numAtoms) { idxAtom -= numAtoms; }
  }
  for (ic = 0; ic < numAtoms; ic++) { atom[ic] = mxa[ic]; }
}

int groupCyclic::restoreFrom(ifstream &ifs) {   // --------- [restoreFrom]
  string s;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "groupCyclic{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numAtoms = atoi(s.c_str()); atom.resize(numAtoms);
  for (int i = 0; i < numAtoms; i++) { atom[i].restoreFrom(ifs); }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}


/*============================================================================
     'groupCycMap' class
============================================================================*/

// -------------- [groupCyclicCFunc::operator()(groupCyclic, groupCyclic)]
int groupCyclicCFunc::operator()(groupCyclic lobj, groupCyclic robj) const
  { return (lobj.getIdxStr() < robj.getIdxStr()); }

groupCycMap::groupCycMap() { clear(); }                 // --- constructor

// ------------------------------------------------------------ [loadFrom]
int groupCycMap::loadFrom(string flname, libPath &libp) {
  int l;
  string s, grpName;
  groupCyclic gc;

  if (!empty()) { clear(); }
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  while (getTokenFrom(ifs, s) == NORM) {
    if ((l = s.length()) < 2) { return ERRE; }
    if ((s[0] != '[') || (s[l - 1] != ']')) { return ERRE; }
    grpName = s.substr(1, l - 2);
    if (gc.restoreFrom(ifs) == ERRE) { return ERRE; }
    gc.beUnique();
    insert(make_pair(gc, grpName));
  }
  ifs.close();
  cout << size() << " cyclic groups loaded from [" << flname << "]\n";
  return NORM;
}

// ------------------------------------------------------------ [findName]
string groupCycMap::findName(groupCyclic &gc) {
  iterator p = find(gc);
  if (p != end()) { return p->second; }
  return "?";
}


/*============================================================================
     'groupLib' class containing all maps of groups
============================================================================*/

groupLib::groupLib() {                                  // --- constructor
  libPath libp;
  clear();
  loadFrom(FL_GRPMAI, FL_GRPCYC, FL_GRPRAD, libp);
}

void groupLib::clear()                                        // --- clear
  { main.clear(); cyc.clear(); rad.clear(); }

// ------------------------------------------------------------ [loadFrom]
int groupLib::loadFrom(string flnm, string flnc, string flnr,
 libPath &libp) {
  if (main.loadFrom(flnm, libp) == ERRE)
    { cout << "Error reading main group list.\n"; return ERRE; }
  if (cyc.loadFrom(flnc, libp) == ERRE)
    { cout << "Error reading cyclic group list.\n"; return ERRE; }
  if (rad.loadFrom(flnr, libp) == ERRE)
    { cout << "Error reading radical group list.\n"; return ERRE; }
  return NORM;
}


/*============================================================================
     'molecByGroups' class
============================================================================*/


// ----------------------------------------------------- [molecByGroups()]
molecByGroups::molecByGroups() { clear(); }

// --------------------------------------------------------------- [clear]
void molecByGroups::clear()
  { cycGrpName = ""; radGrpName = ""; atmGrps.clear(); }

// --------------------------------------------------------------- [print]
void molecByGroups::print(ostream &os) {
  os << "----- molecByGroup -----\n";
  if (cycGrpName != "")
    { os << "cyclic group [" << cycGrpName << "]\n"; }
  if (radGrpName != "")
    { os << "radical group [" << radGrpName << "]\n"; }
  os << "number of atom groups : " << atmGrps.size() << endl;
  map<string, int>::iterator p;
  for (p = atmGrps.begin(); p != atmGrps.end(); p++)
    { os << " [" << p->first << "] " << p->second << endl; }
}

// --------------------------------------------------------- [operator <<]
ostream &operator<<(ostream &os, molecByGroups &mg)
  { mg.print(os); return os; }


/*****************************************************************************
     'molecule' and related classes
*****************************************************************************/


/*============================================================================
     'chemBond' class
============================================================================*/

// ---------------------------------------------------------- [chemBond()]
chemBond::chemBond() { order = 0; idxFirst = 0; idxSecond = 0; }

// --------------------------------------------- [chemBond(int, int, int)]
chemBond::chemBond(int ord, int idx1, int idx2)
  { order = ord; idxFirst = idx1; idxSecond = idx2; }

void chemBond::setFrom(int ord, int idx1, int idx2) {   // ----- [setFrom]
  order = ord; idxFirst = idx1; idxSecond = idx2;
}

int chemBond::operator==(chemBond &robj) {   // ------------ [operator ==]
  if (order != robj.order) { return false; }
  if (idxFirst != robj.idxFirst) { return false; }
  if (idxSecond != robj.idxSecond) { return false; }
  return true;
}


/*============================================================================
     'listBonds' class
============================================================================*/


// --- add a bond to the list (return ERRE for bond already exist) -------
int listBonds::addBond(chemBond bnd) {                       // [addBond]
  if (findBond(bnd) >= 0) { return ERRE; }
  push_back(bnd);
  return NORM;
}

// --- find identical bond in list and return index (-1 for none) --------
int listBonds::findBond(chemBond &bnd) {                     // [findBond]
  int nBonds, ib;
  nBonds = size();
  for (ib = 0; ib < nBonds; ib++) {
    if (bnd == (*this)[ib]) { return ib; }
  }
  return -1;
}


/*============================================================================
     'bondInfo' class
============================================================================*/

bondInfo::bondInfo() { clear(); }                       // --- constructor

void bondInfo::clear()                                        // --- clear
  { order = 0; toAtomicNbr = 0; toAtomIndex = 0; isCommutable = false; }

int bondInfo::operator>(bondInfo &robj) {
  int idxLobj = getIdx(), idxRobj = robj.getIdx();
  return (idxLobj > idxRobj);
}

int bondInfo::operator<(bondInfo &robj) {
  int idxLobj = getIdx(), idxRobj = robj.getIdx();
  return (idxLobj < idxRobj);
}

int bondInfo::getIdx() {
  int idx;
  idx = (toAtomicNbr * 3 + (order - 1)) * MAXN_ATOMS + toAtomIndex;
  return idx;
}

int bondInfo::storeTo(ofstream &ofs) {
  ofs << "      bondInfo{ " << order << " " << toAtomicNbr << " "
      << toAtomIndex << " }\n";
  return NORM;
}

int bondInfo::restoreFrom(ifstream &ifs) {
  string s;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "bondInfo{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  order = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  toAtomicNbr = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  toAtomIndex = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}

ostream &bondInfo::print(ostream &os) {                       // --- print
  os << " ord=" << order << " toAn=" << toAtomicNbr << " toAi="
     << toAtomIndex << " isC=" << isCommutable <<  endl;
  return os;
}

ostream &operator<<(ostream &os, bondInfo &bi) { return bi.print(os); }

/*============================================================================
     'atomInfo' class
============================================================================*/

atomInfo::atomInfo() { clear(); }                       // --- constructor

void atomInfo::clear() {                                      // --- clear
  int i;
  atomicNumber = 0; atomAttrib = 0; numBonds = 0;
  for (i = 0; i < MAXN_BONDS; i++) { bond[i].clear(); }
}

int atomInfo::operator>(atomInfo &robj) {
  string idxLobj = getIdxStr(), idxRobj = robj.getIdxStr();
  return (idxLobj > idxRobj);
}

int atomInfo::operator<(atomInfo &robj) {
  string idxLobj = getIdxStr(), idxRobj = robj.getIdxStr();
  return (idxLobj < idxRobj);
}

string atomInfo::getIdxStr() {
  string idxs;
  int i;
  sortBonds();
  idxs = format(numBonds, 1) + format(atomicNumber, 2)
    + format(atomAttrib, 1);
  for (i = 0; i < MAXN_BONDS; i++) {
    if (i < numBonds) { idxs += format(bond[i].toAtomicNbr, 2); }
    else { idxs += format(0, 2); }
  }
  for (i = 0; i < MAXN_BONDS; i++) {
    if (i < numBonds) { idxs += format(bond[i].order, 1); }
    else { idxs += format(0, 1); }
  }
  for (i = 0; i < MAXN_BONDS; i++) {
    if (i < numBonds) { idxs += format(bond[i].toAtomIndex, 3); }
    else { idxs += format(0, 3); }
  }
  return idxs;
}

void atomInfo::sortBonds() {
  int i, j;
  bondInfo p;
  for (i = 0; i < numBonds - 1; i++) {
    for (j = i + 1; j < numBonds; j++) {
      if (bond[j] < bond[i])
        { p = bond[i]; bond[i]= bond[j]; bond[j] = p; }
    }
  }
}

int atomInfo::storeTo(ofstream &ofs) {
  ofs << "    atomInfo{ " << atomicNumber << " " << atomAttrib << " "
      << numBonds << endl;
  for (int i = 0; i < numBonds; i++) {
    if (bond[i].storeTo(ofs) == ERRE) { return ERRE; }
  }
  ofs << "    }\n";
  return NORM;
}

int atomInfo::restoreFrom(ifstream &ifs) {
  string s;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "atomInfo{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  atomicNumber = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  atomAttrib = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numBonds = atoi(s.c_str());
  for (int i = 0; i < numBonds; i++) {
    if (bond[i].restoreFrom(ifs) == ERRE) { return ERRE; }
  }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}

// ------------------------------------------------------- [getBondToIdcs]
vector<int> atomInfo::getBondToIdcs() {
  int ib;
  vector<int> lstTo;
  lstTo.clear();
  for (ib = 0; ib < numBonds; ib++)
    { lstTo.push_back(bond[ib].toAtomIndex); }
  return lstTo;
}

// -------------------------------------------------------------- [bondTo]
bondInfo *atomInfo::bondTo(int idxAtom) {
  int idxBond;
  if ((idxBond = idxBondTo(idxAtom)) < 0) { return 0; }
  return &bond[idxBond];
}

// ----------------------------------------------------------- [idxBondTo]
int atomInfo::idxBondTo(int idxAtom) {
  int ib;
  for (ib = 0; ib < numBonds; ib++) {
    if (bond[ib].toAtomIndex == idxAtom) { return ib; }
  }
  return (-1);
}

// -------------------------------------------------------- [removeBondTo]
void atomInfo::removeBondTo(int idxAtom) {
  int idxBond;
  idxBond = idxBondTo(idxAtom);
  if (idxBond < 0) { return; }
  removeBond(idxBond);
}

// ---------------------------------------------------------- [removeBond]
void atomInfo::removeBond(int idxBond) {
  int ib, ordRB;
  if ((idxBond < 0) || (idxBond >= numBonds)) { return; }
  ordRB = bond[idxBond].order;
  for (ib = idxBond; ib < (numBonds - 1); ib ++)
    { bond[ib] = bond[ib + 1]; }
  numBonds--;
  atomAttrib += ordRB;
}

// -------------------------------------------------------- [getNumHatoms]

int atomInfo::getNumHatoms() {
  int ib, numH, vlnc = elmt[atomicNumber].valence, sumBO = atomAttrib;
  for (ib = 0; ib < numBonds; ib++) { sumBO += bond[ib].order; }
  numH = vlnc - sumBO;
  return numH;
}

ostream &atomInfo::print(ostream &os) {                       // --- print
  int ib;
  os << " An=" << atomicNumber << " Aa=" << atomAttrib
     << " Nb=" << numBonds << endl;
  for (ib = 0; ib < numBonds; ib++)
    { os << "  bond[" << ib << "] " << bond[ib]; }
  return os;
}

ostream &operator<<(ostream &os, atomInfo &ai) { return ai.print(os); }

/*============================================================================
     'molecInfo' class
============================================================================*/

molecInfo::molecInfo() { clear(); }                     // --- constructor

// --------------------------------------------------------------- [clear]
void molecInfo::clear() { numAtoms = 0; atom.clear(); }

// ---------------------------------------------- [operators ==, !=, <, >]
int molecInfo::operator==(molecInfo &rm)
  { if (compareWith(rm) == 0) { return true; } else { return false; } }
int molecInfo::operator!=(molecInfo &rm)
  { if (compareWith(rm) != 0) { return true; } else { return false; } }
int molecInfo::operator<(molecInfo &rm)
  { if (compareWith(rm) < 0) { return true; } else { return false; } }
int molecInfo::operator>(molecInfo &rm)
  { if (compareWith(rm) > 0) { return true; } else { return false; } }

// ------------------------------------------------------------ [beUnique]
void molecInfo::beUnique(molecByChains &unCh, int &isCyclic, int &nEqCh,
 int &isAroma) {
  int nTrmAtoms, idxa, numCases, i, j;
  vector<molecByChains> molc;
  molecByChains mc;
  string aromaChk;
  vector<int> sortTbl, trnsTbl;

  isCyclic = false;
  nTrmAtoms = getTerminal();
    // check whether cylic or not
  if (followChain(0, 0, mc) == ERRE) { isCyclic = true; }
  if (isCyclic) {
    atomChain refc = mc.chain[mc.numChains - 1];
    listCyclics(numCases, refc, molc);
  } else {
    numCases = nTrmAtoms;
    molc.clear(); molc.push_back(mc);
    for (i = 1; i < nTrmAtoms; i++) {
      mc.clear();
      followChain(0, i, mc);
      molc.push_back(mc);
    }
  }
  for (i = 0; i < numCases; i++) {
    if (isCyclic) { molc[i].sortTree(); }
    else { molc[i].beUnique(); }
  }
  for (i = 0; i < (numCases - 1); i++) {
    for (j = i + 1; j < numCases; j++) {
      if (molc[j] > molc[i])
        { mc = molc[i]; molc[i] = molc[j]; molc[j] = mc; }
    }
  }
  nEqCh = 1;
  for (i = 1; i < numCases; i++) {
    if (molc[i] != molc[0]) { break; }
    nEqCh++;
  }
  unCh = molc[0];
  unCh.getRenumTbl(numAtoms, sortTbl, trnsTbl);
  unCh.renumAtoms(trnsTbl);
  renumAtoms(sortTbl, trnsTbl);
  isAroma = false;
  if (isCyclic) { isAroma = unCh.chain[0].isAromatic(); }
}

// ---------------------------------------------------------- [renumAtoms]
void molecInfo::renumAtoms(vector<int> &sortTbl, vector<int> &trnsTbl) {
  molecInfo newmi;
  int oldIdx, newIdx, ia, ib;
  newmi.numAtoms = numAtoms; newmi.atom.resize(numAtoms);
  for (newIdx = 0; newIdx < newmi.numAtoms; newIdx++) {
    oldIdx = sortTbl[newIdx];
    newmi.atom[newIdx] = atom[oldIdx];
  }
  for (ia = 0; ia < newmi.numAtoms; ia++) {
    for (ib = 0; ib < newmi.atom[ia].numBonds; ib++) {
      oldIdx = newmi.atom[ia].bond[ib].toAtomIndex;
      newIdx = trnsTbl[oldIdx];
      newmi.atom[ia].bond[ib].toAtomIndex = newIdx;
    }
  }
  *this = newmi;
}

// ---------------------------------------------------------- [checkParam]
//   check parameters

int molecInfo::checkParam() {
  int an, numRadCent, nBonds, sumBO, sumV, nBondsTo, nCbonds;
  int ia, ib, iaNxt, ibn, jb;

  numRadCent = 0;
  for (ia = 0; ia < numAtoms; ia++) {
    an = atom[ia].atomicNumber;
    if (atom[ia].atomAttrib > 0) { numRadCent++; }
    nBonds = atom[ia].numBonds; sumBO = 0;
    for (ib = 0; ib < nBonds; ib++) { sumBO += atom[ia].bond[ib].order; }
    sumV = sumBO + atom[ia].atomAttrib;
    if (sumV > (elmt[an].valence + elmt[an].hyperValence)) {
      cout << "ERROR molecInfo::checkParam [invalid bonds]\n";
      return ERRE;
    }
    for (ib = 0; ib < nBonds; ib++) {
      iaNxt = atom[ia].bond[ib].toAtomIndex;
      if (atom[iaNxt].atomicNumber != atom[ia].bond[ib].toAtomicNbr) {
        cout << "ERROR molecInfo::checkParam [bond mismatch]\n";
        return ERRE;
      }
      nBondsTo = atom[iaNxt].numBonds;
      nCbonds = 0;
      for (ibn = 0; ibn < nBondsTo; ibn++) {
        if (atom[iaNxt].bond[ibn].toAtomIndex == ia) {
          nCbonds++;
          if (atom[iaNxt].bond[ibn].order != atom[ia].bond[ib].order) {
            cout << "ERROR molecInfo::checkParam [bond mismatch]\n";
            return ERRE;
          }
          if (atom[iaNxt].bond[ibn].toAtomicNbr != atom[ia].atomicNumber) {
            cout << "ERROR molecInfo::checkParam [bond mismatch]\n";
            return ERRE;
          }
        }
      }
      if (nCbonds != 1) {
        cout << "ERROR molecInfo::checkParam [bond mismatch]\n";
        return ERRE;
      }
      for (jb = ib + 1; jb < nBonds; jb++) {
        if (atom[ia].bond[jb].toAtomIndex == atom[ia].bond[ib].toAtomIndex) {
          cout << "ERROR molecInfo::checkParam [bond duplicate]\n";
          return ERRE;
        }
      }
    }
  }
  if (numRadCent > 1) {
    cout << "ERROR molecInfo::checkParam [too many rad. centers]\n";
    return ERRE;
  }
  return NORM;
}

// --------------------------------------------------------- [getAtomComp]
//   get atomic composition

void molecInfo::getAtomComp(atmcComp &ac) {
  int ia, ib, atmcNbr, numH;
  if (!ac.empty()) { ac.clear(); }
  for (ia = 0; ia < numAtoms; ia++) {
    atmcNbr = atom[ia].atomicNumber; numH = atom[ia].getNumHatoms();
    ac[atmcNbr]++;
    if (numH > 0) { ac[1] += numH; }
  }
}

// --------------------------------------------------- [getAtomCompSorted]
//   get atomic composition sorted for chemical formula

void molecInfo::getAtomCompSorted(atmcCompSorted &acs) {
  atmcComp ac;
  getAtomComp(ac); acs.setFrom(ac);
}

// ---------------------------------------------------- [atmCompToFormula]
//   convert atomic composition to chemical formula
string molecInfo::atmCompToFormula(atmcComp &ac) {
  string rets = "";
  atmcCompSorted acs;
  atmcCompSorted::iterator p;
  acs.setFrom(ac);
  for (p = acs.begin(); p != acs.end(); p++) {
    if (p->second.second > 0) {
      rets += elmt[p->second.first].atomicSymbol;
      if (p->second.second > 1) { rets += format(p->second.second); }
    }
  }
  return rets;
}

// ------------------------------------------------------ [getChemFormula]
//   get chemical formula

string molecInfo::getChemFormula() {
  string rets = "";
  atmcCompSorted acs;
  atmcCompSorted::iterator p;
  getAtomCompSorted(acs);
  for (p = acs.begin(); p != acs.end(); p++) {
    rets += elmt[p->second.first].atomicSymbol;
    if (p->second.second > 1) { rets += format(p->second.second); }
  }
  return rets;
}

// --------------------------------------------------------- [getNumAtoms]
//   get number of atoms of specified atomic number

int molecInfo::getNumAtoms(int atmcNbr) {
  int numAtms;
  atmcComp ac;
  getAtomComp(ac);
  if (ac.find(atmcNbr) == ac.end()) { numAtoms = 0; }
  else { numAtoms = ac[atmcNbr]; }
  return numAtoms;
}

// --------------------------------------------------------------- [print]
void molecInfo::print(ostream *os) {
  *os << "numAtoms = " << numAtoms << "\n";
  for (int i = 0; i < numAtoms; i++) {
    *os << "atom #" << i << " : " << atom[i].atomicNumber
        << ", " << atom[i].atomAttrib
        << ", " << atom[i].numBonds << " (";
    for (int j = 0; j < atom[i].numBonds; j++) {
      *os << atom[i].bond[j].order
          << "/" << atom[i].bond[j].toAtomicNbr
          << "/" << atom[i].bond[j].toAtomIndex << "; ";
    }
    *os << ")\n";
  }
}

int molecInfo::storeTo(ofstream &ofs) {   // ------------------- [storeTo]
  ofs << "  molecInfo{ " << numAtoms << endl;
  for (int i = 0; i < numAtoms; i++) {
    if (atom[i].storeTo(ofs) == ERRE) { return ERRE; }
  }
  ofs << "  }\n";
  return NORM;
}

int molecInfo::restoreFrom(ifstream &ifs) {   // ----------- [restoreFrom]
  string s;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "molecInfo{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numAtoms = atoi(s.c_str());
  atom.resize(numAtoms);
  for (int i = 0; i < numAtoms; i++) {
    if (atom[i].restoreFrom(ifs) == ERRE) { return ERRE; }
  }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}

// -------------------------------------------------------- [beAromaOther]
//   change *this to another localized form of atomatics

void molecInfo::beAromaOther(atomChain &ac) {
  int nAtom, ia, iaBef, iaNxt, idx, idxBef, idxNxt;
  int nBond, ib, bOrd, toIdx, bOrdChange;
  nAtom = ac.numAtoms;
  for (ia = 0; ia < nAtom; ia++) {
    iaBef = ia - 1;
    if (iaBef < 0) { iaBef += nAtom; }
    iaNxt = ia + 1;
    if (iaNxt >= nAtom) { iaNxt -= nAtom; }
    idx = ac.atomIndex[ia];
    idxBef = ac.atomIndex[iaBef];
    idxNxt = ac.atomIndex[iaNxt];
    nBond = atom[idx].numBonds;
    for (ib = 0; ib < nBond; ib++) {
      bOrd = atom[idx].bond[ib].order;
      bOrdChange = bOrd + 1;
      if (bOrdChange > 2) { bOrdChange -= 2; }
      toIdx = atom[idx].bond[ib].toAtomIndex;
      if ((toIdx == idxBef) || (toIdx == idxNxt))
        { atom[idx].bond[ib].order = bOrdChange; }
    }
  }
}

// --------------------------------------------------------- [getResonStr]
//   get resonance structures (returns number found)

int molecInfo::getResonStr(vector<molecInfo> &rs) {
  int numReson, nLast, nLastPrev, idxRC;
  vector<int> idxRadCent;
  int foundFurthRes, nBonds, idxTo, nBndsNxt, idxToNxt, boToNxt;
  int isNew;
  int il, ib, ibn, ip, jb;
  molecInfo r, q;

  idxRadCent.clear();
  rs.clear(); rs.push_back(*this); numReson = 1;
  if (getNumRadCent() != 1) { return numReson; }
  idxRadCent.push_back(getIdxRadCent());
  nLast = 1; nLastPrev = 0;
  do {
    foundFurthRes = false;
    for (il = nLastPrev; il < nLast; il++) {
      idxRC = idxRadCent[il]; r = rs[il];
      nBonds = r.atom[idxRC].numBonds;
      for (ib = 0; ib < nBonds; ib++) {
        idxTo = r.atom[idxRC].bond[ib].toAtomIndex;
        nBndsNxt = r.atom[idxTo].numBonds;
        for (ibn = 0; ibn < nBndsNxt; ibn++) {
          idxToNxt = r.atom[idxTo].bond[ibn].toAtomIndex;
          boToNxt = r.atom[idxTo].bond[ibn].order;
          if (boToNxt > 1) {
            q = r;
            q.atom[idxToNxt].atomAttrib++;
            q.atom[idxRC].atomAttrib--;
            for (jb = 0; jb < q.atom[idxToNxt].numBonds; jb++) {
              if (q.atom[idxToNxt].bond[jb].toAtomIndex == idxTo)
                { q.atom[idxToNxt].bond[jb].order--; }
            }
            q.atom[idxTo].bond[ibn].order--;
            for (jb = 0; jb < nBndsNxt; jb++) {
              if (q.atom[idxTo].bond[jb].toAtomIndex == idxRC)
                { q.atom[idxTo].bond[jb].order++; }
            }
            q.atom[idxRC].bond[ib].order++;
            isNew = true;
            for (ip = 0; ip < numReson; ip++) {
              if (idxToNxt == idxRadCent[ip]) {
                if (q == rs[ip]) { isNew = false; break; }
              }
            }
            if (isNew) {
              rs.push_back(q);
              idxRadCent.push_back(idxToNxt);
              numReson++; foundFurthRes = true;
            }
          }
        }
      }
    }
    nLastPrev = nLast; nLast = numReson;
  } while (foundFurthRes);
  return numReson;
}

// ---------------------------------------------------------- [beResonAvr]
//   be averaged structure

void molecInfo::beResonAvr(int numReson, vector<molecInfo> &rs) {
  int sumAttr, sumBO, nBonds;
  int ia, ir, ib;
  *this = rs[0];
  for (ia = 0; ia < numAtoms; ia++) {
    sumAttr = 0;
    for (ir = 0; ir < numReson; ir++)
      { sumAttr += rs[ir].atom[ia].atomAttrib; }
    atom[ia].atomAttrib = sumAttr;
    nBonds = atom[ia].numBonds;
    for (ib = 0; ib < nBonds; ib++) {
      sumBO = 0;
      for (ir = 0; ir < numReson; ir++)
        { sumBO += rs[ir].atom[ia].bond[ib].order; }
      atom[ia].bond[ib].order = sumBO;
    }
  }
}

// ------------------------------------------------------- [getNumRadCent]
//   get number of radical centers

int molecInfo::getNumRadCent() {
  int nRC = 0;
  for (int ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomAttrib > 0) { nRC++; }
  }
  return nRC;
}

// ------------------------------------------------------- [getIdxRadCent]
//   get index of the radical center (found first)

int molecInfo::getIdxRadCent() {
  for (int ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomAttrib > 0) { return ia; }
  }
  return -1;
}

// ----------------------------------------------------- [getIdcsRadCents]
//   get indices of the all radical centers

void molecInfo::getIdcsRadCents(vector<int> &irc) {
  irc.clear();
  for (int ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomAttrib > 0) { irc.push_back(ia); }
  }
}

// ----------------------------------------------------------- [getStrRSE]
//   get a string representation of R.S.E.

string molecInfo::getStrRSE() {
  /* -------------------------------------------
   representation string format: "VBDTS"
    V ... valence of rad. cent.
    B ... 3 - (max BO connetcted to r.c.)
    D ... number of 'next to' double bonds
    T ... number of 'next to' triple bonds
    S ... number of 'next to' single bonds
  --------------------------------------------*/
  int idxRC, valence, nBonds, maxBO, nntDB, nntTB, nntSB, idxTo;
  int nBndsNxt;
  int ib, bo, ibn, bon;

  string rets = "00000";
  if (getNumRadCent() != 1) { return rets; }
  idxRC = getIdxRadCent();
  valence = elmt[atom[idxRC].atomicNumber].valence;
  nBonds = atom[idxRC].numBonds;
  maxBO = 0; nntDB = 0; nntTB = 0; nntSB = 0;
  for (ib = 0; ib < nBonds; ib++) {
    bo = atom[idxRC].bond[ib].order;
    if (bo > maxBO) { maxBO = bo; }
    idxTo = atom[idxRC].bond[ib].toAtomIndex;
    nBndsNxt = atom[idxTo].numBonds;
    for (ibn = 0; ibn < nBndsNxt; ibn++) {
      bon = atom[idxTo].bond[ibn].order;
      switch (bon) {
       case 1: nntSB++; break;
       case 2: nntDB++; break;
       case 3: nntTB++; break;
      }
    }
  }
  rets = format(valence, 1) + format(3 - maxBO, 1) + format(nntDB, 1)
       + format(nntTB, 1) + format(nntSB, 1);
  return rets;
}

// ------------------------------------------------------- [getGroupAroma]
void molecInfo::getGroupAroma(int idxAtom, atomChain &arc, groupAtom &ga) {
  int numB, bOrd, idNext, nbNext, boNext, isAromAtom, numNeigh = 0;
  int isNextArom;
  if ((idxAtom < 0) || (idxAtom >= numAtoms)) { return; }
  ga.atomicNumber = atom[idxAtom].atomicNumber;
  ga.atomAttrib = atom[idxAtom].atomAttrib;
  numB = atom[idxAtom].numBonds;
  isAromAtom = arc.isIdxInChain(idxAtom);
  for (int ib = 0; ib < numB; ib++) {
    idNext = atom[idxAtom].bond[ib].toAtomIndex;
    isNextArom = arc.isIdxInChain(idNext);
    if ((!isAromAtom) || (isAromAtom && !isNextArom)) {
      ga.neigh[numNeigh].bondOrder = (bOrd = atom[idxAtom].bond[ib].order);
      ga.neigh[numNeigh].toAtomicNbr = atom[idxAtom].bond[ib].toAtomicNbr;
      ga.neigh[numNeigh].UnsBnd2nd = 0;
      if (isNextArom) {
        ga.neigh[numNeigh].UnsBnd2nd = 6;
        ga.neigh[numNeigh].UB2ndAtomicNbr = 6;
      } else if (bOrd == 1) {
        nbNext = atom[idNext].numBonds;
        for (int ibn = 0; ibn < nbNext; ibn++) {
          if ((boNext = atom[idNext].bond[ibn].order) > 1) {
            ga.neigh[numNeigh].UnsBnd2nd += boNext;
            ga.neigh[numNeigh].UB2ndAtomicNbr
             = atom[idNext].bond[ibn].toAtomicNbr;
          }
        }
      }
      numNeigh++;
    }
  }
  if (isAromAtom) {
    switch (atom[idxAtom].atomicNumber) {
    case 7:
      ga.neigh[numNeigh].bondOrder = 1;
      ga.neigh[numNeigh].toAtomicNbr = 6;
      ga.neigh[numNeigh].UnsBnd2nd = 2;
      ga.neigh[numNeigh].UB2ndAtomicNbr = 6;
      numNeigh++;
      ga.neigh[numNeigh].bondOrder = 2;
      ga.neigh[numNeigh].toAtomicNbr = 6;
      ga.neigh[numNeigh].UnsBnd2nd = 0;
      numNeigh++;
      break;
    default:
      ga.neigh[numNeigh].bondOrder = 6;
      ga.neigh[numNeigh].toAtomicNbr = 6;
      ga.neigh[numNeigh].UnsBnd2nd = 0;
      numNeigh++;
      break;
    }
  }
  ga.numNeighbors = numNeigh;
  ga.sortNeighbors();
}

// ---------------------------------------------------- [getGroupNonAroma]
void molecInfo::getGroupNonAroma(int idxAtom, groupAtom &ga) {
  int numB, bOrd, idNext, nbNext, boNext;
  if ((idxAtom < 0) || (idxAtom >= numAtoms)) { return; }
  ga.atomicNumber = atom[idxAtom].atomicNumber;
  ga.atomAttrib = atom[idxAtom].atomAttrib;
  ga.numNeighbors = (numB = atom[idxAtom].numBonds);
  for (int ib = 0; ib < numB; ib++) {
    ga.neigh[ib].bondOrder = (bOrd = atom[idxAtom].bond[ib].order);
    ga.neigh[ib].toAtomicNbr = atom[idxAtom].bond[ib].toAtomicNbr;
    ga.neigh[ib].UnsBnd2nd = 0;
    if (bOrd == 1) {
      idNext = atom[idxAtom].bond[ib].toAtomIndex;
      nbNext = atom[idNext].numBonds;
      for (int ibn = 0; ibn < nbNext; ibn++) {
        if ((boNext = atom[idNext].bond[ibn].order) > 1) {
          ga.neigh[ib].UnsBnd2nd += boNext;
          ga.neigh[ib].UB2ndAtomicNbr
           = atom[idNext].bond[ibn].toAtomicNbr;
        }
      }
    }
  }
  ga.sortNeighbors();
}

// ------------------------------------------------------ [getDissocGroup]
//   get group formed when the bond breaks

void molecInfo::getDissocGroup(int idx, int cIdx, groupAtom &ga) {
  int numB, ib, bo, toIdx, ing, numBNxt, ibn, boNxt;
  ga.atomicNumber = atom[idx].atomicNumber;
  ga.atomAttrib = atom[idx].atomAttrib;
  ga.numNeighbors = (numB = atom[idx].numBonds);
  for (ib = 0, ing = 0; ib < numB; ib++) {
    bo = atom[idx].bond[ib].order; toIdx = atom[idx].bond[ib].toAtomIndex;
    if (toIdx == cIdx) {
      ga.atomAttrib += bo; ga.numNeighbors--;
    } else {
      ga.neigh[ing].bondOrder = bo;
      ga.neigh[ing].toAtomicNbr = atom[idx].bond[ib].toAtomicNbr;
      ga.neigh[ing].UnsBnd2nd = 0;
      if (bo == 1) {
        numBNxt = atom[toIdx].numBonds;
        for (ibn = 0; ibn < numBNxt; ibn++) {
          if ((boNxt = atom[toIdx].bond[ibn].order) > 1) {
            ga.neigh[ing].UnsBnd2nd += boNxt;
            ga.neigh[ing].UB2ndAtomicNbr
             = atom[toIdx].bond[ibn].toAtomicNbr;
          }
        }
      }
      ing++;
    }
  }
  ga.sortNeighbors();
}

// -------------------------------------------------------- [isRingPlanar]
//   check whether the ring is planar or not

int molecInfo::isRingPlanar(atomChain &ring) {
  static string pattPlanar = "111111";
  int nRingAtoms, numPatt, ia;
  string lstPlanarCnt;
  vector<string> ringPatt;

  if (!ring.isCyclic) {
    cout << "ERROR molecInfo::isRingPlanar [no ring]\n";
    return false;
  }
  nRingAtoms = ring.numAtoms;
  if (nRingAtoms > 6) { return false; }
  lstPlanarCnt = getListPlanarCent(ring);
  numPatt = getPattMatchPlanCnt(lstPlanarCnt, &ringPatt);
  if (numPatt == 0) { return false; }
  if (ringPatt[0] == pattPlanar.substr(0, nRingAtoms)) { return true; }
  else { return false; }
}

// ---------------------------------------------------- [getSymIsomCyclic]
//   get informations on rotational symmetry number, number of isomers,
//    and number of internal rotors of a cyclic molecule

int molecInfo::getSymIsomCyclic(atomChain &ring, int &numIsom,
 int &numIntRot) {
  /*-------------------------------------------------
    string representation format: NNAPS0S1B
                                  012345678
     NN .. atomic number, A .. atom attribute,
     P .. pattern (2 .. up / 1 .. middle / 0 .. down)
     S0, S1 .. substitution index, B .. bond order
     substitution index:
       00 .. no subs. (implies inplane subs.)
       01 .. hydrogen,  02 .. lone pare of N atom
       03 and higher .. other substitutions
  -------------------------------------------------*/
  int nRingAtoms, isAroma, numPatts, idxAtom, idxCase, nCases, nCurCases;
  int matchPrev, numSym, gcmSymIsom, iRotSym;
  int idxSubTo, nBonds, bOrd, idx1, idx2, farLin, bSym;
  int ip, ia, js, ic, ks, ib;
  vector<int> ringBO, sym;
  vector<vector<int> > subPat, idxStAt;
  string planCnt, pats, rrep, s;
  vector<string> pattStr, caseStr;
  vector<string> rotR;
  vector<vector<string> > rotRepr;
  listBonds lstBonds;

  if (!ring.isCyclic) {
    cout << "ERROR molecInfo::getSymIsomCyclic [not cyclic]\n";
    return 0;
  }
  nRingAtoms = ring.numAtoms; isAroma = ring.isAromatic();
  planCnt = getListPlanarCent(ring);
  numPatts = getPattMatchPlanCnt(planCnt, &pattStr);
  getRingSubList(ring, subPat, ringBO, idxStAt);
  rotRepr.clear();

  // --- make up base subst. case for all patts. in caseStr and rotRepr
  if (!caseStr.empty()) { caseStr.clear(); }
  for (ip = 0; ip < numPatts; ip++) {
    pats = pattStr[ip];
    for (ia = 0, rrep = ""; ia < nRingAtoms; ia++) {
      idxAtom = ring.atomIndex[ia];
      rrep += format(atom[idxAtom].atomicNumber, 2);
      rrep += format(atom[idxAtom].atomAttrib, 1);
      rrep += pats[ia];
      for (js = 0; js < 2; js++) { rrep += format(subPat[ia][js], 2); }
      if (isAroma) { rrep += 'A'; }
      else { rrep += format(ringBO[ia], 1); }
    }
    caseStr.push_back(rrep); idxCase = caseStr.size() - 1;
    rotR.clear(); getRotatedRepr(nRingAtoms, caseStr[idxCase], &rotR);
    rotRepr.push_back(rotR);
  }

  // --- generate all substitution cases from base cases
  for (ia = 0; ia < nRingAtoms; ia++) {
    if ((subPat[ia][1] > 0) && (subPat[ia][1] != subPat[ia][0])) {
      nCases = caseStr.size();
      for (ic = 0; ic < nCases; ic++) {
        rrep = caseStr[ic]; s = "";
        for (js = 1; js >=0; js--) { s += format(subPat[ia][js], 2); }
        rrep.replace(ia * 9 + 4, 4, s);
        nCurCases = caseStr.size(); matchPrev = false;
        for (idxCase = 0; idxCase < nCurCases; idxCase++) {
          if (numMatchStrList(rrep, &rotRepr[idxCase]) > 0)
            { matchPrev = true; break; }
        }
        if (!matchPrev) {
          caseStr.push_back(rrep);
          idxCase = caseStr.size() - 1;
          rotR.clear(); getRotatedRepr(nRingAtoms, caseStr[idxCase], &rotR);
          rotRepr.push_back(rotR);
        }
      }
    }
  }

  // --- count symmetry numbers and number of isomers
  nCases = caseStr.size(); numSym = 1; sym.clear();
  for (ic = 0; ic < nCases; ic++) {
    sym.push_back(numMatchStrList(caseStr[ic], &rotRepr[ic]));
    numSym = getLCM(numSym, sym[ic]);
  }
  numIsom = 0;
  for (ic = 0; ic < nCases; ic++) {
    numIsom += (numSym / sym[ic]);
  }
  gcmSymIsom = getGCM(numIsom, numSym);
  numIsom /= gcmSymIsom; numSym /= gcmSymIsom;

  // --- cout symmetry numbers of subchain internal rotors
  iRotSym = 1; numIntRot = 0;
  for (ia = 0; ia < nRingAtoms; ia++) {
    idxAtom = ring.atomIndex[ia];
    for (ks = 0; ks < 2; ks++) {
      idxSubTo = idxStAt[ia][ks];
      if (idxSubTo >= 0) {
        lstBonds.clear();
        followBondsSub(idxAtom, idxSubTo, lstBonds);
        nBonds = lstBonds.size();
        for (ib = 0; ib < nBonds; ib++) {
          bOrd = lstBonds[ib].order; idx1 = lstBonds[ib].idxFirst;
          idx2 = lstBonds[ib].idxSecond;
          if (bOrd == 1) {
            farLin = 0;
            bSym = getIntRotSym(idx1, idx2, farLin);
            if ((farLin == 0) && (bSym > 0))
              { iRotSym *= bSym; numIntRot++; }
          }
        }
      }
    }
  }
  numSym *= iRotSym;
  return numSym;
}

// ---------------------------------------------------- [getSymIsomNonCyc]
//   get informations on rotational symmetry number, number of isomers,
//    and number of internal rotors of a non-cyclic molecule

int molecInfo::getSymIsomNonCyc(atomChain &mc, int neqc, int &numIsom,
 int &numIntRot) {
  int numSym, intRotSym;
  if (isTdSymmetric(mc, neqc, numIntRot, intRotSym))
    { numSym = 12 * intRotSym; numIsom = 1; return numSym; }
  if (isC3D3Symmetric(mc, neqc, numIntRot, intRotSym))
    { numSym = 3 * intRotSym; numIsom = 1; return numSym; }
  if (isC2D2Symmetric(mc, neqc, numIntRot, intRotSym))
    { numSym = 2 * intRotSym; numIsom = 1; return numSym; }
  if (isDBC2Symmetric(numIntRot, intRotSym))
    { numSym = 2 * intRotSym; numIsom = 1; return numSym; }
  if (isChiral()) { numIsom = 2; } else { numIsom = 1; }
  numIntRot = countAllIntRotSym(intRotSym);
  numSym = intRotSym;
  return numSym;
}

// --------------------------------------------------------- [getDistance]
//   get shortest distance between two atoms

int molecInfo::getDistance(int id1, int id2) {
  set<int> dstSet, passed;
  dstSet.clear(); passed.clear();
  traceDistance(-1, id1, id2, 0, dstSet, passed);
  if (dstSet.size() == 0) { return -1; }
  return *(dstSet.begin());
}

// --------------------------------------------------------- [getDstRoute]
//   get shortest distance and route between two atoms

int molecInfo::getDstRoute(int id1, int id2, vector<int> &rt) {
  vector<int> wrt;
  map<int, vector<int> > dstMap;
  set<int> passed;
  wrt.clear(); dstMap.clear(); passed.clear();
  traceDstRoute(-1, id1, id2, 0, dstMap, wrt, passed);
  if (dstMap.size() == 0) { rt = wrt; return -1; }
  rt = dstMap.begin()->second;
  return dstMap.begin()->first;
}

// --------------------------------------------------------- [isPeroxyRad]
int molecInfo::isPeroxyRad() {
  int idRC;
  if (getNumRadCent() != 1) { return false; }
  idRC = getIdxRadCent();
  if (atom[idRC].atomicNumber != 8) { return false; }
  if (atom[idRC].atomAttrib != 1) { return false; }
  if (atom[idRC].numBonds != 1) { return false; }
  if (atom[idRC].bond[0].order != 1) { return false; }
  if (atom[idRC].bond[0].toAtomicNbr != 8) { return false; }
  return true;
}

// ------------------------------------------------ [getIdxAtomMatch(int)]
int molecInfo::getIdxAtomMatch(int atmcNbr) {
  int ia;
  for (ia = 0; ia < numAtoms; ia++)
    { if (atom[ia].atomicNumber == atmcNbr) { return ia; } }
  return (-1);
}

// ------------------------------------------- [getIdxAtomMatch(int, int)]
int molecInfo::getIdxAtomMatch(int atmcNbr, int atmAttr) {
  int ia;
  for (ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomicNumber == atmcNbr) {
      if (atom[ia].atomAttrib == atmAttr) { return ia; }
    }
  }
  return (-1);
}

// ---------------------------------------- [getIdxAtomMatch(int,int,int)]
int molecInfo::getIdxAtomMatch(int atmcNbr, int toAtNbr, int bndOrd) {
  int ia, ib, idAtm = -1;
  for (ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomicNumber == atmcNbr) {
      for (ib = 0; ib < atom[ia].numBonds; ib++) {
        if ((atom[ia].bond[ib].toAtomicNbr == toAtNbr) &&
            (atom[ia].bond[ib].order == bndOrd)) {
          idAtm = ia; break;
        }
      }
      if (idAtm >= 0) { break; }
    }
  }
  return idAtm;
}

// ------------------------------------ [getIdxAtomMatch(int,int,int,int)]
int molecInfo::getIdxAtomMatch(int atmcNbr, int atmAttr, int toAtNbr,
 int bndOrd) {
  int ia, ib, idAtm = -1;
  for (ia = 0; ia < numAtoms; ia++) {
    if ( (atom[ia].atomicNumber == atmcNbr) &&
         (atom[ia].atomAttrib == atmAttr) ) {
      for (ib = 0; ib < atom[ia].numBonds; ib++) {
        if ((atom[ia].bond[ib].toAtomicNbr == toAtNbr) &&
            (atom[ia].bond[ib].order == bndOrd)) {
          idAtm = ia; break;
        }
      }
      if (idAtm >= 0) { break; }
    }
  }
  return idAtm;
}

// -------------------------------- [getIdxAtomMatch(int,int,int,int,int)]
int molecInfo::getIdxAtomMatch(int atmcNbr, int toAtNbr, int bndOrd,
 int to3AtNbr, int bnd3Ord) {
  int ia, ib, ian, ibn, idAtm = -1;
  for (ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomicNumber == atmcNbr) {
      for (ib = 0; ib < atom[ia].numBonds; ib++) {
        if ((atom[ia].bond[ib].toAtomicNbr == toAtNbr) &&
            (atom[ia].bond[ib].order == bndOrd)) {
          ian = atom[ia].bond[ib].toAtomIndex;
          for (ibn = 0; ibn < atom[ian].numBonds; ibn++) {
            if ( (atom[ian].bond[ibn].toAtomIndex != ia) &&
                 (atom[ian].bond[ibn].toAtomicNbr == to3AtNbr) &&
                 (atom[ian].bond[ibn].order == bnd3Ord) ) {
              idAtm = ia; break;
            }
            if (idAtm >= 0) { break; }
          }
        }
        if (idAtm >= 0) { break; }
      }
      if (idAtm >= 0) { break; }
    }
  }
  return idAtm;
}

// --------------------------------------------- [getIdxNextAtom(int,int)]
int molecInfo::getIdxNextAtom(int idxAtom, int atmcNbr) {
  int ib, idNxt = -1;
  for (ib = 0; ib < atom[idxAtom].numBonds; ib++) {
    if (atom[idxAtom].bond[ib].toAtomicNbr == atmcNbr) {
      idNxt = atom[idxAtom].bond[ib].toAtomIndex; break;
    }
  }
  return idNxt;
}

// ----------------------------------------- [getIdxNextAtom(int,int,int)]
int molecInfo::getIdxNextAtom(int idxAtom, int atmcNbr, int bndOrd) {
  int ib, idNxt = -1;
  for (ib = 0; ib < atom[idxAtom].numBonds; ib++) {
    if ( (atom[idxAtom].bond[ib].toAtomicNbr == atmcNbr) &&
         (atom[idxAtom].bond[ib].order == bndOrd) ) {
      idNxt = atom[idxAtom].bond[ib].toAtomIndex; break;
    }
  }
  return idNxt;
}

// ------------------------------------- [getIdxNextAtom(int,int,int,int)]
int molecInfo::getIdxNextAtom(int idxAtom, int atmcNbr, int nxBndOrd,
 int nxAtmcNbr) {
  int ia, ib, ibn, idNxt = -1;
  for (ib = 0; ib < atom[idxAtom].numBonds; ib++) {
    if (atom[idxAtom].bond[ib].toAtomicNbr == atmcNbr) {
      ia = atom[idxAtom].bond[ib].toAtomIndex;
      for (ibn = 0; ibn < atom[ia].numBonds; ibn++) {
        if ( (atom[ia].bond[ibn].order == nxBndOrd) &&
             (atom[ia].bond[ibn].toAtomicNbr == nxAtmcNbr) ) {
          idNxt = ia; break;
        }
      }
      if (idNxt > -1) { break; }
    }
  }
  return idNxt;
}

// ------------------------------------ [getIdxAtomTermMatch(int,int,int)]
//   find terminal atom only

int molecInfo::getIdxAtomTermMatch(int atmcNbr, int toAtNbr, int bndOrd) {
  int ia, ib, idAtm = -1, nBnds;
  for (ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomicNumber == atmcNbr) {
      if ((nBnds = atom[ia].numBonds) == 1) {
        for (ib = 0; ib < nBnds; ib++) {
          if ((atom[ia].bond[ib].toAtomicNbr == toAtNbr) &&
              (atom[ia].bond[ib].order == bndOrd)) {
            idAtm = ia; break;
          }
        }
        if (idAtm >= 0) { break; }
      }
    }
  }
  return idAtm;
}

// -------------------------------- [getIdxAtomTermMatch(int,int,int,int)]
//   find terminal atom only (with attrribute)

int molecInfo::getIdxAtomTermMatch(int atmcNbr, int atmAttr, int toAtNbr,
 int bndOrd) {
  int ia, ib, idAtm = -1, nBnds;
  for (ia = 0; ia < numAtoms; ia++) {
    if ( (atom[ia].atomicNumber == atmcNbr) &&
         (atom[ia].atomAttrib == atmAttr) ) {
      if ((nBnds = atom[ia].numBonds) == 1) {
        for (ib = 0; ib < nBnds; ib++) {
          if ((atom[ia].bond[ib].toAtomicNbr == toAtNbr) &&
              (atom[ia].bond[ib].order == bndOrd)) {
            idAtm = ia; break;
          }
        }
        if (idAtm >= 0) { break; }
      }
    }
  }
  return idAtm;
}

// ---------------------------- [getIdxAtomTermMatch(int,int,int,int,int)]
//   find terminal atom only (with attrribute)

int molecInfo::getIdxAtomTermMatch(int atmcNbr, int bndOrd, int toAtNbr,
 int toBdOrd, int nxAtNbr) {
  int ia, ib, ian, ibn, idAtm = -1;
  for (ia = 0; ia < numAtoms; ia++) {
    if ((atom[ia].atomicNumber == atmcNbr) && (atom[ia].numBonds == 1)) {
      if ((atom[ia].bond[0].toAtomicNbr == toAtNbr) &&
          (atom[ia].bond[0].order == bndOrd)) {
        ian = atom[ia].bond[0].toAtomIndex;
        for (ibn = 0; ibn < atom[ian].numBonds; ibn++) {
          if ((atom[ian].bond[ibn].order == toBdOrd) &&
              (atom[ian].bond[ibn].toAtomicNbr == nxAtNbr))
            { idAtm = ia; break; }
        }
        if (idAtm > -1) { break; }
      }
    }
  }
  return idAtm;
}

// ---------------------------- [getIdxAtomTermMatchNumH(int,int,int,int)]
//   find terminal atom matches with condition (with numH)

int molecInfo::getIdxAtomTermMatchNumH(int atmcNbr, int toAtNbr,
 int bndOrd, int toNumH) {
  int ia, ian, ib, idAtm = -1, nBnds;
  for (ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomicNumber == atmcNbr) {
      if ((nBnds = atom[ia].numBonds) == 1) {
        for (ib = 0; ib < nBnds; ib++) {
          if ((atom[ia].bond[ib].toAtomicNbr == toAtNbr) &&
              (atom[ia].bond[ib].order == bndOrd)) {
            ian = atom[ia].bond[ib].toAtomIndex;
            if (atom[ian].getNumHatoms() == toNumH) { idAtm = ia; break; }
          }
        }
        if (idAtm >= 0) { break; }
      }
    }
  }
  return idAtm;
}

// -------------------------------------------------------- [getAtomMatch]
atomInfo *molecInfo::getAtomMatch(int atmcNbr) {
  int idMtcAtm;
  if ((idMtcAtm = getIdxAtomMatch(atmcNbr)) < 0) { return 0; }
  return &atom[idMtcAtm];
}

                                                       // --- getBondMatch
vector<int> molecInfo::getBondMatch(int an1, int an2, int bo) {
  // Returns a pair of atom indices matching the specified bond
  //  which found first (accidentally)
  int ia, ib;
  vector<int> bnd(2, -1);
  for (ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomicNumber == an1) {
      for (ib = 0; ib < atom[ia].numBonds; ib++) {
        if ((atom[ia].bond[ib].toAtomicNbr == an2) &&
            (atom[ia].bond[ib].order == bo)) {
          bnd[0] = ia; bnd[1] = atom[ia].bond[ib].toAtomIndex; break;
        }
      }
      if (bnd[0] >= 0) { break; }
    }
  }
  return bnd;
}

                                                   // --- getAllBondsMatch
vector<vector<int> > molecInfo::getAllBondsMatch(int an1, int an2, int bo) {
  // Returns all pairs of atom indices matching the specified bond
  int ia, ib, ip, dup;
  vector<int> bnd;
  vector<vector<int> > abnds;
  abnds.clear();
  for (ia = 0; ia < numAtoms; ia++) {
    if (atom[ia].atomicNumber == an1) {
      for (ib = 0; ib < atom[ia].numBonds; ib++) {
        if ((atom[ia].bond[ib].toAtomicNbr == an2) &&
            (atom[ia].bond[ib].order == bo)) {
          bnd.clear(); bnd.push_back(ia);
          bnd.push_back(atom[ia].bond[ib].toAtomIndex);
          if (an1 == an2)
            { if (bnd[0] > bnd[1]) { swap(bnd[0], bnd[1]); } }
          dup = false;
          for (ip = 0; ip < abnds.size(); ip++)
            { if (abnds[ip] == bnd) { dup = true; break; } }
          if (!dup) { abnds.push_back(bnd); }
        }
      }
    }
  }
  return abnds;
}

int molecInfo::getAlleneCent(int &ia1, int &ia2) {    // --- getAlleneCent
  // get atom index of allene-center: ia1 & ia2: double-bonded atom-indices
  int idAC = -1, ia;
  for (ia = 0; ia < numAtoms; ia++) {
    if ((atom[ia].atomicNumber == 6) && (atom[ia].numBonds == 2)) {
      if ((atom[ia].bond[0].order == 2) && (atom[ia].bond[0].toAtomicNbr == 6)) {
        if ((atom[ia].bond[1].order == 2) && (atom[ia].bond[1].toAtomicNbr == 6)) {
          idAC = ia;
          ia1 = atom[ia].bond[0].toAtomIndex;
          ia2 = atom[ia].bond[1].toAtomIndex;
          break;
        }
      }
    }
  }
  return idAC;
}

int molecInfo::getBondOrder(int ia1, int ia2) {        // --- getBondOrder
  bondInfo *bnd;
  if ((bnd = atom[ia1].bondTo(ia2)) == 0) { return 0; }
  return bnd->order;
}

int molecInfo::countNextButOneC(int id) {          // --- countNextButOneC
  int nNBOC = 0, nb = atom[id].numBonds, ib, idnx, nbnx, ibnx;
  for (ib = 0; ib < nb; ib++) {
    if (atom[id].bond[ib].toAtomicNbr == 6) {
      idnx = atom[id].bond[ib].toAtomIndex; nbnx = atom[idnx].numBonds;
      for (ibnx = 0; ibnx < nbnx; ibnx++) {
        if (atom[idnx].bond[ibnx].toAtomIndex != id) {
          if (atom[idnx].bond[ibnx].toAtomicNbr == 6) { nNBOC++; }
        }
      }
    }
  }
  return nNBOC;
}

int molecInfo::numNeighQuatC(int id) {                // --- numNeighQuatC
  // - returns the number of neighboring quaternary carbons
  int nNQC = 0, nb = atom[id].numBonds, ib, idnx, nbnx, ibnx, isQuat;
  for (ib = 0; ib < nb; ib++) {
    if (atom[id].bond[ib].toAtomicNbr == 6) {
      idnx = atom[id].bond[ib].toAtomIndex; nbnx = atom[idnx].numBonds;
      if (nbnx == 4) {
        isQuat = true;
        for (ibnx = 0; ibnx < nbnx; ibnx++) {
          if (atom[idnx].bond[ibnx].toAtomicNbr != 6) { isQuat = false; }
        }
        if (isQuat) { nNQC++; }
      }
    }
  }
  return nNQC;
}

int molecInfo::isQuatC(int id) {                            // --- isQuatC
  // - returns 'true' if the atom is quaternary carbon
  int nb = atom[id].numBonds, ib;
  if (atom[id].atomicNumber != 6) { return false; }
  if (nb != 4) { return false; }
  for (ib = 0; ib < nb; ib++) {
    if (atom[id].bond[ib].toAtomicNbr != 6) { return false; }
  }
  return true;
}

int molecInfo::isQuatCincdO(int id) {                  // --- isQuatCincdO
  // - returns 'true' if the atom is quaternary carbon including O
  int nb = atom[id].numBonds, ib;
  if (atom[id].atomicNumber != 6) { return false; }
  if (nb != 4) { return false; }
  for (ib = 0; ib < nb; ib++) {
    if ((atom[id].bond[ib].toAtomicNbr != 6) &&
        (atom[id].bond[ib].toAtomicNbr != 8)) { return false; }
  }
  return true;
}

int molecInfo::isTertC(int id) {                            // --- isTertC
  // - returns 'true' if the atom is tertiary carbon
  int nb = atom[id].numBonds, ib;
  if (atom[id].atomicNumber != 6) { return false; }
  if (nb != 3) { return false; }
  for (ib = 0; ib < nb; ib++) {
    if (atom[id].bond[ib].toAtomicNbr != 6) { return false; }
  }
  return true;
}

int molecInfo::isTertCincdO(int id) {                  // --- isTertCincdO
  // - returns 'true' if the atom is tertiary carbon including O
  int nb = atom[id].numBonds, ib;
  if (atom[id].atomicNumber != 6) { return false; }
  if (nb != 3) { return false; }
  for (ib = 0; ib < nb; ib++) {
    if ((atom[id].bond[ib].toAtomicNbr != 6) &&
        (atom[id].bond[ib].toAtomicNbr != 8)) { return false; }
  }
  return true;
}

void molecInfo::makeNewBond(int ia1, int ia2) {         // --- makeNewBond
  int ic, idAtm[2], anAtm[2], nbAtm[2], idx, cidx, can, nb;
  idAtm[0] = ia1; idAtm[1] = ia2;
  for (ic = 0; ic < 2; ic++) {
    anAtm[ic] = atom[idAtm[ic]].atomicNumber;
    nbAtm[ic] = atom[idAtm[ic]].numBonds;
  }
  for (ic = 0; ic < 2; ic++) {
    idx = idAtm[ic]; cidx = idAtm[(ic + 1) % 2]; nb = nbAtm[ic];
    can = anAtm[(ic + 1) % 2];
    atom[idx].bond[nb].order = 1;
    atom[idx].bond[nb].toAtomicNbr = can;
    atom[idx].bond[nb].toAtomIndex = cidx;
    atom[idx].numBonds++; atom[idx].atomAttrib--;
  }
}

// ---------------------------------------------------------- [removeBond]
void molecInfo::removeBond(int ia1, int ia2) {
  if (   ((ia1 < 0) || (ia1 >= numAtoms))
      || ((ia2 < 0) || (ia2 >= numAtoms)) ) { return; }
  atom[ia1].removeBondTo(ia2); atom[ia2].removeBondTo(ia1);
}

// -------------------------------------------------------- [incBondOrder]
void molecInfo::incBondOrder(int ia1, int ia2) {
  if (getBondOrder(ia1, ia2) == 0) { return; }
  atom[ia1].bondTo(ia2)->order++; atom[ia1].atomAttrib--;
  atom[ia2].bondTo(ia1)->order++; atom[ia2].atomAttrib--;
}

// -------------------------------------------------------- [decBondOrder]
void molecInfo::decBondOrder(int ia1, int ia2) {
  int order;
  if ((order = getBondOrder(ia1, ia2)) == 0) { return; }
  if (order == 1) { removeBond(ia1, ia2); return; }
  atom[ia1].bondTo(ia2)->order--; atom[ia1].atomAttrib++;
  atom[ia2].bondTo(ia1)->order--; atom[ia2].atomAttrib++;
}

// --------------------------------------------------------- [removeHatom]
void molecInfo::removeHatom(int idxAtom) { atom[idxAtom].atomAttrib++; }

// ------------------------------------------------------- [traceDistance]
//   called by getDistance(). recursively calls iteself.

void molecInfo::traceDistance(int cIdx, int idx, int &fIdx, int dst,
 set<int> &dstSet, set<int> passed) {
  int ib, toIdx;
  passed.insert(idx);
  if (idx == fIdx) { dstSet.insert(dst); return; }
  for (ib = 0; ib < atom[idx].numBonds; ib++) {
    toIdx = atom[idx].bond[ib].toAtomIndex;
    if ((toIdx != cIdx) && (passed.find(toIdx) == passed.end()))
      { traceDistance(idx, toIdx, fIdx, (dst + 1), dstSet, passed); }
  }
}

// ------------------------------------------------------- [traceDstRoute]
//   called by getDstRoute(). recursively calls iteself.

void molecInfo::traceDstRoute(int cIdx, int idx, int &fIdx, int dst,
 map<int, vector<int> > &dstMap, vector<int> route, set<int> passed) {
  int ib, toIdx;
  route.push_back(idx);
  passed.insert(idx);
  if (idx == fIdx) { dstMap.insert(make_pair(dst, route)); return; }
  for (ib = 0; ib < atom[idx].numBonds; ib++) {
    toIdx = atom[idx].bond[ib].toAtomIndex;
    if ((toIdx != cIdx) && (passed.find(toIdx) == passed.end())) {
      traceDstRoute(idx, toIdx, fIdx, (dst + 1), dstMap, route, passed);
    }
  }
}

// --------------------------------------------------------- [compareWith]
int molecInfo::compareWith(molecInfo &mi) {
  int nAtomsMin, nBonds, ia, ib;

  if (mi.numAtoms < numAtoms) { nAtomsMin = mi.numAtoms; }
  else { nAtomsMin = numAtoms; }
  for (ia = 0; ia < nAtomsMin; ia++) {
    if (atom[ia].atomicNumber < mi.atom[ia].atomicNumber) { return -1; }
    if (atom[ia].atomicNumber > mi.atom[ia].atomicNumber) { return 1; }
    if (atom[ia].atomAttrib < mi.atom[ia].atomAttrib) { return -1; }
    if (atom[ia].atomAttrib > mi.atom[ia].atomAttrib) { return 1; }
    nBonds = atom[ia].numBonds;
    if (nBonds < mi.atom[ia].numBonds) { return -1; }
    if (nBonds > mi.atom[ia].numBonds) { return 1; }
    for (ib = 0; ib < nBonds; ib++) {
      if (atom[ia].bond[ib].order < mi.atom[ia].bond[ib].order)
        { return -1; }
      if (atom[ia].bond[ib].order > mi.atom[ia].bond[ib].order)
        { return 1; }
      if (atom[ia].bond[ib].toAtomicNbr < mi.atom[ia].bond[ib].toAtomicNbr)
        { return -1; }
      if (atom[ia].bond[ib].toAtomicNbr > mi.atom[ia].bond[ib].toAtomicNbr)
        { return 1; }
      if (atom[ia].bond[ib].toAtomIndex < mi.atom[ia].bond[ib].toAtomIndex)
        { return -1; }
      if (atom[ia].bond[ib].toAtomIndex > mi.atom[ia].bond[ib].toAtomIndex)
        { return 1; }
    }
  }
  if (numAtoms < mi.numAtoms) { return -1; }
  if (numAtoms > mi.numAtoms) { return 1; }
  return 0;
}

// ------------------------------------------------------ [getRingSubList]
//   get list of substitution pattern in the ring

void molecInfo::getRingSubList(atomChain &ring, vector<vector<int> > &sl,
 vector<int> &bor, vector<vector<int> > &isa) {
  int nRingAtoms, idxAtom, atomicNbr, idxNext, nBonds, idxTo;
  int numSubChains = 0, idxChain, sumBO, valence, numH;
  int ia, iaNxt, js, ks, ib, ic, ih;
  vector<int> vi0(2, 0);
  molecByChains subc;
  vector<molecByChains> subChains;

  nRingAtoms = ring.numAtoms;
  sl.clear(); sl.assign(nRingAtoms, vi0);
  bor.clear(); bor.assign(nRingAtoms, 0);
  isa.clear(); isa.assign(nRingAtoms, vi0);
  subChains.clear();
  for (ia = 0; ia < nRingAtoms; ia++) {
    for (js = 0; js < 2; js++) { sl[ia][js] = 0; isa[ia][js] = -1; }
    iaNxt = ia + 1;
    if (iaNxt >= nRingAtoms) { iaNxt -= nRingAtoms; }
    js = 0; ks = 0; sumBO = 0;
    idxAtom = ring.atomIndex[ia]; idxNext = ring.atomIndex[iaNxt];
    atomicNbr = atom[idxAtom].atomicNumber;
    nBonds = atom[idxAtom].numBonds;
    for (ib = 0; ib < nBonds; ib++) {
      idxTo = atom[idxAtom].bond[ib].toAtomIndex;
      sumBO += atom[idxAtom].bond[ib].order;
      if (idxTo == idxNext) {
        bor[ia] = atom[idxAtom].bond[ib].order;
      } else if (!ring.isIdxInChain(idxTo)) {
        isa[ia][ks] = idxTo;
        subc.clear();
        followChain(-idxAtom, -idxTo, subc);
        subc.beUnique();
        for (ic = 0, idxChain = -1; ic < numSubChains; ic++)
          { if (subChains[ic] == subc) { idxChain = ic; break; } }
        if (idxChain < 0) {
          subChains.push_back(subc);
          idxChain = numSubChains;
          numSubChains++;
        }
        sl[ia][js] = idxChain + 3;
        js++; ks++;
      }
    }
    valence = elmt[atomicNbr].valence;
    numH = valence - sumBO - atom[idxAtom].atomAttrib;
    for (ih = 0; ih < numH; ih++) { sl[ia][js] = 1; js++; }
    if ((atomicNbr == 7) && (js == 1)) { sl[ia][js] = 2; js++; }
    if (sl[ia][1] > sl[ia][0]) {
     idxChain = sl[ia][0]; sl[ia][0] = sl[ia][1]; sl[ia][1] = idxChain;
    }
  }
}

// ------------------------------------------------------ [getRotatedRepr]
//   get all rotated representations (for getSymIsomCyclic)

void molecInfo::getRotatedRepr(int nA, string repr, vector<string> *vs) {
  int ia, iaBef, pos, iSt, ii;
  string reprRev, sHead, sPatt, sPattRev, sS0, sS1, sB, sf, sr;
  if (!vs->empty()) { vs->clear(); }
  reprRev = "";
  for (ia = nA - 1; ia >= 0; ia--) {
    iaBef = ia - 1;
    if (iaBef < 0) { iaBef += nA; }
    pos = ia * 9; sHead = repr.substr(pos, 3);
    sPatt = repr.substr(pos + 3, 1);
    sPattRev = format(-(atoi(sPatt.c_str()) - 1) + 1, 1);
    sS0 = repr.substr(pos + 4, 2); sS1 = repr.substr(pos + 6, 2);
    sB = repr.substr(iaBef * 9 + 8, 1);
    if (sS1 == "00") { reprRev += sHead + sPattRev + sS0 + sS1 + sB; }
    else { reprRev += sHead + sPattRev + sS1 + sS0 + sB; }
  }
  for (iSt = 0; iSt < nA; iSt++) {
    sf = ""; sr = "";
    for (ii = 0; ii < nA; ii++) {
      ia = iSt + ii;
      if (ia >= nA) { ia -= nA; }
      sf += repr.substr(ia * 9, 9);
      sr += reprRev.substr(ia * 9, 9);
    }
    vs->push_back(sf); vs->push_back(sr);
  }
}

// ------------------------------------------------------ [followBondsSub]
//   follow and list-up bonds in a substituent

void molecInfo::followBondsSub(int idx, int idxTo, listBonds &lb) {
  int nBonds, idxToNxt, bOrdTo;
  int ib;
  chemBond bnd;
  nBonds = atom[idxTo].numBonds;
  for (ib = 0; ib < nBonds; ib++) {
    idxToNxt = atom[idxTo].bond[ib].toAtomIndex;
    if (idxToNxt == idx) { bOrdTo = atom[idxTo].bond[ib].order; }
  }
  bnd.setFrom(bOrdTo, idx, idxTo);
  if (lb.addBond(bnd) == ERRE) {
    cout << "ERROR molecInfo::followBondsSub [duplication-cyclic?]\n";
    return;
  }
  for (ib = 0; ib < nBonds; ib++) {
    idxToNxt = atom[idxTo].bond[ib].toAtomIndex;
    if (idxToNxt != idx) { followBondsSub(idxTo, idxToNxt, lb); }
  }
  return;
}


// -------------------------------------------------------- [getIntRotSym]
//   get internal rotor symmetry aroung a single bond

int molecInfo::getIntRotSym(int cntPiv, int Piv, int &farLin) {
  int numB, numNB = 0, sumBO = 0, ib;
  int atmNbr, nxtAtmOrd[3], nxtAtmNbr[3], nxtAtmIdx[3], inb;
  int nxnxAtmIdx, valence, numH, numTotNB;
  molecByChains *molc[3];
  numB = atom[Piv].numBonds;
  for (ib = 0; ib < numB; ib++) {
    if (atom[Piv].bond[ib].toAtomIndex == cntPiv) {
      if (atom[Piv].bond[ib].order > 1)
        { cout << "ERROR molecInfo::getIntRotSym\n"; return 0; }
    } else { numNB++; }
    sumBO += atom[Piv].bond[ib].order;
  }
  if (numNB != (numB - 1))
    { cout << "ERROR molecInfo::getIntRotSym\n"; return 0; }
  atmNbr = atom[Piv].atomicNumber;
  valence = elmt[atmNbr].valence;
  numH = valence - sumBO - atom[Piv].atomAttrib;
  if (numH < 0) { numH = 0; }
  numTotNB = numH + numNB;
  if (numTotNB == 0) { return 0; }
  for (ib = 0 ,inb = 0; ib < numB; ib++) {
    if (atom[Piv].bond[ib].toAtomIndex != cntPiv) {
      nxtAtmOrd[inb] = atom[Piv].bond[ib].order;
      nxtAtmNbr[inb] = atom[Piv].bond[ib].toAtomicNbr;
      nxtAtmIdx[inb] = atom[Piv].bond[ib].toAtomIndex;
      inb++;
    }
  }
  switch (atmNbr) {
  case 7:
    if (numNB != 2) { return 1; }
    for (inb = 0; inb < numNB; inb++)
      { if (nxtAtmOrd[inb] != 2) { return 1; } }
    break;
  case 6:
    switch (numTotNB) {
    case 1:
      if (numNB != 1) { return 1; }
      if (nxtAtmOrd[0] != 3) { return 1; }
      if (nxtAtmNbr[0] == 7) { return 0; }
      if (atom[nxtAtmIdx[0]].numBonds == 1) { return 0; }
      for (ib = 0; ib < atom[nxtAtmIdx[0]].numBonds; ib++) {
        if (atom[nxtAtmIdx[0]].bond[ib].toAtomIndex != Piv)
          { nxnxAtmIdx = atom[nxtAtmIdx[0]].bond[ib].toAtomIndex; }
      }
      farLin++;
      return getIntRotSym(nxtAtmIdx[0], nxnxAtmIdx, farLin);
      break;
    case 2:
      if (atom[Piv].atomAttrib != 1) { return 1; }
      if (numH == 2) { return 2; }
      if (numNB != 2) { return 1; }
      break;
    case 3:
      if (numH == 3) { return 3; }
      if (numNB != 3) { return 1; }
      break;
    default: return 1; break;
    }
    break;
  default: return 1; break;
  }
  for (inb = 1; inb < numNB; inb++)
    { if (nxtAtmNbr[inb] != nxtAtmNbr[0]) { return 1; } }
  for (inb = 0; inb < numNB; inb++) {
    molc[inb] = new molecByChains();
    followChain(-Piv, -nxtAtmIdx[inb], *molc[inb]);
    molc[inb]->beUnique();
  }
  for (inb = 1; inb < numNB; inb++) {
    if (*molc[inb] != *molc[0]) { return 1; }
  }
  return numNB;
}

// -------------------------------------------------------- [getDBnRotSym]
//   get rotor symmetry aroud the double bond

int molecInfo::getDBnRotSym(int cntPiv, int Piv) {
  int pvAtmcNb, pvAtmAtt, pvNmBnds, cbOrd, pvIbCB, pvSumBO, pvNumH;
  int pvNmOBs, pvIdxTo[2], pvBOTo[2];
  int ib, bo, toid;
  molecByChains branch[2];

  pvAtmcNb = atom[Piv].atomicNumber; pvAtmAtt = atom[Piv].atomAttrib;
  pvNmBnds = atom[Piv].numBonds; pvSumBO = 0; pvNmOBs = 0;
  for (ib = 0; ib < pvNmBnds; ib++) {
    bo = atom[Piv].bond[ib].order; pvSumBO += bo;
    toid = atom[Piv].bond[ib].toAtomIndex;
    if (toid == cntPiv)
      { cbOrd = bo; pvIbCB = ib; }
    else { pvBOTo[pvNmOBs] = bo; pvIdxTo[pvNmOBs] = toid; pvNmOBs++; }
  }
  if (cbOrd != 2) {
    cout << "ERROR molecInfo::getDBnRotSym [not double bond]\n";
    return 0;
  }
  pvNumH = elmt[pvAtmcNb].valence - pvSumBO - pvAtmAtt;
  if (pvNumH < 0) { pvNumH = 0; }
  if (pvNmOBs == 0) { return pvNumH; }
  if (pvNmOBs == 1) {
    if (pvBOTo[0] == 2) { return getDBnRotSym(Piv, pvIdxTo[0]); }
    else { return 1; }
  }
  for (ib = 0; ib < pvNmOBs; ib++) {
    branch[ib].clear();
    if (followChain(-Piv, -pvIdxTo[ib], branch[ib]) == ERRE) {
      cout << "ERROR molecInfo::getDBnRotSym [follow chain]\n";
      return 0;
    }
    branch[ib].beUnique();
  }
  if (branch[0] == branch[1]) { return 2; }
  return 1;
}

// --------------------------------------------------- [countAllIntRotSym]
//   count all internal rotors and symmetry

int molecInfo::countAllIntRotSym(int &intRotSym) {
  int numIntRot = 0, nBonds, idxTo, frL1, frL2, bSym1, bSym2;
  int ia, ib, ic, nIRres = 0;

  intRotSym = 1;
  for (ia = 0; ia < numAtoms; ia++) {
    nBonds = atom[ia].numBonds;
    for (ib = 0; ib < nBonds; ib++) {
      idxTo = atom[ia].bond[ib].toAtomIndex;
      if (idxTo > ia) { if (atom[ia].bond[ib].order == 1) {
        frL1 = 0; bSym1 = getIntRotSym(ia, idxTo, frL1);
        frL2 = 0; bSym2 = getIntRotSym(idxTo, ia, frL2);
        if ((bSym1 * bSym2) != 0) {
          if ((frL1 + frL2) == 0) { numIntRot++; }
          else if ((frL1 * frL2) == 0) { nIRres++; }
        }
        if ((frL1 == 0) && (bSym1 > 1)) { intRotSym *= bSym1; }
        if ((frL2 == 0) && (bSym2 > 1)) { intRotSym *= bSym2; }
      }}
    }
  }
  if ((nIRres % 2) == 1) {
    cout << "ERROR molecInfo::countAllIntRotSym [unexpected nIRres]\n";
  }
  numIntRot += (nIRres / 2);
  return numIntRot;
}

// --------------------------------------------------- [getListPlanarCent]
//   get a list string of planar centers in a ring

string molecInfo::getListPlanarCent(atomChain &ring) {
  int nRingAtoms, idxAtom, isPlanCnt, nBonds;
  int ia, ib;
  string rets = "";
  nRingAtoms = ring.numAtoms;
  for (ia = 0; ia < nRingAtoms; ia++) {
    idxAtom = ring.atomIndex[ia];
    if ((atom[idxAtom].atomAttrib > 0) &&
        (atom[idxAtom].atomicNumber == 6)) { isPlanCnt = true; }
    else {
      nBonds = atom[idxAtom].numBonds;
      isPlanCnt = false;
      for (ib = 0; ib < nBonds; ++ib) {
        if (atom[idxAtom].bond[ib].order > 1) { isPlanCnt = true; break; }
      }
    }
    if (isPlanCnt) { rets += '*'; }
    else { rets += ' '; }
  }
  return rets;
}

// ------------------------------------------------- [getPattMatchPlanCnt]
//   get pattern strings matches with the string of planar center

int molecInfo::getPattMatchPlanCnt(string &sPlC, vector<string> *patt) {
  /*-----------------------------------------------
    patterns of atom positions:
     * is planar center
     2 .. up / 1 .. middle / 0 .. down
  -----------------------------------------------*/
  static int numRefs = 52;
  static string refs[52][2] = {
   {"   ", "111"}, {"*  ", "111"},
   {"** ", "111"}, {"***", "111"},                         // 4  (3m-ring)
   {"    ", "2020"}, {"*   ", "2020"}, {"* * ", "2020"},
   {"**  ", "1111"}, {"*** ", "1111"}, {"****", "1111"},   // 6  (4m-ring)
   {"     ", "11211"}, {"  *  ", "12101"},
   {" * * ", "12101"}, {"*   *", "11211"},
   {"* * *", "11211"}, {"***  ", "11111"},
   {"**** ", "11111"}, {"*****", "11111"},                 // 8  (5m-ring)
   {"      ", "202020"}, {"      ", "211211"},
   {"*     ", "211011"}, {"*     ", "211211"},
   {"  **  ", "211110"}, {"  * * ", "211011"}, {"  * * ", "211211"},
   {"*  *  ", "211011"}, {"*  *  ", "211211"},
   {"  *** ", "211111"}, {"*  ** ", "201111"}, {"* * * ", "202020"},
   {"  ****", "201111"}, {" ** **", "211211"}, {"* *** ", "211111"},
   {" *****", "211111"}, {"******", "111111"},             // 17 (6m-ring)
   {"       ", "1021021"}, {"   *   ", "1021021"},
   {"*     *", "1121011"}, {"  * *  ", "1021021"}, {" *   * ", "1021021"},
   {"  ***  ", "0111112"}, {"*  *  *", "1101211"}, {"* *   *", "1101211"},
   {" * * * ", "1021021"}, {"**   **", "1112111"}, {" ** ** ", "1001221"},
   {"* ***  ", "0111112"}, {" ***** ", "0111112"}, {"* *** *", "2111112"},
   {"** * **", "1112111"}, {"*** ***", "1112111"}, {"*******", "1021021"}
                                                           // 17 (7m-ring)
  };
  string rfs, pts, rfsRot, ptsRot, ptsRotRev, ptsCut;
  int nRingAtoms, direct, posStart, pos;
  int ir, ia;
  if (!patt->empty()) { patt->clear(); }
  for (ir = 0; ir < numRefs; ir++) {
    rfs = refs[ir][0]; pts = refs[ir][1];
    nRingAtoms = rfs.length();
    for (direct = -1; direct <= 1; direct += 2) {
      for (posStart = 0; posStart < nRingAtoms; posStart++) {
        rfsRot = ""; ptsRot = ""; ptsRotRev = "";
        for (ia = 0; ia < nRingAtoms; ia++) {
          pos = posStart + (direct * ia);
          if (pos < 0) { pos += nRingAtoms; }
          else if (pos >= nRingAtoms) { pos -= nRingAtoms; }
          rfsRot += rfs[pos];
          ptsCut = pts.substr(pos, 1); ptsRot += ptsCut;
          ptsRotRev += format(-(atoi(ptsCut.c_str()) - 1) + 1, 1);
        }
        if (sPlC == rfsRot) {
          if (numMatchStrList(ptsRot, patt) == 0)
            { patt->push_back(ptsRot); }
          if (numMatchStrList(ptsRotRev, patt) == 0)
            { patt->push_back(ptsRotRev); }
        }
      }
    }
  }
  return patt->size();
}

// ------------------------------------------------------- [isTdSymmetric]
//   check whether the molecule has Td (tetrahedral) symmetry

int molecInfo::isTdSymmetric(atomChain &mc, int neqc, int &numIntRot,
 int &intRotSym) {
  int idxSymCnt, idxBrncTo, nBonds, bOrd, idx1, idx2, farLin, bSym;
  int ib, iRotS;
  molecByChains branch[4];
  listBonds lstBonds;

  // --- is CH4 (methane) ?
  if (numAtoms == 1) {
    if ( (atom[0].atomicNumber == 6) && (atom[0].atomAttrib == 0)
         && (atom[0].numBonds == 0) )
      { numIntRot = 0; intRotSym = 1; return true; }
  }

  // --- should have 4n equivalent chains with odd number atoms
  if ((neqc % 4) != 0) { return false; }
  if ((mc.numAtoms % 2) != 1) { return false; }

  // --- follow chains from center
  idxSymCnt = mc.atomIndex[(mc.numAtoms - 1) / 2];
  if (atom[idxSymCnt].numBonds != 4) { return false; }
  for (ib = 0; ib < 4; ib++) {
    idxBrncTo = atom[idxSymCnt].bond[ib].toAtomIndex;
    branch[ib].clear();
    if (followChain(-idxSymCnt, -idxBrncTo, branch[ib]) == ERRE) {
      cout << "ERROR molecInfo::isTdSymmetric [follow chain]\n";
      return false;
    }
    branch[ib].beUnique();
    if (ib > 0) {
      if (branch[ib] != branch[0]) { return false; }
    }
  }

  // --- symm. number of bonds in a branch
  idxBrncTo = atom[idxSymCnt].bond[0].toAtomIndex;
  lstBonds.clear();
  followBondsSub(idxSymCnt, idxBrncTo, lstBonds);
  nBonds = lstBonds.size();
  iRotS = 1; numIntRot = 0;
  for (ib = 0; ib < nBonds; ib++) {
    bOrd = lstBonds[ib].order;
    idx1 = lstBonds[ib].idxFirst; idx2 = lstBonds[ib].idxSecond;
    if (bOrd == 1) {
      farLin = 0;
      bSym = getIntRotSym(idx1, idx2, farLin);
      if ((farLin == 0) && (bSym > 0))
        { iRotS *= bSym; numIntRot++; }
    }
  }
  intRotSym = iRotS * iRotS * iRotS * iRotS;
  numIntRot *= 4;
  return true;
}

// ----------------------------------------------------- [isC3D3Symmetric]
//   detect C3/D3 symmetry except for those detected as internal
//    rotor symmetry (if D3, numIntRot contains 2)

int molecInfo::isC3D3Symmetric(atomChain &mc, int neqc, int &numIntRot,
 int &intRotSym) {
  int idxSymCnt, idxBrncTo, nBonds, bOrd, idx1, idx2, farLin, bSym;
  int ib, iRotS;
  molecByChains branch[3];
  listBonds lstBonds;

  // --- is NH3 or CH3 ?
  if (numAtoms == 1) {
    if ( (atom[0].atomicNumber == 7) && (atom[0].atomAttrib == 0)
         && (atom[0].numBonds == 0) )
      { numIntRot = 0; intRotSym = 1; return true; }
    if ( (atom[0].atomicNumber == 6) && (atom[0].atomAttrib == 1)
         && (atom[0].numBonds == 0) )
      { numIntRot = 0; intRotSym = 2; return true; }
  }

  // --- should have 3n equivalent chains with odd number atoms
  if ((neqc % 3) != 0) { return false; }
  if ((mc.numAtoms % 2) != 1) { return false; }

  // --- follow chains from center
  idxSymCnt = mc.atomIndex[(mc.numAtoms - 1) / 2];
  if (atom[idxSymCnt].numBonds != 3) { return false; }
  for (ib = 0; ib < 3; ib++) {
    idxBrncTo = atom[idxSymCnt].bond[ib].toAtomIndex;
    branch[ib].clear();
    if (followChain(-idxSymCnt, -idxBrncTo, branch[ib]) == ERRE) {
      cout << "ERROR molecInfo::isC3D3Symmetric [follow chain]\n";
      return false;
    }
    branch[ib].beUnique();
    if (ib > 0) {
      if (branch[ib] != branch[0]) { return false; }
    }
  }
  if ( (atom[idxSymCnt].atomicNumber == 6)
       && (atom[idxSymCnt].atomAttrib == 1) ) { intRotSym = 2; }
  else { intRotSym = 1; }

  // --- symm. number of bonds in a branch
  idxBrncTo = atom[idxSymCnt].bond[0].toAtomIndex;
  lstBonds.clear();
  followBondsSub(idxSymCnt, idxBrncTo, lstBonds);
  nBonds = lstBonds.size();
  iRotS = 1; numIntRot = 0;
  for (ib = 0; ib < nBonds; ib++) {
    bOrd = lstBonds[ib].order;
    idx1 = lstBonds[ib].idxFirst; idx2 = lstBonds[ib].idxSecond;
    if (bOrd == 1) {
      farLin = 0;
      bSym = getIntRotSym(idx1, idx2, farLin);
      if ((farLin == 0) && (bSym > 0))
        { iRotS *= bSym; numIntRot++; }
    }
  }
  intRotSym *= iRotS * iRotS * iRotS;
  numIntRot *= 3;
  return true;
}

// ----------------------------------------------------- [isC2D2Symmetric]
//   detect C2/D2 symmetry except for those detected as internal
//    rotor symmetry (if D2, numIntRot contains 2)

int molecInfo::isC2D2Symmetric(atomChain &mc, int neqc, int &numIntRot,
 int &intRotSym) {
  int cbIdx[2], cbOrd, cbNBNxt[2], cbIdxNxt[2][3], cbSubIdx[2][3];
  int caIdx, caNBNxt, caIdxNxt[4], caSubIdx[4];
  int isCentBond, numBrnc, isNewBrnc, posCh, nBonds;
  int iRotS, nIntR, bOrd, idx1, idx2, farLin, bSym;
  int ic, ib, jb, idn, idc, atmcNbr, atmAttr, numH, ir, il;
  molecByChains branch[6], brnc;
  listBonds lstBonds;

  // --- is H2O, NH2 or CH2 ?
  if (numAtoms == 1) {
    if ( (atom[0].atomicNumber == 8) && (atom[0].atomAttrib == 0)
         && (atom[0].numBonds == 0) )
      { numIntRot = 0; intRotSym = 1; return true; }
    if ( (atom[0].atomicNumber == 7) && (atom[0].atomAttrib == 1)
         && (atom[0].numBonds == 0) )
      { numIntRot = 0; intRotSym = 1; return true; }
    if ( (atom[0].atomicNumber == 6) && (atom[0].atomAttrib == 2)
         && (atom[0].numBonds == 0) )
      { numIntRot = 0; intRotSym = 1; return true; }
  }

  // --- should have 2n equivalent chains / check symmetric center
  if ((neqc % 2) != 0) { return false; }
  if ((mc.numAtoms % 2) == 0) { isCentBond = true; }
  else { isCentBond = false; }
  if (isCentBond) {   // --- a bond may be center of symmetry
    posCh = (mc.numAtoms / 2) - 1;
    for (ic = 0; ic < 2; ic++) { cbIdx[ic] = mc.atomIndex[posCh + ic]; }
    for (ic = 0; ic < 2; ic++) {
      nBonds = atom[cbIdx[ic]].numBonds;
      cbNBNxt[ic] = 0;
      for (ib = 0; ib < nBonds; ib++) {
        idn = atom[cbIdx[ic]].bond[ib].toAtomIndex;
        if (ic == 0) { idc = cbIdx[1]; } else { idc = cbIdx[0]; }
        if (idn == idc) { cbOrd = atom[cbIdx[ic]].bond[ib].order; }
        else { cbIdxNxt[ic][cbNBNxt[ic]] = idn; cbNBNxt[ic]++; }
      }
    }
    if (cbNBNxt[0] != cbNBNxt[1]) { return false; }
    if (cbNBNxt[0] == 0) {
      if ( (atom[cbIdx[0]].atomicNumber != atom[cbIdx[1]].atomicNumber)
           || (atom[cbIdx[0]].atomAttrib != atom[cbIdx[1]].atomAttrib) )
        { return false; }
      atmcNbr = atom[cbIdx[0]].atomicNumber;
      atmAttr = atom[cbIdx[0]].atomAttrib;
      numH = elmt[atmcNbr].valence - cbOrd - atmAttr;
      numIntRot = 0; intRotSym = 1;
      if (cbOrd == 1) { numIntRot = 1; }
      switch (numH) {
       case 3: intRotSym = 9; break;
       case 2:
        if (atmcNbr == 6) { intRotSym = 2 * (numIntRot + 1); }
        break;
       case 0: numIntRot = 0; break;
      }
      return true;
    }
  } else {   // --- an atom may be center of symmetry
    posCh = (mc.numAtoms - 1) / 2;
    caIdx = mc.atomIndex[posCh];
    caNBNxt = atom[caIdx].numBonds;
    for (ib = 0; ib < caNBNxt; ib++)
      { caIdxNxt[ib] = atom[caIdx].bond[ib].toAtomIndex; }
    if ((caNBNxt < 2) || (caNBNxt == 3)) { return false; }
    if ((atom[caIdx].atomicNumber == 7) && (atom[caIdx].atomAttrib != 1))
      { return false; }
  }

  // --- follow chains from the center
  if (isCentBond) {   // --- center = bond
    numBrnc = 0;
    for (ic = 0; ic < 2; ic++) {
      for (ib = 0; ib < cbNBNxt[ic]; ib++) {
        brnc.clear();
        if (followChain(-cbIdx[ic], -cbIdxNxt[ic][ib], brnc) == ERRE) {
          cout << "ERROR molecInfo::isC2D2Symmetric [follow chain]\n";
          return false;
        }
        brnc.beUnique(); isNewBrnc = true;
        for (ir = 0; ir < numBrnc; ir++) {
          if (brnc == branch[ir])
            { isNewBrnc = false; cbSubIdx[ic][ib] = ir; break; }
        }
        if (isNewBrnc) {
          branch[numBrnc] = brnc; cbSubIdx[ic][ib] = numBrnc;
          numBrnc++;
        }
      }
      for (ib = 0; ib < cbNBNxt[ic] - 1; ib++) {
        for (jb = ib + 1; jb < cbNBNxt[ic]; jb++) {
          if (cbSubIdx[ic][jb] < cbSubIdx[ic][ib]) {
            ir = cbSubIdx[ic][ib]; cbSubIdx[ic][ib] = cbSubIdx[ic][jb];
            cbSubIdx[ic][jb] = ir;
          }
        }
      }
    }
    for (ib = 0; ib < cbNBNxt[0]; ib++) {
      if (cbSubIdx[0][ib] != cbSubIdx[1][ib]) { return false; }
    }
    numIntRot = 0; intRotSym = 1;
    if (cbOrd == 1) { numIntRot = 1; }
    switch (cbNBNxt[0]) {
     case 3:
      if (cbSubIdx[0][2] == 0) { intRotSym = 9; }
      break;
     case 2:
      atmcNbr = atom[cbIdx[0]].atomicNumber;
      atmAttr = atom[cbIdx[0]].atomAttrib;
      if (cbSubIdx[0][1] == 0) {
        if (atmcNbr == 6) {
          if (atmAttr == 1) { intRotSym = 4; }
          if (cbOrd == 2) { intRotSym = 2; }
        }
      }
      break;
     case 1:
      if ((cbOrd == 2) && isAlleneC2Sym(cbIdx[0], cbIdxNxt[0][0]))
        { intRotSym = 2; }
      if (cbOrd == 1) {
        farLin = 0;
        if (getIntRotSym(cbIdx[0], cbIdx[1], farLin) == 0)
          { numIntRot = 0; }
      }
      break;
    }
  } else {   // --- center = atom
    numIntRot = 0; intRotSym = 1;
    numBrnc = 0;
    for (ib = 0; ib < caNBNxt; ib++) {
      brnc.clear();
      if (followChain(-caIdx, -caIdxNxt[ib], brnc) == ERRE) {
        cout << "ERROR molecInfo::isC2D2Symmetric [follow chain]\n";
        return false;
      }
      brnc.beUnique(); isNewBrnc = true;
      for (ir = 0; ir < numBrnc; ir++) {
        if (brnc == branch[ir])
          { isNewBrnc = false; caSubIdx[ib] = ir; break; }
      }
      if (isNewBrnc) {
        branch[numBrnc] = brnc; caSubIdx[ib] = numBrnc;
        numBrnc++;
      }
    }
    if (caSubIdx[1] != 0) { return false; }
    if (caNBNxt == 4) { if (caSubIdx[3] != 1) { return false; } }
    if (caNBNxt == 2) {
      if (isAlleneC2Sym(caIdx, caIdxNxt[0])) { intRotSym = 2; }
    }
  }

  // --- symm. number of bonds in branches
  if (isCentBond) {
    for (ib = 0; ib < cbNBNxt[0]; ib++) {
      lstBonds.clear();
      followBondsSub(cbIdx[0], cbIdxNxt[0][ib], lstBonds);
      nBonds = lstBonds.size();
      iRotS = 1; nIntR = 0;
      for (il = 0; il < nBonds; il++) {
        bOrd = lstBonds[il].order;
        idx1 = lstBonds[il].idxFirst; idx2 = lstBonds[il].idxSecond;
        if (bOrd == 1) {
          farLin = 0;
          bSym = getIntRotSym(idx1, idx2, farLin);
          if ((farLin == 0) && (bSym > 0))
            { iRotS *= bSym; nIntR++; }
        }
      }
      intRotSym *= (iRotS * iRotS); numIntRot += (nIntR * 2);
    }
  } else {
    for (ib = 0; ib < caNBNxt; ib++) {
      lstBonds.clear();
      followBondsSub(caIdx, caIdxNxt[ib], lstBonds);
      nBonds = lstBonds.size();
      iRotS = 1; nIntR = 0;
      for (il = 0; il < nBonds; il++) {
        bOrd = lstBonds[il].order;
        idx1 = lstBonds[il].idxFirst; idx2 = lstBonds[il].idxSecond;
        if (bOrd == 1) {
          farLin = 0;
          bSym = getIntRotSym(idx1, idx2, farLin);
          if ((farLin == 0) && (bSym > 0))
            { iRotS *= bSym; nIntR++; }
        }
      }
      intRotSym *= iRotS; numIntRot += nIntR;
    }
  }
  return true;
}

// ----------------------------------------------------- [isDBC2Symmetric]
//   detect C2 symmetry centered by a carbon double bond

int molecInfo::isDBC2Symmetric(int &numIntRot, int &intRotSym) {
  int nBonds, sFw, sBc, foundC2cent = false, nBndsLst;
  int ia, ib, idxTo, idxC[2], ic, icc, il;
  int bOrd, idx1, idx2, farLin, bSym;
  listBonds lstBonds;

  // --- search for C2 symmetry center
  for (ia = 0; ia < numAtoms; ia++) {
    nBonds = atom[ia].numBonds;
    for (ib = 0; ib < nBonds; ib++) {
      idxTo = atom[ia].bond[ib].toAtomIndex;
      if ((idxTo > ia) && (atom[ia].bond[ib].order == 2)) {
        sFw = getDBnRotSym(ia, idxTo); sBc = getDBnRotSym(idxTo, ia);
        if (((sFw == 2) && (sBc != 1)) || ((sBc == 2) && (sFw != 1)))
          { foundC2cent = true; idxC[0] = ia; idxC[1] = idxTo; break; }
      }
    }
    if (foundC2cent) { break; }
  }
  if (!foundC2cent) { return false; }

  // --- symm. number of bonds in branches
  numIntRot = 0; intRotSym = 1;
  for (ic = 0; ic < 2; ic++) {
    icc = ic + 1; if (icc >= 2) { icc -= 2; }
    nBonds = atom[idxC[ic]].numBonds;
    for (ib = 0; ib < nBonds; ib++) {
      idxTo = atom[idxC[ic]].bond[ib].toAtomIndex;
      if (idxTo != idxC[icc]) {
        lstBonds.clear();
        followBondsSub(idxC[ic], idxTo, lstBonds);
        nBndsLst = lstBonds.size();
        for (il = 0; il < nBndsLst; il++) {
          bOrd = lstBonds[il].order;
          idx1 = lstBonds[il].idxFirst; idx2 = lstBonds[il].idxSecond;
          if (bOrd == 1) {
            farLin = 0;
            bSym = getIntRotSym(idx1, idx2, farLin);
            if ((farLin == 0) && (bSym > 0))
              { intRotSym *= bSym; numIntRot++; }
          }
        }
      }
    }
  }
  return true;
}

// ------------------------------------------------------------ [isChiral]
//   check whether the molecule is chiral

int molecInfo::isChiral() {
  int ia;
  for (ia = 0; ia < numAtoms; ia++) {
    if (isChiralCent(ia)) { return true; }
  }
  return false;
}

// -------------------------------------------------------- [isChiralCent]
//   check whether the atom is chiral center

int molecInfo::isChiralCent(int idx) {
  int atmcNbr, atmAttr, valence, nBonds, sumBO, nBrnc = 0;
  int ib, idxTo, il;
  molecByChains branch[4], brnc;

  atmcNbr = atom[idx].atomicNumber;
  atmAttr = atom[idx].atomAttrib;
  valence = elmt[atmcNbr].valence;
  nBonds = atom[idx].numBonds;
  sumBO = 0;
  for (ib = 0; ib < nBonds; ib++) { sumBO += atom[idx].bond[ib].order; }
  if (valence < 3) { return false; }
  if (atmAttr != 0) { return false; }
  if (nBonds < (valence - 1)) { return false; }
  if (sumBO != nBonds) { return false; }
  for (ib = 0; ib < nBonds; ib++) {
    idxTo = atom[idx].bond[ib].toAtomIndex;
    brnc.clear();
    if (followChain(-idx, -idxTo, brnc) == ERRE) {
      cout << "ERROR molecInfo::isChiralCent [follow chain]\n";
      return false;
    }
    brnc.beUnique();
    for (il = 0; il < nBrnc; il++)
      { if (brnc == branch[il]) { return false; } }
    branch[nBrnc] = brnc; nBrnc++;
  }
  return true;
}

// ------------------------------------------------------- [isAlleneC2Sym]
//   check whether the part has allene C2v symmetry

int molecInfo::isAlleneC2Sym(int idx1, int idx2) {
  int nBonds, idxTo, nBrnc;
  int ib;
  molecByChains branch[2];

  if (atom[idx1].atomicNumber != 6) { return false; }
  if (atom[idx2].atomicNumber != 6) { return false; }
  nBonds = atom[idx2].numBonds;
  for (ib = 0; ib < nBonds; ib++) {
    if (atom[idx2].bond[ib].toAtomIndex == idx1) {
      if (atom[idx2].bond[ib].order != 2) { return false; }
      break;
    }
  }
  if (nBonds == 1) {
    if (atom[idx2].atomAttrib == 0) { return true; }
    else { return false; }
  }
  if (nBonds == 2) {
    for (ib = 0; ib < nBonds; ib++) {
      idxTo = atom[idx2].bond[ib].toAtomIndex;
      if (idxTo != idx1) { return isAlleneC2Sym(idx2, idxTo); }
    }
  }
  for (ib = 0, nBrnc = 0; ib < nBonds; ib++) {
    idxTo = atom[idx2].bond[ib].toAtomIndex;
    if (idxTo != idx1) {
      branch[nBrnc].clear();
      if (followChain(-idx2, -idxTo, branch[nBrnc]) == ERRE) {
        cout << "molecInfo::isAlleneC2Sym [follow chain]\n";
        return false;
      }
      branch[nBrnc].beUnique();
      nBrnc++;
    }
  }
  if (branch[0] == branch[1]) { return true; }
  else { return false; }
}

// ----------------------------------------------------------- [sortAtoms]
//   pre-sorting of atoms in the molecule before
//    getting unique representation

void molecInfo::sortAtoms() {
  int i, j;
  atomInfo p;
  for (i = 0; i < numAtoms - 1; i++) {
    for (j = i + 1; j < numAtoms; j++) {
      if (atom[j] < atom[i]) {
        p = atom[i]; atom[i] = atom[j]; atom[j] = p;
        swapRefs(i, j);
      }
    }
  }
}

// ------------------------------------------------------------ [swapRefs]
//   swap atom indices in bonfInfo.toAtomIndex

void molecInfo::swapRefs(int ir, int jr) {
  int iAtom, iBond;
  if (ir == jr) { return; }
  if ((ir >= numAtoms) || (jr >= numAtoms)) { return; }
  for (iAtom = 0; iAtom < numAtoms; iAtom++) {
    for (iBond = 0; iBond < atom[iAtom].numBonds; iBond++) {
      if (atom[iAtom].bond[iBond].toAtomIndex == ir) {
        atom[iAtom].bond[iBond].toAtomIndex = jr;
      } else if (atom[iAtom].bond[iBond].toAtomIndex == jr) {
        atom[iAtom].bond[iBond].toAtomIndex = ir;
      }
    }
  }
}

int molecInfo::getTerminal() {   // ------------------------ [getTerminal]
  int i, numTerm = 0;
  sortAtoms();
  for (i = 0; i < numAtoms; i++) {
    if (atom[i].numBonds == 1) { numTerm++; }
    else { break; }
  }
  return numTerm;
}

               // ------------------------------------------ [followChain]
int molecInfo::followChain(int lev, int idx, molecByChains &molc) {
  static vector<int> chidx;
  vector<int> cyidx;
  int nNext, i, nidx, isCyclic = false, startCyc, nAtomCyc;

  if (lev <= 0) { chidx.clear(); }
  if ((lev < 0) || (idx < 0)) { 
    chidx.push_back(-lev);
    if (followChain(1, -idx, molc) == ERRE) { return ERRE; }
    else { return NORM; }
  }
  if (chidx.size() < (lev + 1)) { chidx.resize(lev + 1); }
  chidx[lev] = idx;
  nNext = atom[idx].numBonds;
  if (lev > 0) { nNext -= 1; }
  if (nNext == 0) {
    molc.addChain(atomChain(isCyclic, false, lev + 1, chidx, *this));
    return NORM;
  }
  for (i = 0; i < lev; i++)
    { if (chidx[i] == idx) { isCyclic = true; startCyc = i; break; } }
  if (isCyclic) {
    nAtomCyc = lev - startCyc;
    cyidx.assign(&chidx[startCyc], &chidx[lev]);
    molc.addChain(atomChain(isCyclic, false, nAtomCyc, cyidx, *this));
    return ERRE;
  }
  for (i = 0; i < atom[idx].numBonds; i++) {
    nidx = atom[idx].bond[i].toAtomIndex;
    if (lev > 0) {
      if (nidx != chidx[lev - 1]) {
        if (followChain(lev + 1, nidx, molc) == ERRE) { return ERRE; }
      }
    } else {
      if (followChain(lev + 1, nidx, molc) == ERRE) { return ERRE; }
    }
  }
  return NORM;
}

// --------------------------------------------------------- [listCyclics]
void molecInfo::listCyclics(int &numCases, atomChain &refc,
 vector<molecByChains> &molc) {
  int nAtomsCyc = refc.numAtoms, maxNumSubs = 0, i, j, k;
  int idxa, idxb, nSubs;
  vector<int> cyidx;
  molecByChains mc;

  for (i = 0; i < nAtomsCyc; i++) {
    idxa = refc.atomIndex[i];
    nSubs = atom[idxa].numBonds - 2;
    if (nSubs > maxNumSubs) { maxNumSubs = nSubs; }
  }
  numCases = 0; molc.clear();
  for (i = 0; i < nAtomsCyc; i++) {
    idxa = refc.atomIndex[i];
    if ((atom[idxa].numBonds - 2) == maxNumSubs) {
      // forward cycle
      cyidx.clear();
      for (j = 0; j < nAtomsCyc; j++) {
        k = i + j;
        if (k >= nAtomsCyc) { k -= nAtomsCyc; }
        cyidx.push_back(refc.atomIndex[k]);
      }
      mc.clear();
      mc.addChain(atomChain(true, false, nAtomsCyc, cyidx, *this));
      molc.push_back(mc);
      idxb = cyidx[nAtomsCyc - 1];
      followCycChain(0, idxa, idxb, idxa, molc[numCases]);
      numCases++;
      // backward cycle
      cyidx.clear();
      for (j = 0; j < nAtomsCyc; j++) {
        k = i - j;
        if (k < 0) { k += nAtomsCyc; }
        cyidx.push_back(refc.atomIndex[k]);
      }
      mc.clear();
      mc.addChain(atomChain(true, false, nAtomsCyc, cyidx, *this));
      molc.push_back(mc);
      idxb = cyidx[nAtomsCyc - 1];
      followCycChain(0, idxa, idxb, idxa, molc[numCases]);
      numCases++;
    }
  }
}

// ------------------------------------------------------ [followCycChain]
void molecInfo::followCycChain(int lev, int idxs, int idxe, int cidx,
  molecByChains &molc) {
  static vector<int> chidx;
  int nNext, i, nidx;

  if (lev == 0) { chidx.clear(); }
  if (chidx.size() < (lev + 1)) { chidx.resize(lev + 1); }
  chidx[lev] = cidx;
  nNext = atom[cidx].numBonds - 1;
  if (cidx == idxe) { nNext -= 1; }
  if (nNext <= 0) {
    if (cidx != idxe)
      { molc.addChain(atomChain(false, true, lev + 1, chidx, *this)); }
    return;
  }
  for (i = 0; i < atom[cidx].numBonds; i++) {
    nidx = atom[cidx].bond[i].toAtomIndex;
    if (cidx == idxs) {
      if (nidx != idxe)
        { followCycChain(lev + 1, idxs, idxe, nidx, molc); }
    } else if (cidx == idxe) {
      if ((nidx != idxs) && (nidx != chidx[lev - 1]))
        { followCycChain(lev + 1, idxs, idxe, nidx, molc); }
    } else {
      if (nidx != chidx[lev - 1])
        { followCycChain(lev + 1, idxs, idxe, nidx, molc); }
    }
  }
}

ostream &operator<<(ostream &os, molecInfo &m)   // -------- [operator <<]
  { m.print(&os); return os; }

/*============================================================================
     'molSMILES' and related classes - for SMILES import
============================================================================*/

// --- atomSMILES ------------------------------------------------------------

atomSMILES::atomSMILES() { clear(); }                   // --- constructor

void atomSMILES::clear() {                                    // --- clear
  atomInfo::clear();
  numRadcElc = 0; charge = 0; numDeloElc = 0;
  bracket = false;
  numHexp = 0; ringIdcs.clear();
}

void atomSMILES::corrDeloc() {                            // --- corrDeloc
   // correct number of delocalized electrons (numDeloElc)
  int ib;
  for (ib = 0; ib < numBonds; ib++) {
    if (bond[ib].isCommutable) { numDeloElc = 1; break; }
  }
}

int atomSMILES::sumOrderBonds() {                     // --- sumOrderBonds
  int ib, sumBO = 0;
  for (ib = 0; ib < numBonds; ib++) { sumBO += bond[ib].order; }
  return sumBO;
}

int atomSMILES::numBondingElec() {                   // --- numBondingElec
  return sumOrderBonds() + numRadcElc + numDeloElc;
}

int atomSMILES::numHatoms() {                             // --- numHatoms
  int valnc, hypvl, usedValnc = numBondingElec();
  valnc = elmt.ionValence(atomicNumber, charge, hypvl);
  while (((valnc - usedValnc) < 0) && (hypvl >= 2))
    { valnc += 2; hypvl -= 2; }
  return valnc - usedValnc;
}

ostream &atomSMILES::print(ostream &os) {                     // --- print
  os << " nRe=" << numRadcElc << " ch=" << charge << " nDe=" << numDeloElc;
  return atomInfo::print(os);
}

ostream &operator<<(ostream &os, atomSMILES &as) { return as.print(os); }

// --- molSMILES -------------------------------------------------------------

molSMILES::molSMILES() {                                // --- constructor
  numAtoms = 0; numRings = 0; atoms.clear();
  chBonds = "$:.-=#*";
  chOrgSb1st = "BbCcNnOoFPpSsI";
  str2cOrgSb.clear();
  str2cOrgSb.insert("Br"); str2cOrgSb.insert("Cl");
  sPars.erase(); cpPars = 0;
}

int molSMILES::setMol(string s, molecInfo &mi) {             // --- setMol
  sPars = s; cpPars = 0;
  int pend = sPars.length(), nodBranch = -1, ia;
  if (parseChain(pend, nodBranch) == ERRE)
    { cout << "molSMILES: SMILES parse error.\n"; return ERRE; }
  if (resolvRingClosure() == ERRE)
    { cout << "molSMILES: SMILES parse error (ring).\n"; return ERRE; }
  if (resolvBrcktExp() == ERRE)
    { cout << "molSMILES: SMILES parse error (valence).\n"; return ERRE; }
  resolvDeloElc();
   // check consistency with KUCRS
  if (isCharged()) {
    cout << "molSMILES: charged molecule is not supported by KUCRS.\n"; return ERRE;
  } else if (numRings > 1) {
    cout << "molSMILES: two or more ring molecule is not supported by KUCRS.\n"; return ERRE;
  }
  mi.clear();
  mi.numAtoms = numAtoms;
  for (ia = 0; ia < numAtoms; ia++) { mi.atom.push_back(atoms[ia]); }
  return NORM;
}

bool molSMILES::isCharged() {                             // --- isCharged
  int ia;
  for (ia = 0; ia < numAtoms; ia++) {
    if (atoms[ia].charge != 0) { return true; }
  }
  return false;
}

int molSMILES::parseChain(int pend, int nodBranch) {     // --- parseChain
   /*
      'chain' is either a whole SMILES string or a part enclosed in
     parentheses.  This function is called recursively to resolve the
     SMILES string.
      nodBranch is the note(atom) number from which the current chain
     is branched, and is -1 for the main chain. nodBnext is the node
     number of the next chain (if exist).
   */
  int pclose, nodBnext = nodBranch;
  bondInfo bnd;
  atomSMILES atm;
   // - loop bond-atom parsing
  while (cpPars < pend) {
    if (sPars[cpPars] == '(') {
      if (nodBnext == nodBranch) { return ERRE; }
      if ((pclose = findEndPrnt(cpPars, pend)) < 0) { return ERRE; }
      cpPars++;
      if (parseChain(pclose, nodBnext) == ERRE) { return ERRE; }
      cpPars = pclose + 1;
    } else {
      if (nodBnext >= 0) { findBond(pend, bnd); }
      if (findAtom(pend, atm) == ERRE) { return ERRE; }
      if (nodBnext >= 0) {
        bnd.toAtomIndex = numAtoms;
         atoms[nodBnext].bond[atoms[nodBnext].numBonds] = bnd;
         atoms[nodBnext].numBonds++;
        bnd.toAtomIndex = nodBnext; atm.bond[atm.numBonds] = bnd;
         atm.numBonds++;
      }
      atoms.push_back(atm); nodBnext = numAtoms; numAtoms++;
    }
  }
  if (nodBnext == nodBranch) { return ERRE; }
  return NORM;
}

int molSMILES::findEndPrnt(int pstart, int pend) {      // --- findEndPrnt
  int pos, level = 1;
  for (pos = pstart + 1; pos < pend; pos++) {
    if (sPars[pos] == ')') { level--; if (level == 0) { return pos; } }
    else if (sPars[pos] == '(') { level++; }
  }
  return -1;
}

void molSMILES::findBond(int pend, bondInfo &bnd) {        // --- findBond
  int idxf, ord;
  bnd.clear(); bnd.order = 1; bnd.isCommutable = false;
  if ((idxf = chBonds.find(sPars[cpPars])) == string::npos) { return; }
  ord = idxf - 2;
  bnd.order = abs(ord);
  if (ord < 0) { bnd.isCommutable = true; }
  cpPars++;
  return;
}

int molSMILES::findAtom(int pend, atomSMILES &atm) {         // --- findAtom
  int idxf, pket, rcfound;
  string sym;
  atm.clear();
  if (cpPars >= pend) { ERRE; }
  if (sPars[cpPars] == '[') {
    if ((pket = sPars.find(']', cpPars + 1)) != string::npos) {
      cpPars++;
      if (parseBracketAtom(pket, atm) == ERRE) { return ERRE; }
      cpPars = pket + 1;
    } else { return ERRE; }
  } else if ((idxf = chOrgSb1st.find(sPars[cpPars])) != string::npos) {
    sym = sPars[cpPars];
    if (str2cOrgSb.find(sPars.substr(cpPars, 2)) != str2cOrgSb.end())
      { sym = sPars.substr(cpPars, 2); }
    if (islower(sym[0])) { atm.numDeloElc = 1; }
    atm.atomicNumber = elmt.atmcNbrBySymb(sym);
    cpPars += sym.length();
  } else { return ERRE; }
  if (cpPars >= pend) { return NORM; }
   // - ring closures
  do {
    rcfound = false;
    if (isdigit(sPars[cpPars])) {
      rcfound = true;
      atm.ringIdcs.insert(atoi(sPars.substr(cpPars, 1).c_str()));
      cpPars++;
    } else if (sPars[cpPars] == '%') {
      if ((cpPars + 2) >= pend) { return ERRE; }
      if (isdigit(sPars[cpPars + 1]) && isdigit(sPars[cpPars + 2])) {
        rcfound = true;
        atm.ringIdcs.insert(atoi(sPars.substr(cpPars + 1, 2).c_str()));
        cpPars += 3;
      } else { return ERRE; }
    }
  } while (rcfound);
  return NORM;
}

                                                   // --- parseBracketAtom
int molSMILES::parseBracketAtom(int pend, atomSMILES &atm) {
  int sgnCharge;
  char symchrg;
  string sym, snH, sChrg, snCh, sRad, snRd;
  atm.clear(); atm.bracket = true;
  if ((pend - cpPars) < 1) { return ERRE; }
  if (!isalpha(sPars[cpPars])) { return ERRE; }
  sym = sPars[cpPars];
  if ((cpPars + 1) < pend) {
    if (isalpha(sPars[cpPars + 1]) && sPars[cpPars + 1] != 'H')
      { sym += sPars[cpPars + 1]; }
  }
  if (islower(sym[0])) { atm.numDeloElc = 1; }
  if ((atm.atomicNumber = elmt.atmcNbrBySymb(sym)) < 0) { return ERRE; }
  cpPars += sym.length();
  if (cpPars >= pend) { return NORM; }
  if (sPars[cpPars] == 'H') {
    cpPars++; snH = "";
    while (isdigit(sPars[cpPars])) { snH += sPars[cpPars]; cpPars++; }
    if (snH == "") { atm.numHexp = 1; }
    else { atm.numHexp = atoi(snH.c_str()); }
  } else if (sPars[cpPars] == '.') {
    sRad = "."; snRd = ""; cpPars++;
    while (sPars[cpPars] == '.') { sRad += sPars[cpPars]; cpPars++; }
    while (isdigit(sPars[cpPars])) { snRd += sPars[cpPars]; cpPars++; }
    if ((sRad.length() > 1) && (snRd.length() > 0)) { return ERRE; }
    if (snRd.length() > 0) { atm.numRadcElc = atoi(snRd.c_str()); }
    else { atm.numRadcElc = sRad.length(); }
  }
  if (cpPars >= pend) { return NORM; }
  if ((sPars[cpPars] == '+') || (sPars[cpPars] == '-')) {
    symchrg = sPars[cpPars]; sChrg = symchrg; snCh = "";
    sgnCharge = atoi((sChrg + "1").c_str());
    cpPars++; 
    while (sPars[cpPars] == symchrg) { sChrg += sPars[cpPars]; cpPars++; }
    while (isdigit(sPars[cpPars])) { snCh += sPars[cpPars]; cpPars++; }
    if ((sChrg.length() > 1) && (snCh.length() > 0)) { return ERRE; }
    if (snCh.length() > 0) { atm.charge = sgnCharge * atoi(snCh.c_str()); }
    else { atm.charge = sgnCharge * sChrg.length(); }
  }
  if (cpPars < pend) { return ERRE; }
  return NORM;
}

int molSMILES::resolvRingClosure() {              // --- resolvRingClosure
  int ia, nsame, ipair, idatm0, idatm1;
  vector<int> ida;
  map<int, vector<int> > cnnt;
  map<int, vector<int> >::iterator q;
  set<int>::iterator p;
  bondInfo bnd;
  cnnt.clear();
  for (ia = 0; ia < numAtoms; ia++) {
    for (p = atoms[ia].ringIdcs.begin(); p != atoms[ia].ringIdcs.end();
     p++) { cnnt[*p].push_back(ia); }
    atoms[ia].ringIdcs.clear();
  }
  numRings = 0;
  for (q = cnnt.begin(); q != cnnt.end(); q++) {
    if ((nsame = q->second.size()) < 2) { return ERRE; }
    if ((nsame % 2) == 1) { return ERRE; }
    ida = q->second; bnd.order = 1;
    for (ipair = 0; ipair < (nsame / 2); ipair++) {
      idatm0 = ida[ipair * 2]; idatm1 = ida[ipair * 2 + 1];
      bnd.toAtomIndex = idatm1;
       atoms[idatm0].bond[atoms[idatm0].numBonds] = bnd;
       atoms[idatm0].numBonds++;
      bnd.toAtomIndex = idatm0;
       atoms[idatm1].bond[atoms[idatm1].numBonds] = bnd;
       atoms[idatm1].numBonds++;
      numRings++;
    }
  }
  return NORM;
}

int molSMILES::resolvBrcktExp() {                    // --- resolvBrcktExp
  int ia;
   // correct numDeloElc in atoms
  for (ia = 0; ia < numAtoms; ia++) { atoms[ia].corrDeloc(); }
   // resolve the bracket explicits (number of H atoms or radical
   //  electrons)
  for (ia = 0; ia < numAtoms; ia++) {
    if (atoms[ia].bracket) {
      if (atoms[ia].numRadcElc == 0) {
        if (atoms[ia].numHexp < atoms[ia].numHatoms()) {
          atoms[ia].numRadcElc = atoms[ia].numHatoms()
                               - atoms[ia].numHexp;
        } else if (atoms[ia].numHexp > atoms[ia].numHatoms()) { return ERRE; }
      } else {
        if (atoms[ia].numHatoms() < 0) { return ERRE; }
      }
    } else {
      if (atoms[ia].numHatoms() < 0) { return ERRE; }
    }
  }
  return NORM;
}

void molSMILES::resolvDeloElc() {                     // --- resolvDeloElc
  int ia, ib, iaTo, ibBack;
  for (ia = 0; ia < numAtoms; ia++) {
    if (atoms[ia].numDeloElc > 0) {
      for (ib = 0; ib < atoms[ia].numBonds; ib++) {
        iaTo = atoms[ia].bond[ib].toAtomIndex;
        if (atoms[iaTo].numDeloElc > 0) {
          ibBack = atoms[iaTo].idxBondTo(ia);
          atoms[ia].bond[ib].order++; atoms[ia].numDeloElc = 0;
          atoms[iaTo].bond[ibBack].order++; atoms[iaTo].numDeloElc = 0;
          break;
        }
      }
    }
  }
  for (ia = 0; ia < numAtoms; ia++) {
    // unresolved electrons must be radical electrons
    if (atoms[ia].numDeloElc > 0)
      { atoms[ia].numRadcElc++; atoms[ia].numDeloElc = 0; }
    // some remaining tasks
    atoms[ia].atomAttrib = atoms[ia].numRadcElc;
    for (ib = 0; ib < atoms[ia].numBonds; ib++) {
      iaTo = atoms[ia].bond[ib].toAtomIndex;
      atoms[ia].bond[ib].toAtomicNbr = atoms[iaTo].atomicNumber;
    }
  }
}

ostream &molSMILES::print(ostream &os) {                      // --- print
  int ia;
  os << "atomSMILES{\n";
  os << " nA=" << numAtoms << " nR=" << numRings << endl;
  for (ia = 0; ia < numAtoms; ia++)
    { os << " atom[" << ia << "] " << atoms[ia]; }
  os << "}\n";
  return os;
}

ostream &operator<<(ostream &os, molSMILES &ms) { return ms.print(os); }

/* =========================================================================
     exSMILESmol and related classes
========================================================================= */

exSMLSatom::exSMLSatom() { clear(); }                   // --- constructor

void exSMLSatom::clear() {                                    // --- clear
  atomicNumber = 0; atomAttrib = 0; bondOrder = 0;
  numHatoms = 0; cycIndex = 0; idBranch.clear();
}

exSMLSchain::exSMLSchain() {                            // --- constructor
  clear();
  branchedFrom = -1; branchPos = -1;
}

exSMLSmol::exSMLSmol() {                                // --- constructor
  clear(); strSMILES = "";
  chBonds = "=#*";
  anOrgSb.clear();
  // SMILES's organic subset is B,C,N,O,F, P,S,Cl, Br, I
  anOrgSb.insert(5); anOrgSb.insert(6); anOrgSb.insert(7);
  anOrgSb.insert(8); anOrgSb.insert(9);
  anOrgSb.insert(15); anOrgSb.insert(16); anOrgSb.insert(17);
  anOrgSb.insert(35); anOrgSb.insert(53);
}

int exSMLSmol::setFrom(molecInfo &mi, molecByChains &mc) {  // --- setFrom
  int ia, na, idxa, idxaLeft, idxb, isCyclic, ich, bpos = 0, borgch, boffs;
  exSMLSatom atom;
  exSMLSchain chain;
  if (mc.numChains < 1) { return ERRE; }
  // set 1st chain
  chain.clear();
  na = mc.chain[0].numAtoms; isCyclic = mc.chain[0].isCyclic;
  idxaLeft = -1;
  for (ia = 0; ia < na; ia++) {
    idxa = mc.chain[0].atomIndex[ia];
    atom.clear();
    atom.atomicNumber = mi.atom[idxa].atomicNumber;
    atom.atomAttrib = mi.atom[idxa].atomAttrib;
    if (ia == 0) {
      if (isCyclic) { atom.bondOrder = 1; atom.cycIndex = 1; }
      else { atom.bondOrder = 0; }
    } else {
      idxb = mi.atom[idxa].idxBondTo(idxaLeft);
      atom.bondOrder = mi.atom[idxa].bond[idxb].order;
      if ((ia == na - 1) && isCyclic) { atom.cycIndex = 1; }
    }
    atom.numHatoms = mi.atom[idxa].getNumHatoms();
    chain.push_back(atom);
    idxaLeft = idxa;
  }
  push_back(chain);
  // set 2nd and later chains
  for (ich = 1; ich < mc.numChains; ich++) {
    borgch = mc.getChainOrigin(ich, bpos);
    boffs = (*this)[borgch].branchPos + 1;
    (*this)[borgch][bpos - boffs].idBranch.push_back(ich);
    chain.clear();
    na = mc.chain[ich].numAtoms; idxaLeft = mc.chain[ich].atomIndex[bpos];
    for (ia = bpos + 1; ia < na; ia++) {
      idxa = mc.chain[ich].atomIndex[ia];
      atom.clear();
      atom.atomicNumber = mi.atom[idxa].atomicNumber;
      atom.atomAttrib = mi.atom[idxa].atomAttrib;
      idxb = mi.atom[idxa].idxBondTo(idxaLeft);
      atom.bondOrder = mi.atom[idxa].bond[idxb].order;
      atom.numHatoms = mi.atom[idxa].getNumHatoms();
      chain.push_back(atom);
      idxaLeft = idxa;
    }
    chain.branchedFrom = borgch; chain.branchPos = bpos;
    push_back(chain);
  }
  return NORM;
}

string exSMLSmol::getSMILES() {                           // --- getSMILES
  strSMILES = "";
  if (size() < 1) { return strSMILES; }
  strSMILES = chainSMLS(0);
  return strSMILES;
}

string exSMLSmol::chainSMLS(int cch) {                    // --- chainSMLS
  string rs = "", asym, ssubc;
  int ia, na = (*this)[cch].size(), an, aa, bo, nH, cI, nB, iB, idbc;
  for (ia = 0; ia < na; ia++) {
    an = (*this)[cch][ia].atomicNumber;
    asym = elmt.canonSymbol(an);
    aa = (*this)[cch][ia].atomAttrib;
    bo = (*this)[cch][ia].bondOrder;
    cI = (*this)[cch][ia].cycIndex;
    nB = (*this)[cch][ia].idBranch.size();
    if (bo > 1) { rs += chBonds[bo - 2]; }
    if ((anOrgSb.find(an) == anOrgSb.end()) || (aa != 0)) {
      // non-organic subset or radical
      nH = (*this)[cch][ia].numHatoms;
      rs += "[" + asym + Hstr(nH) + "]";
    } else {
      // organic subset with normal valence
      rs += asym;
    }
    if (cI > 0) { rs += format(cI); }
    for (iB = 0; iB < nB; iB++) {
      idbc = (*this)[cch][ia].idBranch[iB];
      ssubc = chainSMLS(idbc);
      rs += "(" + ssubc + ")";
    }
  }
  return rs;
}

string exSMLSmol::Hstr(int nH) {                               // --- Hstr
  string rs = "";
  if (nH < 1) { return rs; }
  rs += "H";
  if (nH > 1) { rs += format(nH); }
  return rs;
}

/*============================================================================
     'typeMolec' class
============================================================================*/

typeMolec::typeMolec() {                                // --- constructor
  libPath libp;
  clear();
  if (loadFrom(FL_TYPMOL, libp) == ERRE)
    { cout << "Error reading type library.\n"; }
}

void typeMolec::clear() { mCodStr.clear(); mStrCod.clear(); } // --- clear

int typeMolec::loadFrom(string flname, libPath &libp) {    // --- loadFrom
  int endTable = false;
  long tCod;
  string s, tNam;

  if (!mCodStr.empty()) { mCodStr.clear(); }
  if (!mStrCod.empty()) { mStrCod.clear(); }
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) {
    cout << "Error reading " << flname << ".\n";
    ifs.close(); return ERRE;
  }
  if (s != "typeMolec{") {
    cout << "Error reading " << flname << ".\n";
    ifs.close(); return ERRE;
  }
  do {
    if (getTokenFrom(ifs, tNam) == ERRE) {
      cout << "Error reading " << flname << ".\n";
      ifs.close(); return ERRE;
    }
    if (tNam == "}") { endTable = true; }
    else {
      if (getTokenFrom(ifs, s) == ERRE) {
        cout << "Error reading " << flname << ".\n";
        ifs.close(); return ERRE;
      }
      tCod = valHexStr(s);
      mCodStr.insert(make_pair(tCod, tNam));
      mStrCod.insert(make_pair(tNam, tCod));
    }
  } while (!endTable);
  ifs.close();
  cout << mCodStr.size() << " types loaded from [" << flname << "]\n";
  return NORM;
}

// ------------------------------------------------------------- [typName]
string typeMolec::typName(long tcod) {
  if (mCodStr.find(tcod) == mCodStr.end()) { return ""; }
  return mCodStr[tcod];
}

string typeMolec::typNameExt(long tcod) {                // --- typNameExt
  // - extended type name for cyclic species
  static const long bitCyc = 1L << 27, mskCyc = ~bitCyc;
  if (mCodStr.find(tcod) != mCodStr.end()) { return mCodStr[tcod]; }
  if (mCodStr.find(tcod & mskCyc) != mCodStr.end())
    { return mCodStr[tcod & mskCyc] + ".cyc"; }
  return "";
}

// ------------------------------------------------------------- [typCode]
long typeMolec::typCode(string tnam) {
  if (mStrCod.find(tnam) == mStrCod.end()) { return 0L; }
  return mStrCod[tnam];
}

/*============================================================================
     'molecule' class
============================================================================*/

molecule::molecule() { clear(); }                       // --- constructor

// ---------------------------------------------- [setFrom(alkaneChain &)]
int molecule::setFrom(alkaneChain &alc) {
  int nAtoms, nBonds, toIdx, pci, ppc, ich;
  vector<vector<int> > atomIdx;

  molName = alc.alkaneName;
  isUnique = false; isCyclic = false; numEquivChain = 1;
  isAromatic = false; numResStruct = 0; isRxnDone = false;
  if (alc.numChains < 1) { return ERRE; }
  if ((alc.isCyclic) && (alc.chain[0].lenChain < 3)) { return ERRE; }
  nAtoms = 0;
  for (ich = 0; ich < alc.numChains; ich++)
    { nAtoms += alc.chain[ich].lenChain; }
  molec.atom.resize(nAtoms);
  nAtoms = 0; atomIdx.clear(); atomIdx.resize(alc.numChains);
  for (ich = 0; ich < alc.numChains; ich++) {
    if (ich > 0) {
      pci = alc.chain[ich].parentChIdx;
      ppc = alc.chain[ich].posSubInPCh - 1;
      toIdx = atomIdx[pci][ppc];
      nBonds = molec.atom[toIdx].numBonds;
      molec.atom[toIdx].bond[nBonds].order = 1;
      molec.atom[toIdx].bond[nBonds].toAtomicNbr = 6;
      molec.atom[toIdx].bond[nBonds].toAtomIndex = nAtoms;
      molec.atom[toIdx].numBonds++;
    }
    atomIdx[ich].resize(alc.chain[ich].lenChain);
    for (int iat = 0; iat < alc.chain[ich].lenChain; iat++) {
      molec.atom[nAtoms].atomicNumber = 6;
      molec.atom[nAtoms].atomAttrib = 0;
      nBonds = 0;
      if (iat == (alc.chain[ich].lenChain - 1)) {
        if ((ich == 0) && (alc.isCyclic)) {
          molec.atom[nAtoms].bond[nBonds].toAtomIndex = atomIdx[0][0];
          nBonds++;
        }
      } else {
        molec.atom[nAtoms].bond[nBonds].toAtomIndex = nAtoms + 1;
        nBonds++;
      }
      if (iat == 0) {
        if (ich == 0) {
          if (alc.isCyclic) {
            molec.atom[nAtoms].bond[nBonds].toAtomIndex
             = nAtoms + alc.chain[ich].lenChain - 1;
            nBonds++;
          }
        } else {
          molec.atom[nAtoms].bond[nBonds].toAtomIndex = toIdx;
          nBonds++;
        }
      } else {
        molec.atom[nAtoms].bond[nBonds].toAtomIndex = nAtoms - 1;
        nBonds++;
      }
      molec.atom[nAtoms].numBonds = nBonds;
      for (int ib = 0; ib < nBonds; ib++) {
          molec.atom[nAtoms].bond[ib].order = 1;
          molec.atom[nAtoms].bond[ib].toAtomicNbr = 6;
      }
      atomIdx[ich][iat] = nAtoms;
      nAtoms++;
    }
  }
  molec.numAtoms = nAtoms;
  string chForm = molec.getChemFormula();
  molName += chForm;
  return NORM;
}

// ---------------------------------------- [setFrom(molecInfo &, string)]
void molecule::setFrom(molecInfo &mi, string nam) {
  molName = nam; molec = mi; resavr.clear(); uniqueRepr.clear();
  isUnique = false;
}

                                                     // --- setFromPsmiles
int molecule::setFromPsmiles(string psm, groupLib &gl, typeMolec &tym) {
  // set from string with 'prefix' and SMILES
  int pclose;
  string pref = "", sm;
  molSMILES mS;
  clear();
  if (psm[0] == '\'') {
    if ((pclose = psm.find('\'', 1)) == string::npos) { return ERRE; }
    pref = psm.substr(1, pclose - 1);
    sm = psm.substr(pclose + 1);
  } else { sm = psm; }
  if (mS.setMol(sm, molec) == ERRE) { return ERRE; }
  string tchForm = getTypeFormula(gl, tym);
  molName = pref + tchForm;
  return NORM;
}

int molecule::setFromSMILES(string sm, string nam) {  // --- setFromSMILES
  // set from SMILES string and name
  molSMILES mS;
  clear();
  if (mS.setMol(sm, molec) == ERRE) { return ERRE; }
  molName = nam;
  return NORM;
}

string molecule::getSMILES() {                            // --- getSMILES
  string rs = "(error)";
  exSMLSmol esm;
  if (!isUnique) { beUnique(); }
  if (esm.setFrom(molec, uniqueRepr) == ERRE) { return rs; }
  rs = esm.getSMILES();
  return rs;
}

// --------------------------------------------------------------- [clear]
void molecule::clear() {
  molName.erase(); molec.clear(); resavr.clear(); uniqueRepr.clear();
  isUnique = false; isCyclic = false; numEquivChain = 0;
  isAromatic = false; numResStruct = 0; isRxnDone = false;
}

void molecule::print(ostream *os) {   // ------------------------- [print]
  *os << "molName = " << molName << "\n";
  *os << molec;
  *os << "unique flag : " << isUnique << endl;
  if (isUnique) { *os << uniqueRepr; }
}

void molecule::beUnique() {   // ------------------------------ [beUnique]
  int numReson, nResLast, flgCyc, nEqCh, flgArm, fC, nEC, fA, isNew;
  int idxRC, findRCout, numRCout, si, idxMax, needSwap;
  int ir, irc, jr;
  vector<int> isRCinCyc, sortIdx;
  vector<string> rse;
  molecInfo rss;
  molecByChains rcc;
  vector<molecInfo> rs;
  vector<molecByChains> rc;

  rs.clear(); rc.clear(); numReson = molec.getResonStr(rs);
  rc.resize(numReson); nResLast = numReson;
  for (ir = 0; ir < nResLast; ir++) {
    rs[ir].beUnique(rc[ir], flgCyc, nEqCh, flgArm);
    if (flgArm) {
      rss = rs[ir]; rss.beAromaOther(rc[ir].chain[0]);
      rcc.clear(); rss.beUnique(rcc, fC, nEC, fA);
      isNew = true;
      for (irc = 0; irc < numReson; irc++) {
        if (rss == rs[irc]) { isNew = false; break; }
      }
      if (isNew) {
        rs.push_back(rss); rc.push_back(rcc); numReson++;
      }
    }
  }
  // radical center out of the ring is preferable for unique structure
  findRCout = false;
  if (flgCyc && (molec.getNumRadCent() == 1)) {
    numRCout = 0; isRCinCyc.resize(numReson);
    for (ir = 0; ir < numReson; ir++) {
      idxRC = rs[ir].getIdxRadCent();
      if ( (rc[ir].chain[0].isIdxInChain(idxRC)) ||
           (rs[ir].atom[idxRC].atomicNumber != 6) )
        { isRCinCyc[ir] = true; }
      else { isRCinCyc[ir] = false; numRCout++; }
    }
    if (numRCout > 0) { findRCout = true; }
  }
  // sorting
  sortIdx.resize(numReson); rse.resize(numReson);
  for (ir = 0; ir < numReson; ir++)
    { sortIdx[ir] = ir; rse[ir] = rs[ir].getStrRSE(); }
  for (ir = 0; ir < (numReson - 1); ir++) {
    for (jr = (ir + 1); jr < numReson; jr++) {
      needSwap = false;
      if (rse[sortIdx[jr]] > rse[sortIdx[ir]]) { needSwap = true; }
      else if (rse[sortIdx[jr]] == rse[sortIdx[ir]]) {
        if (rc[sortIdx[jr]] > rc[sortIdx[ir]]) { needSwap = true; }
      }
      if (needSwap)
        { si = sortIdx[ir]; sortIdx[ir] = sortIdx[jr]; sortIdx[jr] = si; }
    }
  }
  idxMax = sortIdx[0];
  if (findRCout) {
    for (ir = 0; ir < numReson; ir++) {
      if (!isRCinCyc[sortIdx[ir]]) { idxMax = sortIdx[ir]; break; }
    }
  }
  molec = rs[idxMax];
  rs.clear(); rc.clear();
  molec.beUnique(uniqueRepr, isCyclic, numEquivChain, isAromatic);
  // prepare resonant average structure
  numReson = molec.getResonStr(rs);
  if (isAromatic) {
    rss = molec;
    rss.beAromaOther(uniqueRepr.chain[0]);
    isNew = true;
    for (ir = 0; ir < numReson; ir++)
      { if (rss == rs[ir]) { isNew = false; } }
    if (isNew) { rs.push_back(rss); numReson++; }
  }
  numResStruct = numReson;
  if (numReson > 1) {
    resavr.beResonAvr(numReson, rs);
  }
  if (isAromatic) {
    if (uniqueRepr.chain[0].numAtoms != 6) { isAromatic = false; }
  }
  isUnique = true;
}

void molecule::getUniqueView(vector<string> &vs) {   // -- [getUniqueView]
  if (!vs.empty()) { vs.clear(); }
  if (!isUnique) { return; }
  uniqueRepr.getStrucView(vs);
}

int molecule::storeTo(ofstream &ofs) {   // -------------------- [storeTo]
  string s;
  s = "!----- " + molec.getChemFormula() + " [" + molName + "] ";
  ofs << s << setw(70 - s.length()) << setfill('-') << "-"
      << endl << endl;
  ofs << "molecule{ " << "[" << molName << "] " << isUnique << " "
      << isCyclic << " " << numEquivChain << " " << isAromatic << " "
      << numResStruct << " " << isRxnDone << endl;
  if (molec.storeTo(ofs) == ERRE) { return ERRE; }
  if (isUnique) {
    if (uniqueRepr.storeTo(ofs) == ERRE) { return ERRE; }
  }
  if (numResStruct > 1) {
    if (resavr.storeTo(ofs) == ERRE) { return ERRE; }
  }
  ofs << "}\n\n";
  return NORM;
}

                                                         // --- storeLogTo
void molecule::storeLogTo(ostream &os, groupLib &gl, typeMolec tm) {
  long mtyp;
  string chemForm, typ;
  vector<string> vs;
  getUniqueView(vs);
  chemForm = molec.getChemFormula();
  os << "-----------------------------------------------------------\n";
  os << " name:    " << molName << endl;
  os << " formula: " << chemForm << endl;
  os << " SMILES:  " << getSMILES() << endl;
  os << " struct:\n";
  os << vs;
  os << " groups: " << getThermInput(gl) << endl;
  mtyp = getType(gl); typ = tm.typNameExt(mtyp);
  os << " code:    " << hex << mtyp << dec;
  if (typ != "") { os << " (" << typ << ")"; }
  os << endl;
}

// --------------------------------------------------------- [restoreFrom]
int molecule::restoreFrom(ifstream &ifs) {
  int l;
  string s;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "molecule{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  l = s.length();
  if (l < 2) { return ERRE; }
  if ((s[0] != '[') || (s[l - 1] != ']')) { return ERRE; }
  molName = s.substr(1, l - 2);
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  isUnique = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  isCyclic = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numEquivChain = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  isAromatic = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numResStruct = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  isRxnDone = atoi(s.c_str());
  if (molec.restoreFrom(ifs) == ERRE) { return ERRE; }
  if (isUnique) {
    if (uniqueRepr.restoreFrom(ifs, molec) == ERRE) { return ERRE; }
  }
  if (numResStruct > 1) {
    if (resavr.restoreFrom(ifs) == ERRE) { return ERRE; }
  }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  if (molec.checkParam() == ERRE) { return ERRE; }
  return NORM;
}

// -------------------------------------------------------- [getGroupAtom]
void molecule::getGroupAtom(int idxAtom, groupAtom &ga) {
  atomChain arc;
  if (isAromatic) {
    arc = uniqueRepr.chain[0];
    molec.getGroupAroma(idxAtom, arc, ga);
  }
  else { molec.getGroupNonAroma(idxAtom, ga); }
}

// ------------------------------------------------------ [getGroupCyclic]
void molecule::getGroupCyclic(groupCyclic &gc) {
  int nAtom, ia, iaNext, idxAtom, idxNext, nBond, ib;
  atomChain ac;
  gc.numAtoms = 0;
  if (!isCyclic) { return; }
  ac = uniqueRepr.chain[0];
  gc.numAtoms = (nAtom = ac.numAtoms); gc.atom.resize(nAtom);
  for (ia = 0; ia < nAtom; ia++) {
    iaNext = ia + 1;
    if (iaNext >= nAtom) { iaNext -= nAtom; }
    idxAtom = ac.atomIndex[ia]; idxNext = ac.atomIndex[iaNext];
    gc.atom[ia].atomicNumber = molec.atom[idxAtom].atomicNumber;
    gc.atom[ia].atomAttrib = 0;
    nBond = molec.atom[idxAtom].numBonds;
    for (ib = 0; ib < nBond; ib++) {
      if (molec.atom[idxAtom].bond[ib].toAtomIndex == idxNext) {
        gc.atom[ia].bondOrdToNext = molec.atom[idxAtom].bond[ib].order;
      }
    }
  }
  gc.beUnique();
}

// ---------------------------------------------------- [getMolecByGroups]
void molecule::getMolecByGroups(molecByGroups &mg, groupLib &grp) {
  int nAtom = molec.numAtoms, ia;
  string gnam;
  groupCyclic gc; groupAtom ga;

  mg.clear();
  if (isCyclic && (!isAromatic))
    { getGroupCyclic(gc); mg.cycGrpName = grp.cyc.findName(gc); }
  for (ia = 0; ia < nAtom; ia++) {
    getGroupAtom(ia, ga);
    if (ga.isRadical())
      { mg.radGrpName = grp.rad.findName(ga); ga.beNonRadical(); }
    gnam = grp.main.findName(ga);
    if (gnam != "none") { mg.atmGrps[gnam]++; }
  }
}

                                                 // --- getIdxAtomGrpMatch
int molecule::getIdxAtomGrpMatch(string gs, groupLib &grp) {
  int idx = -1, ia, na = molec.numAtoms;
  string gnam;
  groupAtom ga;

  for (ia = 0; ia < na; ia++) {
    getGroupAtom(ia, ga);
    if (ga.isRadical()) { gnam = grp.rad.findName(ga); }
    else { gnam = grp.main.findName(ga); }
    if (gnam == gs) { idx = ia; break; }
  }
  return idx;
}

// ------------------------------------------------------- [getNumSymIsom]
//   get informations on rotational symmetry number, number of isomers,
//    and number of internal rotors of a molecule

int molecule::getNumSymIsom(int &numIsom, int &numIntRot) {
  if (!isUnique) {
    cout << "ERROR molecule::getNumSymIsom [not uniqu]\n";
    return 0;
  }
  if (isCyclic) {
    return molec.getSymIsomCyclic(uniqueRepr.chain[0], numIsom,
     numIntRot);
  } else {
    return molec.getSymIsomNonCyc(uniqueRepr.chain[0], numEquivChain,
     numIsom, numIntRot);
  }
}

// ------------------------------------------------------- [getThermInput]
//   get thermgen type group output string

string molecule::getThermInput(groupLib &grp) {
  int numSym, numIsom, numIntRot, isRadical = false;
  string rets;
  atmcCompSorted acs;
  atmcCompSorted::iterator p;
  molecByGroups molGrp;
  map<string, int>::iterator q;

  molec.getAtomCompSorted(acs);
  getMolecByGroups(molGrp, grp);
  numSym = getNumSymIsom(numIsom, numIntRot);
  if (molGrp.radGrpName != "") { isRadical = true; }
  rets = " " + molName;
  if (isRadical) { rets += " R "; } else { rets += " M "; }
  for (p = acs.begin(); p != acs.end(); p++) {
    rets += elmt[p->second.first].atomicSymbol + " "
         + format(p->second.second) + " ";
  }
  rets += "$ ";
  if (molGrp.cycGrpName != "") { rets += molGrp.cycGrpName + " 1 "; }
  for (q = molGrp.atmGrps.begin(); q != molGrp.atmGrps.end(); q++)
    { rets += q->first + " " + format(q->second) + " "; }
  rets += "$ ";
  if (isRadical) { rets += molGrp.radGrpName + " "; }
  rets += format(numIntRot) + " " + format(numSym);
  // --- disabled for consistency with THERM
  // if (numIsom > 1) { rets += " $ " + format(numIsom); }
  return rets;
}

// ------------------------------------------------------------- [getType]
//   Get type of molecule. Returns 0 for unsupported molecule.
//    May returns type currently not supported.

long molecule::getType(groupLib &gl) {
  /* ---------------------------------------------------------------------
   return code:
    3322222222221111111111
    10987654321098765432109876543210
    <______><______><______><______> (32 bits)
    cnof       -- constituent elements (C/N/O/F)
        ca     -- Cyclic/Aromatic
          dat          -- C bond info. (double/allene/triple)
             kakaeaehhp   -- O bond info. (ketene/aldehyde/ketone/acid/
                              ester/alcohol/ether/hydroperoxide(2bit)/
                              peroxide)
                       icnnha   -- N bond info. (imino/cyano/
                                    nitroso/nitro/hydrazino/azo)
                             rr     -- radical center (01-C/10-N/11-O)
                               riiii   -- radical info.
   riiii : 000xx - at NB center mono(00)/di(01)/tri(10) radical
         : 001xx - at SB center to C(00)/N(01)/O(10)/F(11)
         : 010xx - at DB center to C(00)/N(01)/O(10)/birad-O(11)
         : 011xx - at TB center to C(00)/N(01)
         : 1ddtt - reson stab by near-by unsat bonds
            (dd/tt - number of double/triple bonds)
         : 11111 - by benzene
  --------------------------------------------------------------------- */
  static const long   bitF   = 1L << 28, bitCyc = 1L << 27,
   bitArm = 1L << 26, bitCD  = 1L << 25, bitCA  = 1L << 24,
   bitCT  = 1L << 23, bitKte = 1L << 22, bitAld = 1L << 21;
  static const long   bitKto = 1L << 20, bitAcd = 1L << 19,
   bitEst = 1L << 18, bitAlc = 1L << 17, bitEth = 1L << 16,
   bitHyp = 1L << 15, bitHy2 = 1L << 14, bitPer = 1L << 13;
  static const long   bitImi = 1L << 12, bitCya = 1L << 11,
   bitNO  = 1L << 10, bitNO2 = 1L <<  9, bitHrz = 1L <<  8,
   bitAzo = 1L <<  7, bitRSF = 1L <<  4, bytRBZ = 0xfL;
  static const int bshCAN = 5, bshRBO = 2, bshNDB = 2;
  long typ = 0L;
  int atmcNum, nRadCnt, idxRadCnt, ng, nOOH = 0, nOOR = 0;
  int nNeig, ing, nDB2nd, nTB2nd, nBZ2nd, bo, maxBO, toANmxBO;
  int toAN, uB2nd, sbMxToAN, rcAttrib;
  string sg;
  atmcCompSorted acs;
  atmcCompSorted::iterator pacs;
  molecByGroups mg;
  map<string, int>::iterator p;
  vector<string> vs;
  groupAtom rg;

  nRadCnt = molec.getNumRadCent();
  if ((nRadCnt > 1) || (nRadCnt < 0)) { return 0L; }
  molec.getAtomCompSorted(acs);
  for (pacs = acs.begin(); pacs != acs.end(); pacs++) {
    atmcNum = pacs->second.first;
    if (((atmcNum >= 6) && (atmcNum <= 9)) && (pacs->second.second > 0))
      { typ |= (bitF << (9 - atmcNum)); }
  }
  if (isCyclic) { typ |= bitCyc; }
  if (isAromatic) { typ |= bitArm; }
  getMolecByGroups(mg, gl);
  for (p = mg.atmGrps.begin(); p != mg.atmGrps.end(); p++) {
    splitToken((sg = p->first), "/", vs); ng = p->second;
    if (numMatchStrList("CD", &vs) > 0) { typ |= bitCD; }
    if (numMatchStrList("CA", &vs) > 0) { typ |= bitCA; }
    if (numMatchStrList("CT", &vs) > 0) { typ |= bitCT; }
    if (numMatchStrList("NI", &vs) > 0) { typ |= bitImi; }
    if (numMatchStrList("CN", &vs) > 0) { typ |= bitCya; }
    if (numMatchStrList("NO", &vs) > 0) { typ |= bitNO; }
    if (numMatchStrList("NO2", &vs) > 0) { typ |= bitNO2; }
    if (numMatchStrList("NA", &vs) > 0) { typ |= bitAzo; }
    if (sg == "KETENE") { typ |= bitKte; }
    else if (vs[0] == "CO") {
      if (numMatchStrList("O", &vs) == 0) {
        if ((numMatchStrList("H", &vs) > 0) ||
                 (numMatchStrList("H2", &vs) > 0)) { typ |= bitAld; }
        else { typ |= bitKto; }
      }
    } else if (vs[0] == "O") {
      if (numMatchStrList("CO", &vs) > 0) {
        if (numMatchStrList("H", &vs) > 0) { typ |= bitAcd; }
        else { typ |= bitEst; }
      } else if (numMatchStrList("O", &vs) > 1) {
        if (numMatchStrList("H", &vs) > 0) { nOOH += ng; typ |= bitHyp; }
        else { nOOR += ng; }
      } else if (numMatchStrList("H", &vs) > 0) { typ |= bitAlc; }
      else { typ |= bitEth; }
    } else if (vs[0] == "N") {
      if (numMatchStrList("N", &vs) > 1) { typ |= bitHrz; }
    }
  }
  if (nOOR > nOOH) { typ |= bitPer; }
  if (nOOH > 1) { typ |= bitHy2; }
  if (nRadCnt != 1) { return typ; }     // --- exit for non-radical

  idxRadCnt = molec.getIdxRadCent(); getGroupAtom(idxRadCnt, rg);
  atmcNum = rg.atomicNumber; nNeig = rg.numNeighbors;
  rcAttrib = molec.atom[idxRadCnt].atomAttrib;
  if ((atmcNum >= 6) && (atmcNum <= 8))
    { typ |= ((atmcNum - 5) << bshCAN); }
  else { typ |= 1L; return typ; }
  nDB2nd = 0; nTB2nd = 0; nBZ2nd = 0; maxBO = 0; sbMxToAN = 0; toANmxBO =0;
  for (ing = 0; ing < nNeig; ing++) {
    bo = rg.neigh[ing].bondOrder; toAN = rg.neigh[ing].toAtomicNbr;
    uB2nd = rg.neigh[ing].UnsBnd2nd;
    if (bo > maxBO) { maxBO = bo; toANmxBO = toAN; }
    if (bo == 1) { if (toAN > sbMxToAN) { sbMxToAN = toAN; } }
    if (uB2nd == 2) { nDB2nd++; }
    else if (uB2nd == 3) { nTB2nd++; }
    else if (uB2nd == 6) { nBZ2nd++; }
  }
  if ((nDB2nd + nTB2nd + nBZ2nd) > 0) {
    typ |= bitRSF;
    if (nBZ2nd > 0) { typ |= bytRBZ; }
    else { typ |= ((nDB2nd << bshNDB) | nTB2nd); }
    return typ;      // --- exit for resonant radical
  }
  typ |= (maxBO << bshRBO);
  if (maxBO == 1) { toANmxBO = sbMxToAN; }
  if ((toANmxBO >= 6) && (toANmxBO <= 9)) { typ |= (toANmxBO - 6); }
  if (maxBO == 0) {  // --- for CH3/CH2/CH
    if (rcAttrib > 1) { typ |= (rcAttrib - 1); }
  }
  if (maxBO == 2) {  // --- for CO
    if (rcAttrib == 2) { typ |= 1L; }
  }
  return typ;
}

// ------------------------------------------------------ [getTypeFormula]
//   Get type specific chemical formula
string molecule::getTypeFormula(groupLib &gl, typeMolec &tym) {
  static const long bitCyc = 1L << 27, mskCyc = ~bitCyc;
  long mtyp;
  string typ, typCm, head = "", tail = "", typForm;
  atmcComp ac;

  mtyp = getType(gl);
  typ = tym.typName(mtyp);
  typCm = tym.typName(mtyp & mskCyc);
  molec.getAtomComp(ac);
  if (typCm == "Alcohol" || typCm == "RdOH") { ac[1]--; ac[8]--; tail = "OH"; }
  else if (typ == "CycEther" || typ == "CycEthDB") { ac[8]--; tail = "cyO"; }
  else if (typ == "CycEthOH") { ac[1]--; ac[8] -= 2; tail = "OHcyO"; }
  else if (typ == "RCOCyEth") { ac[6]--; ac[8] -= 2; tail = "COcyO"; }
  else if (typ == "CyEthCHO") { ac[1]--; ac[6]--; ac[8] -= 2; tail = "CHOcyO"; }
  else if (typCm == "RCHO" || typCm == "CdQCHO") { ac[1]--; ac[6]--; ac[8]--; tail = "CHO"; }
  else if (typCm == "RCOR" || typCm == "CdQCOR") { ac[6]--; ac[8]--; tail = "CO"; }
  else if (typCm == "hoRCHO" || typCm == "hoCdQCHO") { ac[1] -= 2; ac[6]--; ac[8] -= 2; head = "HO"; tail = "CHO"; }
  else if (typCm == "hoRCOR" || typCm == "hoCdQCRO") { ac[1]--; ac[6]--; ac[8] -= 2; head = "HO"; tail = "CO"; }
  else if (typCm == "RCOQCHO") { ac[1]--; ac[6] -= 2; ac[8] -= 2; tail = "COCHO"; }
  else if (typCm == "OCCQCHO") { ac[1]--; ac[6] -= 3; ac[8] -= 2; head = "OCC"; tail = "CHO"; }
  else if (typCm == "RCOQCCO") { ac[6] -= 3; ac[8] -= 2; head = "OCC"; tail = "CO"; }
  else if (typCm == "RCCO" || typCm == "RdCCO") { ac[6] -= 2; ac[8]--; tail = "CCO"; }
  else if (typCm == "hoRCCO") { ac[1]--; ac[6] -= 2; ac[8] -= 2; head = "HO"; tail = "CCO"; }
  else if (typCm == "ROOH" || typCm == "RdOOH") { ac[1]--; ac[8] -= 2; tail = "OOH"; }
  else if (typCm == "hoROOH" || typCm == "hoRdOOH") { ac[1 ]-= 2; ac[8] -= 3; head = "HO"; tail = "OOH"; }
  else if (typCm == "AldOOH" || typCm == "AldDOOH") { ac[1] -= 2; ac[6]--; ac[8] -= 3; head = "CHO"; tail = "OOH"; }
  else if (typCm == "KetOOH" || typCm == "KetDOOH") { ac[1]--; ac[6]--; ac[8] -= 3; head = "CO"; tail = "OOH"; }
  else if (typCm == "hoAldOOH") { ac[1] -= 3; ac[6]--; ac[8] -= 4; head = "HOCHO"; tail = "OOH"; }
  else if (typCm == "hoKetOOH") { ac[1] -= 2; ac[6]--; ac[8] -= 4; head = "HOCO"; tail = "OOH"; }
  else if (typCm == "R" || typCm == "Alkenyl" || typCm =="Vinyl") { tail = ""; }
  else if (typCm == "Allyl" || typCm == "AllAllyl" || typCm == "Pntdnyl") { head = "a"; tail = ""; }
  else if (typCm == "hoR" || typCm == "hoAlkenyl" || typCm == "hoVinyl") { ac[1]--; ac[8]--; tail = "OH"; }
  else if (typCm == "hoRalp" || typCm == "hoRDalp") { ac[1]--; ac[8]--; tail = "aOH"; }
  else if (typCm == "hoAllyl") { ac[1]--; ac[8]--; head = "a"; tail = "OH"; }
  else if (typCm == "coAlkenyl" || typCm == "rcoVinyl") { ac[6]--; ac[8]--; head = "CO"; tail = ""; }
  else if (typCm == "hcoVinyl") { ac[1]--; ac[6]--; ac[8]--; head = "CHO"; tail = ""; }
  else if (typCm == "Ketyl") { ac[6] -= 2; ac[8]--; head = "CCO"; tail = ""; }
  else if (typCm == "HOOhoRalp") { ac[1] -= 2; ac[8] -= 3; head = "HOO"; tail = "aOH"; }
  else if (typCm == "ROO" || typCm == "RdOO") { ac[8] -= 2; tail = "OO"; }
  else if (typCm == "hoROO") { ac[1]--; ac[8] -= 3; head = "HO"; tail = "OO"; }
  else if (typCm == "RO" || typCm == "RdO") { ac[8]--; tail = "O"; }
  else if (typCm == "HOORO") { ac[1]--; ac[8] -= 3; head = "HOO"; tail = "O"; }
  else if (typCm == "QOOH" || typCm == "QdOOH" || typCm == "QaOOH" || typCm == "alpQOOH") { ac[1]--; ac[8] -= 2; tail = "OOH"; }
  else if (typCm == "hoQOOH") { ac[1] -= 2; ac[8] -= 3; head = "HO"; tail = "OOH"; }
  else if (typCm == "RCOPOOH") { ac[1]--; ac[6]--; ac[8] -= 3; head = "CO"; tail = "OOH"; }
  else if (typCm == "CHOPOOH") { ac[1] -= 2; ac[6]--; ac[8] -= 3; head = "CHO"; tail = "OOH"; }
  else if (typCm == "OOQOOH" || typCm == "OOQdOOH") { ac[1]--; ac[8] -= 4; head = "HOO"; tail = "OO"; }
  else if (typCm == "OOhoQOOH") { ac[1] -= 2; ac[8] -= 5; head = "HOOHO"; tail = "OO"; }
  else if (typCm == "RCO" || typCm == "CdQCO") { ac[6]--; ac[8]--; tail = "CO"; }
  else if (typCm == "hoRCO" || typCm == "hoCdQCO") { ac[1]--; ac[6] -= 2; ac[8]--; head = "HO"; tail = "CO"; }
  else if (typCm == "RCOQCO") { ac[6] -= 2; ac[8] -= 2; head = "CO"; tail = "CO"; }
  else if (typCm == "OCCQCO") { ac[6] -= 3; ac[8] -= 2; head = "OCC"; tail = "CO"; }
  else if (typCm == "OCCV") { ac[6] -= 2; ac[8]--; head = "OCC"; }
  else if (typCm == "hoOCCV") { ac[1]--; ac[6] -= 2; ac[8] -= 2; head = "OCC"; tail = "OH"; }
  else if (typCm == "RCOQ" || typCm == "RdCOQ") { ac[6]--; ac[8]--; head = "CO"; tail = ""; }
  else if (typCm == "HCOQ") { ac[1]--; ac[6]--; ac[8]--; head = "CHO"; tail = ""; }
  else if (typCm == "CHOQCOQ") { ac[1]--; ac[6] -= 2; ac[8] -= 2; head = "CHOCO"; tail = ""; }
  else if (typCm == "RCOV" || typCm == "CdQCOV") { ac[6]--; ac[8]--; tail = "vCO"; }
  else if (typCm == "HCOV" || typCm == "HCOVQCd") { ac[1]--; ac[6]--; ac[8]--; tail = "vCHO"; }
  else if (typCm == "hoRCOV" || typCm == "hoRdCOV") { ac[1]--; ac[6]--; ac[8] -= 2; head = "HO"; tail = "vCO"; }
  else if (typCm == "HCOVQOH") { ac[1] -= 2; ac[6]--; ac[8] -= 2; head = "HO"; tail = "vCHO"; }
  else if (typCm == "RCOVAlly") { ac[6]--; ac[8]--; head = "CO"; }
  else if (typCm == "HCOVAlly") { ac[1]--; ac[6]--; ac[8]--; head = "CHO"; }
  else if (typCm == "CHOQCOV") { ac[1]--; ac[6] -= 2; ac[8] -= 2; head = "CHO"; tail = "vCO"; }
  else if (typCm == "OCCQ") { ac[6] -= 2; ac[8]--; head = "OCC"; }
  else if (typCm == "hoQCHO") { ac[1] -= 2; ac[6]--; ac[8] -= 2; head = "HO"; tail = "CHO"; }
  else if (typCm == "hoQCRO") { ac[1]--; ac[6]--; ac[8] -= 2; head = "HO"; tail = "CO"; }
  else if (typCm == "RCO2") { ac[6]--; ac[8] -= 2; tail = "CO2"; }
  typForm = head + molec.atmCompToFormula(ac) + tail;
  return typForm;
}

// --------------------------------------------------------- [beRadicalAt]
void molecule::beRadicalAt(int idxAtom)
  { molec.atom[idxAtom].atomAttrib = 1; beUnique(); }

// -------------------------------------------------------- [beNonRadical]
void molecule::beNonRadical() {
int idRC;
  idRC = molec.getIdxRadCent();
  molec.atom[idRC].atomAttrib = 0;
  beUnique();
}

// -------------------------------------------------------- [isAtomInRing]
int molecule::isAtomInRing(int idxAtom) {
  if (!isCyclic) { return false; }
  return uniqueRepr.chain[0].isIdxInChain(idxAtom);
}

// -------------------------------------------------------- [getDstInRing]
int molecule::getDstInRing(int id1, int id2, int &dstInRing) {
  int dst, ir, inRing = false;
  vector<int> rt;
  dst = molec.getDstRoute(id1, id2, rt);
  dstInRing = 0;
  if (isCyclic) {
    for (ir = 0; ir <= dst; ir++) {
      if (uniqueRepr.chain[0].isIdxInChain(rt[ir])) {
        if (inRing) { dstInRing++; }
        else { inRing = true; }
      }
    }
  }
  return dst;
}

// --------------------------------------------------- [getDstInRingRoute]
int molecule::getDstInRingRoute(int id1, int id2, int &dstInRing,
 vector<int> &rt) {
  int dst, ir, inRing = false;
  dst = molec.getDstRoute(id1, id2, rt);
  dstInRing = 0;
  if (isCyclic) {
    for (ir = 0; ir <= dst; ir++) {
      if (uniqueRepr.chain[0].isIdxInChain(rt[ir])) {
        if (inRing) { dstInRing++; }
        else { inRing = true; }
      }
    }
  }
  return dst;
}

int molecule::nearestCH2noO(int idx) {                // --- nearestCH2noO
  // get index of nearest carbon with more than 2 H and without O
  int found = false, ia, ixp, ix, ib, ibn, cnxO;
  set<int> checked;
  vector<int> lastch, inchck;
  checked.clear(); lastch.clear();
  checked.insert(idx); lastch.push_back(idx);
  while (!found) {
    inchck.clear();
    for (ia = 0; ia < lastch.size(); ia++) {
      ixp = lastch[ia];
      for (ib = 0; ib < molec.atom[ixp].numBonds; ib++) {
        ix = molec.atom[ixp].bond[ib].toAtomIndex;
        if (checked.find(ix) == checked.end()) {
          if (molec.atom[ix].atomicNumber == 6) {
            if (molec.atom[ix].getNumHatoms() > 1) {
              cnxO = false;
              for (ibn = 0; ibn < molec.atom[ix].numBonds; ibn++) {
                if (molec.atom[ix].bond[ibn].toAtomicNbr == 8) { cnxO = true; break; }
              }
              if (!cnxO) { found = true; }
            }
          }
          if (found) { break; }
          checked.insert(ix); inchck.push_back(ix);
        }
      }
      if (found) { break; }
    }
    lastch = inchck;
    if (checked.size() >= molec.numAtoms) { break; }
  }
  if (!found) { return -1; }
  return ix;
}

void molecule::moveOOH(int ixfrom, int ixto, int ixo) {     // --- moveOOH
  molec.removeBond(ixfrom, ixo);
  molec.atom[ixfrom].atomAttrib--;
  molec.atom[ixto].atomAttrib++;
  molec.makeNewBond(ixto, ixo);
}

int molecule::isAldehyde() {                             // --- isAldehyde
  int idCo;
  if ((idCo = molec.getIdxAtomMatch(6, 8, 2)) < 0) { return false; }
  if (molec.atom[idCo].getNumHatoms() < 1) { return false; }
  return true;
}

int molecule::isEnol(int &idCdOH, int &idCd, int &idOH) {    // --- isEnol
  int idb, ia, iax;
  vector<vector<int> > dblBnds = molec.getAllBondsMatch(6, 6, 2);
  if (dblBnds.size() == 0) { return false; }
  for (idb = 0; idb < dblBnds.size(); idb++) {
    for (ia = 0; ia < 2; ia++) {
      if ((idOH = molec.getIdxNextAtom(dblBnds[idb][ia], 8, 1)) >= 0) {
        if (molec.atom[idOH].numBonds == 1) {
          idCdOH = dblBnds[idb][ia];
          iax = (ia + 1) % 2;
          idCd = dblBnds[idb][iax];
          return true;
        }
      }
    }
  }
  return false;
}

int molecule::isEnol() {                                     // --- isEnol
  int idCdOH, idCd, idOH;
  return isEnol(idCdOH, idCd, idOH);
}

int molecule::isVinyl() {                                   // --- isVinyl
  int idRC;
  if ((idRC = molec.getIdxRadCent()) < 0) { return false; }
  if (molec.getIdxNextAtom(idRC, 6, 2) < 0) { return false; }
  return true;
}

int molecule::isAllyl() {                                   // --- isAllyl
  int idRC;
  if ((idRC = molec.getIdxRadCent()) < 0) { return false; }
  if (molec.getIdxNextAtom(idRC, 6, 2, 6) < 0) { return false; }
  return true;
}

int molecule::isVinoxy() {                                 // --- isVinoxy
  int idRC;
  if ((idRC = molec.getIdxRadCent()) < 0) { return false; }
  if (molec.getIdxNextAtom(idRC, 6, 2, 8) < 0) { return false; }
  return true;
}

int molecule::isAcyl() {                                     // --- isAcyl
  int idRC;
  if ((idRC = molec.getIdxRadCent()) < 0) { return false; }
  if (molec.getIdxNextAtom(idRC, 8, 2) < 0) { return false; }
  return true;
}

int molecule::isBetaHOAlkyl() {                       // --- isBetaHOAlkyl
  int idRC, idCo, idOc;
  if ((idRC = molec.getIdxRadCent()) < 0) { return false; }
  if ((idCo = molec.getIdxNextAtom(idRC, 6, 1, 8)) < 0) { return false; }
  if (molec.getBondOrder(idRC, idCo) != 1) { return false; }
  idOc = molec.getIdxNextAtom(idCo, 8, 1);
  if (molec.atom[idOc].numBonds > 1) { return false; }
  return true;
}

ostream &operator<<(ostream &os, molecule &m)   // --------- [operator <<]
  { m.print(&os); return os; }


