/*--------------------------------------------------------------------------
     "reaction.cc" - 'reaction' and library classes (implementations)
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2022 by A. Miyoshi, Hiroshima University
                                       created: Sept. 14, 2002
                                        edited: July  30, 2007
                                        edited: Apr.  23, 2011
                                        edited: Aug.  24, 2011
                                        edited: Oct.  14, 2011
                                        edited: Apr.  17, 2016
                                        edited: Dec.   4, 2021
                                        edited: July  20, 2022
                                        edited: Jan.  12, 2023
                                   last edited: Apr.  19, 2023
--------------------------------------------------------------------------*/

#include "reaction.h"

/*****************************************************************************
     'rxnTunePar' and related classes
*****************************************************************************/

/*===== rateTuning class ===================================================*/

rateTuning::rateTuning() { clear(); }                   // --- constructor

void rateTuning::clear() { factA = 1.; shiftE = 0.; }         // --- clear

int rateTuning::restoreFrom(ifstream &ifs) {            // --- restoreFrom ---
  vector<string> vs;
  if (getTokensFrom(ifs, 2, vs) == ERRE) { return ERRE; }
  factA = atof(vs[0].c_str());
  shiftE = atof(vs[1].c_str());
  return NORM;
}

/*===== rxnTunePar class ===================================================*/


rxnTunePar::rxnTunePar() {                              // --- constructor
  libPath libp;
  clear();
  if (loadFrom(FL_RXNTUN, libp) == ERRE)
    { cout << "Error reading rxn tuning library.\n"; }
}

int rxnTunePar::loadFrom(string flname, libPath &libp) {   // --- loadFrom
  string s;
  rateTuning rtune;
  if (!empty()) { clear(); }
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  while (getTokenFrom(ifs, s) == NORM) {
    if (s == "") { break; }
    if (rtune.restoreFrom(ifs) == ERRE) { ifs.close(); return ERRE; }
    insert(make_pair(s, rtune));
  }
  ifs.close();
  cout << size() << " tuning parameters loaded from [" << flname << "]\n";
  return NORM;
}


/*****************************************************************************
     'rxnLib' and related classes
*****************************************************************************/

// ===== reagentInfo =======================================================

reagentInfo::reagentInfo() { clear(); }                // --- constructors

reagentInfo::reagentInfo(string r, string p, int f)
  { rct = r; prd = p; isRevrs = f; }

void reagentInfo::clear()                                     // --- clear
  { rct = ""; prd = ""; isRevrs = false; }

// ===== reagentMast =======================================================


// --------------------------------------------------------- [restoreFrom]
int reagentMast::restoreFrom(ifstream &ifs) {
  string s;
  vector<string> vs;
  if (!empty()) { clear(); }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "reagentMast{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numPrior = atoi(s.c_str());
  do {
    if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
    if (s != "}") {
      splitToken(s, "/", vs);
      if (vs.size() != 3) { return ERRE; }
      push_back(reagentInfo(vs[0], vs[1], valBoolStr(vs[2])));
    }
  } while (s != "}");
  return NORM;
}

// --------------------------------------------------------------- [print]
void reagentMast::print(ostream &os) {
  os << "reagentMast " << numPrior << " :";
  for (iterator p = begin(); p != end(); p++) {
    os << " " << p->rct << "/" << p->prd << "/" << formatBool(p->isRevrs);
  }
  os << endl;
}

// --------------------------------------------------------- [operator <<]
ostream &operator<<(ostream &os, reagentMast &rm)
  { rm.print(os); return os; }


/*============================================================================
     'reagentList' class
============================================================================*/


// --------------------------------------------------------- [restoreFrom]
int reagentList::restoreFrom(ifstream &ifs) {
  string s;
  if (!empty()) { clear(); }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "reagentList{") { return ERRE; }
  do {
    if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
    if (s != "}") { if (find(s) == end()) { insert(s); } }
  } while (s != "}");
  return NORM;
}

// --------------------------------------------------------------- [print]
void reagentList::print(ostream &os) {
  os << " reagentList:";
  for (iterator p = begin(); p != end(); p++) { os << " " << *p; }
  os << endl;
}

// --------------------------------------------------------- [operator <<]
ostream &operator<<(ostream &os, reagentList &rg)
  { rg.print(os); return os; }


/*============================================================================
     'rxnSite' class
============================================================================*/


// --------------------------------------------------------- [restoreFrom]
int rxnSite::restoreFrom(ifstream &ifs) {
  int l;
  string s;
  reagentList rg;
  reagentList::iterator p;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if ((l = s.length()) < 2) { return ERRE; }
  if ((s[0] != '[') || (s[l - 1] != ']')) { return ERRE; }
  snam = s.substr(1, l - 2);
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  prior = atoi(s.c_str());
  if (grpa.restoreFrom(ifs) == ERRE) { return ERRE; }
  if (rg.restoreFrom(ifs) == ERRE) { return ERRE; }
  if (!reag.empty()) { reag.clear(); }
  if (!fullrg.empty()) { fullrg.clear(); }
  for (p = rg.begin(); p != rg.end(); p++) {
    s = *p;
    if (s.substr(0, 1) == ".") { fullrg.insert(s.substr(1)); }
    else { reag.insert(s); fullrg.insert(s); }
  }
  return NORM;
}

// --------------------------------------------------------------- [print]
void rxnSite::print(ostream &os)
  { os << "rxnSite: " << snam << " " << prior << reag; }

// --------------------------------------------------------- [operator <<]
ostream &operator<<(ostream &os, rxnSite &rc)
  { rc.print(os); return os; }


/*============================================================================
     'rxnSiteList' class
============================================================================*/


// ---------------------------------------------------------------- [find]
rxnSiteList::iterator rxnSiteList::find(groupAtom &ga) {
  string idxs = ga.getIdxStr();
  for (iterator p = begin(); p != end(); p++) {
    if (p->grpa.getIdxStr() == idxs) { return p; }
  }
  return end();
}

// --------------------------------------------------------- [restoreFrom]
int rxnSiteList::restoreFrom(ifstream &ifs) {
  int numSites, ic;
  string s;
  rxnSite rc;
  if (!empty()) { clear(); }
  if (rgMst.restoreFrom(ifs) == ERRE) { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "rxnSiteList{") { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  priorLim = atoi(s.c_str());
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  numSites = atoi(s.c_str());
  for (ic = 0; ic < numSites; ic++) {
    if (rc.restoreFrom(ifs) == ERRE) { return ERRE; }
    push_back(rc);
  }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "}") { return ERRE; }
  return NORM;
}

// --------------------------------------------------------------- [print]
void rxnSiteList::print(ostream &os) {
  rgMst.print(os);
  os << "Priority limit : " << priorLim << endl;
  for (iterator p = begin(); p != end(); p++) { os << *p; }
}

// --------------------------------------------------------- [operator <<]
ostream &operator<<(ostream &os, rxnSiteList &rcl)
  { rcl.print(os); return os; }


/*============================================================================
     'rxnLib' class
============================================================================*/

rxnLib::rxnLib() {                                      // --- constructor
  libPath libp;
  clear();
  if (loadFrom(FL_RXNSIT, libp) == ERRE)
    { cout << "Error reading rxn site library.\n"; }
}

// ------------------------------------------------------------ [loadFrom]
int rxnLib::loadFrom(string flname, libPath &libp) {
  int l;
  string s, libName;
  rxnSiteList rcl;
  if (!empty()) { clear(); }
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  while (getTokenFrom(ifs, s) == NORM) {
    if ((l = s.length()) < 2) { return ERRE; }
    if ((s[0] != '[') || (s[l - 1] != ']')) { return ERRE; }
    libName = s.substr(1, l - 2);
    if (rcl.restoreFrom(ifs) == ERRE) { return ERRE; }
    insert(make_pair(libName, rcl));
    cout << " site list [" << libName << "] of " << rcl.size()
         << " sites restored.\n";
  }
  ifs.close();
  cout << size() << " reaction site lists loaded from [" << flname
       << "]\n";
  return NORM;
}

// --------------------------------------------------------------- [print]
void rxnLib::print(ostream &os) {
  for (iterator p = begin(); p != end(); p++) {
    os << "site list " << p->first << " contain " << p->second.size()
       << " sites.\n";
    os << p->second;
  }
}

// --------------------------------------------------------- [operator <<]
ostream &operator<<(ostream &os, rxnLib &rl)
  { rl.print(os); return os; }


/*****************************************************************************
     classes related to 'reaction'
*****************************************************************************/


/*============================================================================
     'rxnDescr' class
============================================================================*/


// ---------------------------------------------------------- [rxnDescr()]
rxnDescr::rxnDescr() { clear(); }

// --------------------------------------------------------------- [clear]
void rxnDescr::clear() {
  reac.clear(); prod.clear(); rate = ""; shiftE = 0.; factA = 1.; shiftB = 0.;
  dupc = 1; isRevrs = false; modTroe = false;
  modAL = 0.; modbL = 0.; modEaL = 0.;
  modaFc = 0.; modT3Fc = 0.; modT1Fc = 0.; modT2Fc = 0.;
}

// --------------------------------------------------------------- [print]
void rxnDescr::print(ostream &os) {
  vector<string>::iterator p;
  for (p = reac.begin(); p != reac.end(); p++) { os << " " << *p; }
  os << " =";
  for (p = prod.begin(); p != prod.end(); p++) { os << " " << *p; }
  os << " $ " << rate << " " << shiftE << " " << (factA * dupc) << " "
     << formatBool(isRevrs);
  if (modTroe) {
    // always put shiftB !
    os << " " << shiftB;
    os << " " << modAL;
    os << " " << modbL;
    os << " " << modEaL;
    os << " " << modaFc;
    os << " " << modT3Fc;
    os << " " << modT1Fc;
    os << " " << modT2Fc;
  } else if (fabs(shiftB) > 0.0004) { os << " " << shiftB; }
  os << endl;
}

// --------------------------------------------------------- [operator <<]
ostream &operator<<(ostream &os, rxnDescr &r) { r.print(os); return os; }


/*============================================================================
     'rxnList' class
============================================================================*/

// ----------------------------------------------- [appendFrom(rxnList &)]
void rxnList::appendFrom(rxnList &alst) {
  for (iterator p = alst.begin(); p != alst.end(); p++) { push_back(*p); }
}

// ------------------------------------------------- [procDupProd(string)]
//   if one of the products matches with pnam, add 1 to dupc

int rxnList::procDupProd(string pnam) {
  int dupFound = false;
  vector<string>::iterator ps;
  for (iterator p = begin(); p != end(); p++) {
    for (ps = p->prod.begin(); ps != p->prod.end(); ps++) {
      if (*ps == pnam) { p->dupc += 1; dupFound = true; break; }
    }
  }
  return dupFound;
}

// ------------------------------------------ [procDupProd(int, string[])]
//   if products matches with n and pnam[], add 1 to dupc
//    (only for the case all products match)

int rxnList::procDupProd(int n, string pnam[]) {
  int dupFound = false, diff, ip, foundMatch;
  vector<string>::iterator ps;
  for (iterator p = begin(); p != end(); p++) {
    if (p->prod.size() == n) {
      diff = false;
      for (ps = p->prod.begin(); ps != p->prod.end(); ps++) {
        foundMatch = false;
        for (ip = 0; ip < n; ip++) {
          if (*ps == pnam[ip]) { foundMatch = true; }
        }
        if (!foundMatch) { diff = true; break; }
      }
      if (!diff) { p->dupc += 1; dupFound = true; break; }
    }
  }
  return dupFound;
}

// ----------------------------------------------------- [procDupProdPart]
//   if part of products matches with n and pnam[], add 1 to dupc
//    (require part matching only)

int rxnList::procDupProdPart(int n, string pnam[]) {
  int ip, dupFound = false, foundMatch, diff;
  vector<string>::iterator ps;
  for (iterator p = begin(); p != end(); p++) {
    diff = false;
    for (ip = 0; ip < n; ip++) {
      foundMatch = false;
      for (ps = p->prod.begin(); ps != p->prod.end(); ps++) {
        if (*ps == pnam[ip]) { foundMatch = true; break; }
      }
      if (!foundMatch) { diff = true; break; }
    }
    if (!diff) { p->dupc += 1; dupFound = true; }
  }
  return dupFound;
}

// --------------------------------------------------------- [findDupRevs]
//   if reverse reaction is found, returns true

int rxnList::findDupRevs(string rnam, string pnam) {
  for (iterator p = begin(); p != end(); p++) {
    if ((p->reac.size() == 1) && (p->prod.size() == 1)) {
      if ((*(p->reac.begin()) == pnam) && (*(p->prod.begin()) == rnam))
        { return true; }
    }
  }
  return false;
}

// --------------------------------------------------------------- [print]
void rxnList::print(ostream &os)
  { for (iterator p = begin(); p != end(); p++) { p->print(os); } }

// --------------------------------------------------------- [operator <<]
ostream &operator<<(ostream &os, rxnList &rl) { rl.print(os); return os; }


/*****************************************************************************
     'reaction' class - reaction processor
*****************************************************************************/


/****** public member functions *********************************************/

// --------------------------------------------------------------- [clear]
void reaction::clear() {
  lst.clear(); rsvLst.clear();
  namPrdGen.clear(); namPrdRef.clear(); numRxnGen = 0;
  spAd = 0; grLb = 0; rxLb = 0; tyMol = 0; oStrm = 0; fuels = 0;
  radFuel.clear(); ro2Fuel.clear();
  // - for debugging
  // tmpFlag = false;
  //
}

// -------------------------------------------------------- [initPrivMemb]
//   initialize private members

void reaction::initPrivMemb(specAdmin &sa, groupLib &g, rxnLib &rl,
 rxnTunePar &rt, typeMolec &tm, outStrms &os, vector<molecule> &f) {
  int ix;
  clear();
  for (ix = 0; ix < f.size(); ix++)
    { if (!f[ix].isRxnDone) { namPrdGen.insert(f[ix].molName); } }
  spAd = &sa; grLb = &g; rxLb = &rl; rxTp = &rt; tyMol = &tm;
  oStrm = &os; fuels = &f;
  prepRadRo2Fuel();
}

// ------------------------------------------------------- [reportGenStat]
//   output report on generated reactions and molecules

void reaction::reportGenStat() {
  int numC; long mtyp;
  set<string>::iterator p;
  molecList::iterator q;
  map<int, map<long, set<string> > > lstMol;
  map<int, map<long, set<string> > >::iterator r;
  map<long, set<string> >::iterator u;

  cout << numRxnGen << " reactions generated.\n";
  cout << namPrdGen.size() << " species generated.\n";
  lstMol.clear();
  for (p = namPrdGen.begin(); p != namPrdGen.end(); p++) {
    q = spAd->molAd.find(*p); mtyp = q->mtyp;
    numC = q->molec.getNumAtoms(6);
    lstMol[-numC][mtyp].insert(*p);
  }
  for (r = lstMol.begin(); r != lstMol.end(); r++) {
    cout << " C" << -(r->first) << " molecules\n";
    for (u = r->second.begin(); u != r->second.end(); u++) {
      cout << "  " << u->second.size() << " " << tyMol->typName(u->first)
        << " (" << hex << u->first << dec << ") :";
      for (p = u->second.begin(); p != u->second.end(); p++)
        { cout << ' ' << *p; }
      cout << endl;
    }
  }
  cout << namPrdRef.size() << " base species referred.\n";
  lstMol.clear();
  for (p = namPrdRef.begin(); p != namPrdRef.end(); p++) {
    q = spAd->molAd.find(*p); mtyp = q->mtyp;
    numC = q->molec.getNumAtoms(6);
    lstMol[-numC][mtyp].insert(*p);
  }
  for (r = lstMol.begin(); r != lstMol.end(); r++) {
    cout << " C" << -(r->first) << " molecules\n";
    for (u = r->second.begin(); u != r->second.end(); u++) {
      cout << "  " << u->second.size() << " " << tyMol->typName(u->first)
        << " (" << hex << u->first << dec << ") :";
      for (p = u->second.begin(); p != u->second.end(); p++)
        { cout << ' ' << *p; }
      cout << endl;
    }
  }
}

// =========================================================== [reactQTop]
//   react molecule at the top in the queue
// =======================================================================

int reaction::reactQTop() {
  static const long bitCyc = 1L << 27, mskCyc = ~bitCyc;
  int isFuel = false, cNRxnGen = numRxnGen, cNPrdGen = namPrdGen.size();
  long mtyp;
  string typ, typCm;
  molecule tmol;

  if (!lst.empty()) { lst.clear(); }
  if (spAd->rxnQue.getPopQueTop(tmol) == ERRE) { return ERRE; }
  mtyp = tmol.getType(*grLb);
  typ = tyMol->typName(mtyp); typCm = tyMol->typName(mtyp & mskCyc);
  isFuel = checkIsFuel(tmol);
// - for debudding
//  if (tmol.molName == "C7H13O3e") { tmpFlag = true; }
//  if (tmpFlag) {
//    cout << "type:" << hex << mtyp << dec << " - [" << tmol.molName << "]";
//    int dummy;
//    cin >> dummy;
//    cout << endl;
//  }
//
  if (typCm == "Alkane") {
    if (isFuel) { rxnHabsFuel(tmol); } else { rxnHabsPart(tmol); }
    rxnDecmpNonRad(tmol); }
  else if (typCm == "Alkene") {
    rxnDecmpNonRad(tmol);
    if (isFuel) { rxnHabsFuel(tmol);
      rxnDecmpCHnonR(tmol); rxnAddRadAlkene(tmol);
    } else { rxnHabsAlkene(tmol); }
    rxnOaddAlkene(tmol); rxnAddHelAlkene(tmol);
  }
  else if (typCm == "Alcohol") {
    if (isFuel) { rxnHabsFuel(tmol); } else { rxnHabsPart(tmol); }
    rxnDecmpNonRad(tmol); }
  else if (typCm == "RdOH") {
    if (tmol.isEnol()) { rxnEnol(tmol); }   // - RdOH may be enol
    rxnHabsAlkene(tmol); rxnOaddAlkene(tmol);
  }
  else if (typCm == "Allene") {
    // 221014 disabled
    // rxnOaddAlkene(tmol);
    //
    rxnDecmpNonRad(tmol);
  }
  else if ((typ == "Alkenyne") || (typ == "Allenyne") || (typ == "Aldalknyn"))
    { rxnDecmpNonRad(tmol); }
  else if ((typ == "CycEther") || (typ == "CycEthDB") || (typ == "CyEthCHO")
   || (typ == "RCOCyEth") || (typ == "CyCEthOH")) {
    rxnHabsCycEther(tmol);
    if (numRxnGen == cNRxnGen) { rxnDecmpCycEther(tmol); }
  }
  else if ((typCm == "RCHO") || (typCm == "RCOR") || (typCm == "hoRCHO") ||
    (typCm == "hoRCOR") || (typCm == "hoRKetCHO") || (typCm == "RCOOH") ||
    (typCm == "HCOQCOOH") || (typCm == "CdCOOH")) {
    rxnHabsKetAld(tmol); rxnDecmpKetAld(tmol);
    if (numRxnGen == cNRxnGen) { rxnDecmpNonRad(tmol); }
  }
  else if (typCm == "RCOQCOOH") { rxnHabsKetAld(tmol); }
  else if ((typCm == "CdQCOR") || (typ == "hoCdQCRO")) {
    rxnHabsKetAld(tmol);
    if (numRxnGen == cNRxnGen) { rxnOaddAlkene(tmol); }
  }
  else if (typCm == "hoRaldCCO") {
    rxnHabsAldOnly(tmol);
    if (numRxnGen == cNRxnGen) { rxnOaddAlkene(tmol); }
  }
  else if ((typCm == "OCCQCHO") || (typCm == "RCOQCHO") || (typ == "hoCdQCHO") ||
     (typ == "OCCKetCHO") || (typ == "AllCHO") || (typ == "Aldactln"))
    { rxnHabsAldOnly(tmol); }
  else if ((typCm == "CdQCHO") || (typCm == "RCOCdQCHO")) {
    if (tmol.isAldehyde()) { rxnHabsAldOnly(tmol); }
    else { rxnOaddAlkene(tmol); }
  }
  else if ((typCm == "RCCO") || (typCm == "RdCCO") || (typ == "RCOQCCO") ||
    (typCm == "hoRCCO") || (typCm == "hoRCOQCCO") || (typ == "hoRCCCO") ||
    (typ == "OCCCQCHO") || (typCm == "RCORdCCO") || (typ == "RCCCO") ||
    (typ == "OCCO"))
    { rxnKetene(tmol); }
  else if ((typCm == "ROOH") || (typCm == "RdOOH") || (typCm == "hoROOH") ||
    (typCm == "hoRdOOH")) { rxnDecmpROOH(tmol, 0.); /* rxnDecmpSpec(tmol, "O-O"); */ }
  else if ((typCm == "KetOOH") || (typCm == "AldOOH") || (typCm == "AcdOOH") ||
    (typCm == "KetDOOH") || (typCm == "AldDOOH") || (typCm == "hoAldOOH") ||
    (typCm == "hoKetOOH") || (typCm == "KetenOOH")) { rxnDecmpROOH(tmol, 0.); /* rxnDecmpKetOOH(tmol); */ }
  else if (typCm == "R") { rxnDecmpRad(tmol); rxnIsomRad(tmol);
                           rxnRecmbRadO2(tmol); rxnOabsNewR(tmol); }
  else if (typCm == "hoR") { rxnDecmpRad(tmol); rxnIsomRad(tmol);
                             rxnRecmbRadO2(tmol); rxnOabsNewR(tmol); }
  else if (typCm == "HOOhoRalp")
    { rxnDecmpRad(tmol); rxnO2addHORalp(tmol); rxnDecmpQOOH(tmol); }
  else if ((typCm == "hoRalp") || (typCm == "hoRDalp") ||
    (typCm == "hoAldRalp") || (typCm == "hoKetRalp"))
    { rxnDecmpRad(tmol); rxnO2addHORalp(tmol); }
  else if (typCm == "Alkenyl") {
    if (checkIsRadFuel(tmol))
      { rxnIsomRad(tmol); rxnRecmbRadO2(tmol); rxnOabsNewR(tmol); }
    rxnDecmpRad(tmol); }
  else if ((typCm == "Allyl") || (typCm == "AllAllyl") || (typCm == "AlleAllyl")) {
    rxnDecmpRad(tmol);
    if (checkIsRadFuel(tmol)) {
      if ((*rxTp)["AllylP/enabled"].factA > 0.)
        { rxnRecmbRadO2(tmol); }
      rxnOabsNewR(tmol);
    }
    rxnO2addAllyl(tmol);
  }
  else if (typCm == "Allenyl") { rxnDecmpRad(tmol); }
  else if ((typCm == "Propargyl") || (typCm == "ppgAllyl") ||
    (typCm == "Aldppgyl") || (typCm == "Alleppgyl") || (typCm == "AlAlyppg") ||
    (typCm == "Keteppgyl"))
    { rxnDecmpRad(tmol); }
  else if ((typCm == "Pntdnyl") || (typCm == "hoPntdnyl") || (typCm == "acPntdnyl"))
    { rxnO2addAllyl(tmol); }
  else if ((typCm == "coAlkenyl") || (typCm == "hoAlkenyl"))
    { rxnDecmpRad(tmol); }
  else if (typCm == "hoAllyl") { rxnO2addAllyl(tmol); }
  else if ((typCm == "Vinyl") || (typ == "rcoVinyl") || (typ == "hcoVinyl") ||
    (typ == "hoVinyl") || (typ == "Cumulenyl") || (typ == "acVinyl"))
    { rxnO2addVinyl(tmol); }
  else if ((typ == "Ketyl") || (typ == "RCOKetyl")) { rxnO2addKetyl(tmol); }
  else if (typCm == "ROO") { rxnElimHO2RO2(tmol); rxnIsomRad(tmol);
                             rxnHabsNewRO2(tmol); rxnOabsNewRO2(tmol);
                             rxnRO2RO2(tmol); }
  else if (typCm == "hoROO") { rxnElimHO2RO2(tmol); rxnIsomRad(tmol);
                             rxnHabsNewRO2(tmol); rxnOabsNewRO2(tmol);
                             rxnRO2RO2(tmol); }
  else if (typCm == "RdOO") { rxnElimHO2RO2(tmol); rxnIsomRad(tmol);
                              rxnIntAddRdO2(tmol);
                              rxnHabsNewRO2(tmol); rxnOabsNewRO2(tmol);
                              rxnRO2RO2(tmol); }
  else if ((typCm == "RO") || (typCm == "RdO") || (typCm == "HOOQO") ||
           (typCm == "AldRO") || (typCm == "AcdRO") || (typCm == "KetRO") ||
           (typCm == "HCOQOOH") || (typCm == "KetylRO") ||
           (typCm == "KetRdO") || (typCm == "AldRdO"))
    { rxnDecmpRad(tmol); }
  else if ((typCm == "QOOH") || (typCm == "QdOOH"))
    { rxnDecmpQOOH(tmol); rxnRecmbQOOHO2(tmol); }
  else if (typCm == "QaOOH") {
    rxnDecmpQaOOH(tmol);
    if ((*rxTp)["AllylP/enabled"].factA > 0.)
      { rxnRecmbQOOHO2(tmol); }
  }
  else if ((typCm == "alpQOOH") || (typCm == "OCCQOOH")) { rxnDecmpRad(tmol); }
  else if (typCm == "hoQOOH") {
    // if it is beta-OH-alkyl, dissociate C-O
    if (tmol.isBetaHOAlkyl()) { rxnDecmpRad(tmol); }
    // - Jan 7, 2011
    // else { rxnDecmpQOOH(tmol); rxnRecmbQOOHO2(tmol); }
    rxnDecmpQOOH(tmol); rxnRecmbQOOHO2(tmol);
    // - 
  }
  else if ((typ == "CHOPOOH") || (typ == "RCOPOOH")) { rxnDecmpQOOH(tmol); }
  else if ((typCm == "OOQOOH") || (typCm == "OOQdOOH") ||
    (typCm == "OOhoQOOH")) { rxnIsomOOQOOH(tmol); }
  else if ((typCm == "RCO") || (typCm == "hoRCO") || (typCm == "AllCO") ||
    (typCm == "AllCOa")) { rxnDecmpRCO(tmol); }
  else if ((typCm == "RCOQCO") || (typCm == "CdQCO") || (typCm == "OCCQCO") ||
    (typ == "hoCdQCO") || (typCm == "hoRCOQCO") || (typ == "hoKtQCO") ||
    (typCm == "KetCdQCO") || (typ == "OCCKetCO")) { rxnDecmpRCO(tmol); }
  else if (typ == "OCCAcyl") {
    if (tmol.isAcyl()) { rxnDecmpRCO(tmol); }
    else { rxnDecmpRad(tmol); }
  }
//  - Dec. 15, 2016
// else if ((typ == "OCCV") || (typ == "hoOCCV")) {
  else if ((typCm == "OCCV") || (typ == "hoOCCV")) {
    if (tmol.isAllyl()) { rxnO2addAllyl(tmol); }
    else { rxnDecmpOCCV(tmol); }
  }
  else if ((typCm == "RCOQ") || (typCm == "RdCOQ") || (typCm == "HCOQ"))
    { rxnDecmpKetR(tmol); rxnO2addKetR(tmol); }
  else if ((typ == "CHOQCOQ") || (typ == "HOCOQ")) { rxnDecmpKetR(tmol); }
  else if ((typCm == "RCOV") || (typCm == "hoRCOV"))
    { rxnDecmpVinoxy(tmol); rxnO2addVinoxy(tmol); }
  else if (typCm == "HCOVQOH") {
    rxnDecmpRad(tmol);
    if (numRxnGen == cNRxnGen) { rxnDecmpVinoxy(tmol); rxnO2addVinoxy(tmol); }
  }
  else if (typCm == "OCCQVin") {
    if (tmol.isVinoxy()) { rxnDecmpVinoxy(tmol); rxnO2addVinoxy(tmol); }
    else { rxnDecmpRad(tmol); }
  }
  else if ((typCm == "RCOVCHO") || (typCm == "RCOVohCHO") ||
    (typCm == "AK3Vinoxy"))
    { rxnDecmpVinoxy(tmol); rxnO2addVinoxy(tmol); }
  else if ((typCm == "HCOV") || (typCm == "HOCOV") || (typCm == "HOCOVCHO")) {
    // this may be acyl
    if (tmol.isAcyl()) { rxnDecmpRCO(tmol); }
    else { rxnDecmpVinoxy(tmol); rxnO2addVinoxy(tmol); }
  }
  else if (typCm == "HCOVinoxy") { rxnDecmpVinoxy(tmol); }
  else if ((typCm == "HCOVQCd") || (typCm == "CdQCOV") || (typ == "hoRdCOV") ||
   (typ == "hoRdHCOV") || (typ == "KAKvin")) {
    // these may be acyl, vinoxy, or allyl
    if (tmol.isAcyl()) { rxnDecmpRCO(tmol); }
    else if (tmol.isVinoxy()) { rxnDecmpVinoxy(tmol); rxnO2addVinoxy(tmol); }
    else if (tmol.isAllyl()) { rxnO2addAllyl(tmol); }
  }
  else if ((typCm == "HCOVAlly") || (typCm == "hoHCOVAlly") ||
   (typCm == "RCOVAlly") || (typCm == "hoRCOVAlly")) {
    if (tmol.isCyclic) { rxnO2addAllyl(tmol); }
    else {
      rxnO2addVinoxy(tmol); 
      if (numRxnGen == cNRxnGen) { rxnO2addAllyl(tmol); }
    }
  }
  else if ((typCm == "OCCVAlly") || (typCm == "OCCVhoAlly") ||
    (typCm == "OCCkVAlly") || (typCm == "OCCaVAlly"))
    { rxnO2addAllyl(tmol); }
// - Dec. 15, 2016
// else if ((typ == "CHOQCOV") || (typ == "CHOQohCOV") || (typCm == "CdQHCOV")
  else if ((typCm == "CHOQCOV") || (typ == "CHOQohCOV") || (typCm == "CdQHCOV")
   || (typCm == "HOCOQCOV") || (typCm == "AllVinoxy")) {
    if (tmol.isAcyl()) { rxnDecmpRCO(tmol); }
    else { rxnDecmpVinoxy(tmol); }
  }
  else if ((typCm == "OCCQ") || (typCm == "hoQCHO") || (typCm == "hoQCRO") ||
   (typCm == "RCO2") || (typ == "OCCQKet") || (typCm == "OCQCO2") ||
   (typCm == "CHOQCO2") || (typ == "hoQCCO") || (typ == "CdCO2"))
    { rxnDecmpRad(tmol); }
  else
    { cout << "ERROR [unknown type of molecule: "
       << hex << mtyp << dec << " - " << tmol.molName << "].\n";
      return ERRE; }
  if ((numRxnGen == cNRxnGen) && (namPrdGen.size() == cNPrdGen)) {
    cout << "## WARNING ## no reaction has been generated for species ["
         << tmol.molName << "].\n";
  }
  endReact(tmol);
  return NORM;
}


/****** general purpose private functions ***********************************/

// ------------------------------------------------------ [prepRadRo2Fuel]
//   prepare radFuel and ro2Fuel for fuels in library

void reaction::prepRadRo2Fuel() {
  static const long bitCyc = 1L << 27, mskCyc = ~bitCyc;
  long mtyp, rtyp;
  string typCm, rtypS, namRad, namRO2;
  int ia, idba;
  vector<int> dblBnd;
  vector<molecule>::iterator p;
  molecule im, rad, ro2;
  groupAtom ga;

  for (p = fuels->begin(); p != fuels->end(); p++) {
    if ((p->isRxnDone) && (p->molName != "CH4")) {
      mtyp = p->getType(*grLb);
      typCm = tyMol->typName(mtyp & mskCyc);
      // - H-abstraction products
      for (ia = 0; ia < p->molec.numAtoms; ia++) {
        p->getGroupAtom(ia, ga);
        if ((*rxLb)["Habs"].find(ga) != (*rxLb)["Habs"].end()) {
          rad = *p; rad.beRadicalAt(ia);
          // -------------------------------- Aug. 24, 2011
          // - avoid to store "RO" in radFuel
          rtyp = rad.getType(*grLb);
          rtypS = tyMol->typName(rtyp);
          if (rtypS != "RO") {
            if (spAd->getMolStat(rad, namRad) == 3) {
              rad.molName = namRad;
              if (radFuel.find(namRad) == radFuel.end()) {
                radFuel.push_back(molecRef(rad, *grLb));
                molAddRcO2(rad, ro2);
                if (spAd->getMolStat(ro2, namRO2) == 3) {
                  ro2.molName = namRO2;
                  if (ro2Fuel.find(namRO2) == ro2Fuel.end()) {
                    ro2Fuel.push_back(molecRef(ro2, *grLb));
                  }
                }
              }
            }
          }
          // --------------------------------
        }
      }
      // - OH-addition products
      if (typCm == "Alkene") {
        dblBnd = p->molec.getBondMatch(6, 6, 2);
        im = *p; im.molec.decBondOrder(dblBnd[0], dblBnd[1]);
        for (idba = 0; idba < 2; idba++) {
          molAddOH(im, rad, dblBnd[idba]); rad.beUnique();
          if (spAd->getMolStat(rad, namRad) == 3) {
            rad.molName = namRad;
            if (radFuel.find(namRad) == radFuel.end()) {
              radFuel.push_back(molecRef(rad, *grLb));
              molAddRcO2(rad, ro2);
              if (spAd->getMolStat(ro2, namRO2) == 3) {
                ro2.molName = namRO2;
                if (ro2Fuel.find(namRO2) == ro2Fuel.end()) {
                  ro2Fuel.push_back(molecRef(ro2, *grLb));
                }
              }
            }
          }
        }
      }
    }
  }
}

// --------------------------------------------------------- [checkIsFuel]

int reaction::checkIsFuel(molecule &m) {
  vector<molecule>::iterator p;
  for (p = fuels->begin(); p != fuels->end(); p++) {
    if (m.molec == p->molec) { return true; }
  }
  return false;
}

int reaction::checkIsRadFuel(molecule &m) {          // --- checkIsRadFuel
  if (radFuel.find(m) != radFuel.end()) { return true; }
  return false;
}

int reaction::checkIsRo2Fuel(molecule &m) {          // --- checkIsRo2Fuel
  if (ro2Fuel.find(m) != ro2Fuel.end()) { return true; }
  return false;
}

int reaction::checkIsRxnDone(molecule &m) {          // --- checkIsRxnDone
  string nam;
  int mstat = spAd->getMolStat(m, nam);
  if (mstat == 3) { return true; }
  return false;
}

// ------------------------------------------------------------ [procProd]
//   process products

void reaction::procProd(molecule &prod, string bnam) {
  prod.beUnique();
  string namProd;
  int mstat = spAd->getMolStat(prod, namProd);
  if (mstat == 0) { namProd = spAd->namAd.regBaseAuto(bnam); }
  prod.molName = namProd;
  if (mstat == 2) { spAd->molAd.removeMol(prod); }
  if ((mstat % 2) == 0) {
    oStrm->thm << prod.getThermInput(*grLb) << endl;
    prod.storeTo(oStrm->mol);
    prod.storeLogTo(oStrm->mll, *grLb, *tyMol);
    spAd->rxnQue.push_back(molecInQue(prod));
    namPrdGen.insert(namProd);
  }
  if (mstat == 3) {
    if (namPrdGen.find(namProd) == namPrdGen.end())
      { namPrdRef.insert(namProd); }
  }
}

// ---------------------------------------------------- [procProdAutoBase]
//   process products - automatically setting basename

void reaction::procProdAutoBase(molecule &prod) {
  string basNam;
  basNam = prod.getTypeFormula(*grLb, *tyMol);
  procProd(prod, basNam);
}

// --------------------------------------------------------- [outLstClear]
//   output lst to oStrm->rxn, and clear lst.

void reaction::outLstClear() {
  numRxnGen += lst.size();
  oStrm->rxn << lst; lst.clear();
}

// --------------------------------------------------------- [outLstResrv]
//   output lst to oStrm->rxn, and reserve lst in rsvLst.

void reaction::outLstResrv() {
  numRxnGen += lst.size();
  oStrm->rxn << lst;
  rsvLst.appendFrom(lst);
  lst.clear();
}

// ------------------------------------------------------------ [endReact]
void reaction::endReact(molecule &m) {
  m.isRxnDone = true; spAd->molAd.push_back(molecRef(m, *grLb));
}


/****** type-specific reaction processors ***********************************/

// ========================================================= [rxnHabsFuel]
//   H-atom abstraction from the fuel

void reaction::rxnHabsFuel(molecule &m) {
  int pos, ia, mstat, flagGen, isCCrad, idRC, ib, nCnbo;
  double eShftNBO = 0.;
  string prfNam = "", basNam, namPrd, siteName, namRct = m.molName;
  string namRO2, namRO2H;
  groupAtom ga;
  rxnSiteList::iterator p;
  molecule prod, pRO2H, RO2;
  reagentMast::iterator prg;
  rxnDescr rxd;
  molecList::iterator pm;

  if ((pos = m.molName.find(m.getTypeFormula(*grLb, *tyMol))) != string::npos) {
    if (pos > 0) { prfNam = m.molName.substr(0, pos); }
  }
  for (ia = 0; ia < m.molec.numAtoms; ++ia) {
    m.getGroupAtom(ia, ga);
    if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
      prod = m; prod.beRadicalAt(ia);
      basNam = prfNam + prod.getTypeFormula(*grLb, *tyMol);
      procProd(prod, basNam); namPrd = prod.molName;
      isCCrad = true;
      idRC = prod.molec.getIdxRadCent();
      if (prod.molec.atom[idRC].atomicNumber != 6) { isCCrad = false; }
      else {
        for (ib = 0; ib < prod.molec.atom[idRC].numBonds; ib++) {
          if (prod.molec.atom[idRC].bond[ib].order != 1) { isCCrad = false; }
        }
      }
      if (isCCrad) {
        if (radFuel.find(namPrd) == radFuel.end())
          { radFuel.push_back(molecRef(prod, *grLb)); }
      }
      if (lst.procDupProd(namPrd) == false) {
        siteName = p->snam;
        nCnbo = prod.molec.countNextButOneC(idRC);
        for (prg = (*rxLb)["Habs"].rgMst.begin();
             prg != (*rxLb)["Habs"].rgMst.end(); prg++) {
          if (p->fullrg.find(prg->rct) != p->fullrg.end()) {
            if (prg->rct == "RO2") {
              // === re-activated Nov. 10, 2003 ===
              for (pm = spAd->molAd.begin(); pm != spAd->molAd.end();
               pm++) {
                if (pm->isRO2ofDone()) {
                  namRO2 = pm->name; pRO2H.clear();
                  pRO2H.molec = pm->molec; pRO2H.beNonRadical();
                  mstat = spAd->getMolStat(pRO2H, namRO2H);
                  if (mstat == 3) { flagGen = true; }
                  else {
                    RO2.molec = pm->molec; RO2.molName = pm->name;
                    RO2.beUnique();
                    if (checkIsRo2Fuel(RO2)) {
                      procProdAutoBase(pRO2H); namRO2H = pRO2H.molName;
                      flagGen = true;
                    } else { flagGen = false; }
                  }
                  if (flagGen) {
                    pRO2H.molName = namRO2H;
                    rxd.clear();
                    rxd.reac.push_back(namRct); rxd.reac.push_back(namRO2);
                    rxd.prod.push_back(namPrd); rxd.prod.push_back(namRO2H);
                    rxd.rate = "Habs/by-" + prg->rct + "/" + siteName;
                    rxd.isRevrs = prg->isRevrs;
                    rxd.shiftE = eShftNBO * nCnbo;
                    lst.push_back(rxd);
                  }
                }
              }
              // ===
            } else {
              rxd.clear();
              rxd.reac.push_back(namRct); rxd.reac.push_back(prg->rct);
              rxd.prod.push_back(namPrd); rxd.prod.push_back(prg->prd);
              rxd.rate = "Habs/by-" + prg->rct + "/" + siteName;
              rxd.isRevrs = prg->isRevrs;
              rxd.shiftE = eShftNBO * nCnbo;
              lst.push_back(rxd);
            }
          }
        }
      }
    }
  }
  outLstClear();
}

// ========================================================= [rxnHabsPart]
//   H-atom abstraction reactions (lists only high priority sites)

void reaction::rxnHabsPart(molecule &m) {
  int pos, ia, maxPri = 0, priThres, numRgtM, irgm;
  string prfNam, basNam, namPrd, siteName, namRct = m.molName;
  groupAtom ga;
  rxnSiteList::iterator prs;
  reagentMast rgtMst;
  molecule prod;
  rxnDescr rxd;

  if ((pos = m.molName.find(m.molec.getChemFormula())) != string::npos)
    { if (pos > 0) { prfNam = m.molName.substr(0, pos); } }
  for (ia = 0; ia < m.molec.numAtoms; ++ia) {
    m.getGroupAtom(ia, ga);
    if ((prs = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
      if (prs->prior > maxPri) { maxPri = prs->prior; }
    }
  }
  priThres = maxPri - (*rxLb)["Habs"].priorLim;
  rgtMst = (*rxLb)["Habs"].rgMst; numRgtM = rgtMst.numPrior;
  for (ia = 0; ia < m.molec.numAtoms; ++ia) {
    m.getGroupAtom(ia, ga);
    if ((prs = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
      if (prs->prior >= priThres) {
        prod = m; prod.beRadicalAt(ia);
        basNam = prfNam + prod.molec.getChemFormula();
        procProd(prod, basNam); namPrd = prod.molName;
        if (lst.procDupProd(namPrd) == false) {
          siteName = prs->snam;
          for (irgm = 0; irgm < numRgtM; irgm++) {
            if (prs->reag.find(rgtMst[irgm].rct) != prs->reag.end()) {
              rxd.clear();
              rxd.reac.push_back(namRct);
              rxd.reac.push_back(rgtMst[irgm].rct);
              rxd.prod.push_back(namPrd);
              rxd.prod.push_back(rgtMst[irgm].prd);
              rxd.rate = "Habs/by-" + rgtMst[irgm].rct + "/" + siteName;
              rxd.isRevrs = rgtMst[irgm].isRevrs;
              lst.push_back(rxd);
            }
          }
        }
      }
    }
  }
  outLstClear();
}

// ======================================================= [rxnHabsNewRO2]
//   H-atom abstraction by newly added RO2 from fuel, HO2, and H2O2

void reaction::rxnHabsNewRO2(molecule &ro2) {
  int pos, ix, ia, nogen;
  string namRO2 = ro2.molName, namRO2H, namFuel, namRadF;
  string prfNam, prfNamRO2 = "", basNam, siteName;
  groupAtom ga;
  rxnSiteList::iterator p;
  molecule pRadF, pRO2H;
  rxnDescr rxd;

  if ((pos = ro2.molName.find(ro2.getTypeFormula(*grLb, *tyMol))) != string::npos) {
    if (pos > 0) { prfNamRO2 = ro2.molName.substr(0, pos); }
  }
  pRO2H.clear(); pRO2H.molec = ro2.molec; pRO2H.beNonRadical();
// procProdAutoBase(pRO2H);
  basNam = prfNamRO2 + pRO2H.getTypeFormula(*grLb, *tyMol);
  procProd(pRO2H, basNam);
//
  namRO2H = pRO2H.molName;
  if (checkIsRo2Fuel(ro2)) {
    for (ix = 0; ix < fuels->size(); ix++) {
      namFuel = (*fuels)[ix].molName; prfNam = "";
      if ((pos = (*fuels)[ix].molName.find((*fuels)[ix].molec.getChemFormula()))
          != string::npos)
        { if (pos > 0) { prfNam = (*fuels)[ix].molName.substr(0, pos); } }
      for (ia = 0; ia < (*fuels)[ix].molec.numAtoms; ++ia) {
        (*fuels)[ix].getGroupAtom(ia, ga);
        if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
          pRadF = (*fuels)[ix]; pRadF.beRadicalAt(ia);
          nogen = false;
          if ((*fuels)[ix].isRxnDone)
            { if (spAd->getMolStat(pRadF, namRadF) != 3) { nogen = true; } }
          if (!nogen) {
            basNam = prfNam + pRadF.molec.getChemFormula();
            procProd(pRadF, basNam); namRadF = pRadF.molName;
            if (lst.procDupProd(namRadF) == false) {
              siteName = p->snam;
              if (p->fullrg.find("RO2") != p->fullrg.end()) {
                rxd.clear();
                rxd.reac.push_back(namFuel); rxd.reac.push_back(namRO2);
                rxd.prod.push_back(namRadF); rxd.prod.push_back(namRO2H);
                rxd.rate = "Habs/by-RO2/" + siteName;
                rxd.isRevrs = true;
                lst.push_back(rxd);
              }
            }
          }
        }
      }
    }
  }
  rxd.clear();
  rxd.reac.push_back("HO2"); rxd.reac.push_back(namRO2);
  rxd.prod.push_back("O2"); rxd.prod.push_back(namRO2H);
  rxd.rate = "Habs/by-RO2/HO2"; rxd.isRevrs = true;
  lst.push_back(rxd);
  rxd.clear();
  rxd.reac.push_back("H2O2"); rxd.reac.push_back(namRO2);
  rxd.prod.push_back("HO2"); rxd.prod.push_back(namRO2H);
  rxd.rate = "Habs/by-RO2/H2O2"; rxd.isRevrs = true;
  lst.push_back(rxd);
  outLstClear();
}

// ======================================================= [rxnHabsAlkene]
//   H-atom abstraction from alkene
void reaction::rxnHabsAlkene(molecule &m) {
  int ia, ib, idx, cidx, idTo, idMxP, maxPri = 0;
  string namRct = m.molName, namPrd;
  vector<int> dbnd;
  rxnSiteList::iterator p, pmx;
  reagentMast::iterator prg;
  groupAtom ga;
  molecule prd;
  rxnDescr rxd;

  dbnd = m.molec.getBondMatch(6, 6, 2);
  for (ia = 0; ia < 2; ia++) {
    idx = dbnd[ia]; cidx = dbnd[(ia + 1) % 2];
    for (ib = 0; ib < m.molec.atom[idx].numBonds; ib++) {
      if ((idTo = m.molec.atom[idx].bond[ib].toAtomIndex) != cidx) {
        m.getGroupAtom(idTo, ga);
        if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
          if (p->prior > maxPri)
            { idMxP = idTo; pmx = p; maxPri = p->prior; }
        }
      }
    }
  }
  if (maxPri == 0) { return; }
  prd = m; prd.beRadicalAt(idMxP);
  procProdAutoBase(prd); namPrd = prd.molName;
  for (prg = (*rxLb)["Habs"].rgMst.begin();
       prg != (*rxLb)["Habs"].rgMst.end(); prg++) {
    if (pmx->reag.find(prg->rct) != pmx->reag.end()) {
      rxd.clear();
      rxd.reac.push_back(namRct); rxd.reac.push_back(prg->rct);
      rxd.prod.push_back(namPrd); rxd.prod.push_back(prg->prd);
      rxd.rate = "Habs/by-" + prg->rct + "/" + pmx->snam;
      rxd.isRevrs = prg->isRevrs;
      lst.push_back(rxd);
    }
  }
  outLstClear();
}

// ===================================================== [rxnHabsCycEther]
//   H-atom abstraction from cyclic ethers
void reaction::rxnHabsCycEther(molecule &m) {
  int ia, ib, ibn, ibc, ip, idxO, idx, idxn, cidx, cidxn, cidxnb, numPrds;
  string namPrd[2], namRct = m.molName;
  vector<int> idxCo, idxCCo, cidxCCo;
  set<int> idxDone;
  rxnSiteList::iterator p;
  reagentMast::iterator prg;
  groupAtom ga;
  molecule imPrd, prd[2];
  rxnDescr rxd;

  idxO = m.getIdxAtomGrpMatch("O/C2", *grLb);
  idxCo = m.molec.atom[idxO].getBondToIdcs();
  for (ib = 0; ib < 2; ib++) {
    idx = idxCo[ib]; cidx = idxCo[(ib + 1) % 2]; m.getGroupAtom(idx, ga);
    if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
      imPrd = m; imPrd.molec.removeHatom(idx);
      imPrd.molec.removeBond(idxO, cidx);
      imPrd.molec.incBondOrder(idx, idxO);
      imPrd.isCyclic = false;              // Mar. 7, 2010
      // imPrd.beUnique();                 // Mar. 7, 2010
      numPrds = molDisKetR(imPrd, prd[0], prd[1]);
      for (ip = 0; ip < numPrds; ip++)
        { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
      if (lst.procDupProdPart(numPrds, namPrd) == false) {
        for (prg = (*rxLb)["Habs"].rgMst.begin();
             prg != (*rxLb)["Habs"].rgMst.end(); prg++) {
          if (p->reag.find(prg->rct) != p->reag.end()) {
            rxd.clear();
            rxd.reac.push_back(namRct); rxd.reac.push_back(prg->rct);
            rxd.prod.push_back(prg->prd);
            for (ip = 0; ip < numPrds; ip++)
              { rxd.prod.push_back(namPrd[ip]); }
            rxd.rate = "Habs/by-" + prg->rct + "/" + p->snam;
            rxd.isRevrs = prg->isRevrs;
            lst.push_back(rxd);
          }
        }
      }
    }
  }
  if (lst.size() > 0) { outLstClear(); return; }
  // - for the case no CH bond found in O-neighbor (July 30, 2007)
  idxDone.clear();
  for (ib = 0; ib < 2; ib++) {
    idx = idxCo[ib]; cidx = idxCo[(ib + 1) % 2];
    idxCCo = m.molec.atom[idx].getBondToIdcs();
    for (ibn = 0; ibn < idxCCo.size(); ibn++) {
      idxn = idxCCo[ibn];
      if ( (idxn != idxO) && (idxn != cidx) &&
           (idxDone.find(idxn) == idxDone.end()) ) {
        m.getGroupAtom(idxn, ga);
        if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
          idxDone.insert(idxn);
          imPrd = m; imPrd.molec.removeHatom(idxn);
          imPrd.molec.removeBond(idx, idxO);
          imPrd.molec.incBondOrder(idx, idxn);
          cidxCCo = m.molec.atom[cidx].getBondToIdcs();
          cidxnb = -1;
          for (ibc = 0; ibc < cidxCCo.size(); ibc++) {
            cidxn = cidxCCo[ibc];
            if (cidxn != idxO) {
              if (m.isAtomInRing(cidxn)) { cidxnb = cidxn; break; }
            }
          }
          imPrd.isCyclic = false;
          numPrds = molDissoc(imPrd, cidx, cidxnb, prd[0], prd[1], true);
          for (ip = 0; ip < numPrds; ip++)
            { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
          if (lst.procDupProdPart(numPrds, namPrd) == false) {
            for (prg = (*rxLb)["Habs"].rgMst.begin();
                 prg != (*rxLb)["Habs"].rgMst.end(); prg++) {
              if ((prg->rct == "OH") || (prg->rct == "HO2")) {
                rxd.clear();
                rxd.reac.push_back(namRct); rxd.reac.push_back(prg->rct);
                rxd.prod.push_back(prg->prd);
                for (ip = 0; ip < numPrds; ip++)
                  { rxd.prod.push_back(namPrd[ip]); }
                rxd.rate = "Habs/by-" + prg->rct + "/" + p->snam;
                rxd.isRevrs = prg->isRevrs;
                lst.push_back(rxd);
              }
            }
          }
        }
      }
    }
  }
  outLstClear();
}

// ======================================================= [rxnHabsKetAld]
//   H-atom abstraction from ketones and aldehydes (H-C=O, vinoxy pos.)

void reaction::rxnHabsKetAld(molecule &m) {
  int ib, iCx, iSt, idxCo, idxCnx[2], idxSite[2];
  int numCnx = 0, numSites = 0;
  string namRct = m.molName, namPrd;
  rxnSiteList::iterator p;
  reagentMast::iterator prg;
  groupAtom ga;
  molecule prd;
  rxnDescr rxd;

  idxCo = m.molec.getIdxAtomMatch(6, 8, 2);
  for (ib = 0; ib < m.molec.atom[idxCo].numBonds; ib++) {
    if (m.molec.atom[idxCo].bond[ib].order == 1) {
      idxCnx[numCnx] = m.molec.atom[idxCo].bond[ib].toAtomIndex; numCnx++;
    }
  }
  if (m.molec.atom[idxCo].getNumHatoms() > 0)
    { idxSite[numSites] = idxCo; numSites++; }
  for (iCx = 0; iCx < numCnx; iCx++) {
    if (m.molec.atom[idxCnx[iCx]].getNumHatoms() > 0)
      { idxSite[numSites] = idxCnx[iCx]; numSites++; }
  }
  for (iSt = 0; iSt < numSites; iSt++) {
    m.getGroupAtom(idxSite[iSt], ga);
    if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
      prd = m; prd.beRadicalAt(idxSite[iSt]);
      procProdAutoBase(prd); namPrd = prd.molName;
      if (lst.procDupProd(namPrd) == false) {
        for (prg = (*rxLb)["Habs"].rgMst.begin();
             prg != (*rxLb)["Habs"].rgMst.end(); prg++) {
          if (p->reag.find(prg->rct) != p->reag.end()) {
            rxd.clear();
            rxd.reac.push_back(namRct); rxd.reac.push_back(prg->rct);
            rxd.prod.push_back(namPrd); rxd.prod.push_back(prg->prd);
            rxd.rate = "Habs/by-" + prg->rct + "/" + p->snam;
            rxd.isRevrs = prg->isRevrs;
            lst.push_back(rxd);
          }
        }
      }
    }
  }
  outLstClear();
}

// ====================================================== [rxnHabsAldOnly]
//   H-atom abstraction from aldehydes (H-C=O only)

void reaction::rxnHabsAldOnly(molecule &m) {
  int idxOc, idxCo;
  string namRct = m.molName, namPrd;
  rxnSiteList::iterator p;
  reagentMast::iterator prg;
  groupAtom ga;
  molecule prd;
  rxnDescr rxd;

  idxOc = m.molec.getIdxAtomTermMatchNumH(8, 6, 2, 1);
  idxCo = m.molec.getIdxNextAtom(idxOc, 6);
  if (m.molec.atom[idxCo].getNumHatoms() < 1) { return; }

  m.getGroupAtom(idxCo, ga);
  if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
    prd = m; prd.beRadicalAt(idxCo);
    procProdAutoBase(prd); namPrd = prd.molName;
    for (prg = (*rxLb)["Habs"].rgMst.begin();
         prg != (*rxLb)["Habs"].rgMst.end(); prg++) {
      if (p->reag.find(prg->rct) != p->reag.end()) {
        rxd.clear();
        rxd.reac.push_back(namRct); rxd.reac.push_back(prg->rct);
        rxd.prod.push_back(namPrd); rxd.prod.push_back(prg->prd);
        rxd.rate = "Habs/by-" + prg->rct + "/" + p->snam;
        rxd.isRevrs = prg->isRevrs;
        lst.push_back(rxd);
      }
    }
  }
  outLstClear();
}

// ========================================================= [rxnOabsNewR]
//   O-atom abstraction from HO2 and CH3OO by newly added R

void reaction::rxnOabsNewR(molecule &r) {
  int pos;
  string namPrd, namRct = r.molName;
  molecule prod, RO2t, ROt;
  rxnDescr rxd;
  string rate, namRO2t, namROt, prfNam = "", basNam;
  molecList::iterator pmr;

  if (r.isAllyl()) { rate = "Oabs/by-Allyl/RO2"; }
  else { rate = "Oabs/by-R/RO2"; }
  if ((pos = r.molName.find(r.getTypeFormula(*grLb, *tyMol))) != string::npos) {
    if (pos > 0) { prfNam = r.molName.substr(0, pos); }
  }
  molAddRcO(r, prod);
// procProdAutoBase(prod); namPrd = prod.molName;
   basNam = prfNam + prod.getTypeFormula(*grLb, *tyMol);
   procProd(prod, basNam); namPrd = prod.molName;
//
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("HO2");
  rxd.prod.push_back(namPrd); rxd.prod.push_back("OH");
  rxd.rate = rate;
  rxd.isRevrs = false;
  lst.push_back(rxd);
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("CH3OO");
  rxd.prod.push_back(namPrd); rxd.prod.push_back("CH3O");
  rxd.rate = rate;
  rxd.isRevrs = false;
  lst.push_back(rxd);
  if (checkIsRadFuel(r)) {
    for (pmr = ro2Fuel.begin(); pmr != ro2Fuel.end(); pmr++) {
      if (spAd->getMolStat(pmr->molec, namRO2t) == 3) {
        RO2t.molec = pmr->molec; RO2t.molName = namRO2t;
        molRmvRcO(RO2t, ROt);
        procProdAutoBase(ROt); namROt = ROt.molName;
        rxd.clear();
        rxd.reac.push_back(namRct); rxd.reac.push_back(namRO2t);
        rxd.prod.push_back(namPrd); rxd.prod.push_back(namROt);
        rxd.rate = rate;
        rxd.isRevrs = false;
        lst.push_back(rxd);
      }
    }
  }
  outLstClear();
}

// ======================================================= [rxnOabsNewRO2]
//   O-atom abstraction from newly added RO2 by 'radFuel' and CH3

void reaction::rxnOabsNewRO2(molecule &ro2) {
  string namRO2 = ro2.molName, namRO, namRt, namRtO;
  molecule prdRO, rctRt, prdRtO;
  molecList::iterator pm;
  rxnDescr rxd;

  molRmvRcO(ro2, prdRO);
  procProdAutoBase(prdRO); namRO = prdRO.molName;
  if (checkIsRo2Fuel(ro2)) {
    for (pm = radFuel.begin(); pm != radFuel.end(); pm++) {
      namRt = pm->name; rctRt.setFrom(pm->molec, pm->name);
      molAddRcO(rctRt, prdRtO);
      procProdAutoBase(prdRtO); namRtO = prdRtO.molName;
      rxd.clear();
      rxd.reac.push_back(namRO2); rxd.reac.push_back(namRt);
      rxd.prod.push_back(namRO); rxd.prod.push_back(namRtO);
      rxd.rate = "Oabs/by-R/RO2";
      rxd.isRevrs = false;
      lst.push_back(rxd);
    }
  }
  rxd.clear();                         // ----- RO2 + CH3
  rxd.reac.push_back(namRO2); rxd.reac.push_back("CH3");
  rxd.prod.push_back(namRO); rxd.prod.push_back("CH3O");
  rxd.rate = "Oabs/by-R/RO2";
  rxd.isRevrs = false;
  lst.push_back(rxd);
//  rxd.clear();                         // ----- RO2 + Allyl
//  rxd.reac.push_back(namRO2); rxd.reac.push_back("C3H5-A");
//  rxd.prod.push_back(namRO); rxd.prod.push_back("C3H5O");
//  rxd.rate = "Oabs/by-Allyl/RO2";
//  rxd.isRevrs = false;
//  lst.push_back(rxd);
  outLstClear();
}

// ====================================================== [rxnDecmpNonRad]
//   Decomposition reactions for non-radical

void reaction::rxnDecmpNonRad(molecule &m) {
  int ia, ib, toIdx, bo, prior, unknown, maxPri = 0, priThres, numPrds;
  int ip, foundCH3;
  string namPrd[2], namRct = m.molName;
  groupAtom ga;
  rxnSiteList::iterator p;
  listBonds dcmpBnds;       // here member 'order' is used for priority
  listBonds::iterator pb;
  molecule prd[2];
  rxnDescr rxd;

  dcmpBnds.clear();
  for (ia = 0; ia < m.molec.numAtoms; ia++) {
    for (ib = 0; ib < m.molec.atom[ia].numBonds; ib++) {
      toIdx = m.molec.atom[ia].bond[ib].toAtomIndex;
      bo = m.molec.atom[ia].bond[ib].order;
      if ((bo == 1) && (toIdx > ia)) {
        unknown = false; prior = 0;
        m.molec.getDissocGroup(ia, toIdx, ga);
        if ((p = (*rxLb)["Decomp"].find(ga)) != (*rxLb)["Decomp"].end())
          { prior += p->prior; } else { unknown = true; }
        m.molec.getDissocGroup(toIdx, ia, ga);
        if ((p = (*rxLb)["Decomp"].find(ga)) != (*rxLb)["Decomp"].end())
          { prior += p->prior; } else { unknown = true; }
        if (!unknown) {
          dcmpBnds.addBond(chemBond(prior, ia, toIdx));
          if (prior > maxPri) { maxPri = prior; }
        }
      }
    }
  }
  if (dcmpBnds.size() == 0) { return; }
  priThres = maxPri - (*rxLb)["Decomp"].priorLim;
  for (pb = dcmpBnds.begin(); pb != dcmpBnds.end(); pb++) {
    if (pb->order >= priThres) {
      numPrds =
       molDissoc(m, pb->idxFirst, pb->idxSecond, prd[0], prd[1], false);
      foundCH3 = false;
      for (ip = 0; ip < numPrds; ip++) {
        procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName;
        if (namPrd[ip] == "CH3") { foundCH3 = true; }
      }
      if (numPrds > 0) {
        if (lst.procDupProd(numPrds, namPrd) == false) {
          rxd.clear();
          rxd.reac.push_back(namRct);
          for (ip = 0; ip < numPrds; ip++)
            { rxd.prod.push_back(namPrd[ip]); }
          if (numPrds == 1) { rxd.rate = "Dcmp/CycC-C"; }
          else if (foundCH3) { rxd.rate = "Rv/Dcmp/C-Me"; }
          else { rxd.rate = "Rv/Dcmp/C-C"; }
          rxd.isRevrs = true;
          lst.push_back(rxd);
        }
      }
    }
  }
  outLstClear();
}


// ====================================================== [rxnDecmpCHnonR]
//   C-H Decomposition reactions for non-radical

void reaction::rxnDecmpCHnonR(molecule &m) {
  int pos, ia, maxPri = 0, priThres;
  string prfNam = "", basNam, namPrd, namRct = m.molName, siteName;
  groupAtom ga;
  rxnSiteList::iterator p;
  molecule prod;
  rxnDescr rxd;

  if ((pos = m.molName.find(m.molec.getChemFormula())) != string::npos)
    { if (pos > 0) { prfNam = m.molName.substr(0, pos); } }
  for (ia = 0; ia < m.molec.numAtoms; ++ia) {
    m.getGroupAtom(ia, ga);
    if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
      if (p->prior > maxPri) { maxPri = p->prior; }
    }
  }
  priThres = maxPri - (*rxLb)["Habs"].priorLim;
  for (ia = 0; ia < m.molec.numAtoms; ++ia) {
    m.getGroupAtom(ia, ga);
    if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
      if (p->prior >= priThres) {
        prod = m; prod.beRadicalAt(ia);
        basNam = prfNam + prod.molec.getChemFormula();
        procProd(prod, basNam); namPrd = prod.molName;
        if (lst.procDupProd(namPrd) == false) {
          siteName = p->snam;
          rxd.clear();
          rxd.reac.push_back(namRct);
          rxd.prod.push_back(namPrd);
          rxd.prod.push_back("H");
          rxd.rate = "Rv/Dcmp/C-H";
          rxd.isRevrs = true;
          if (siteName.length() == 5) {
            if (siteName.substr(0, 4) == "Ally")
              { rxd.rate = "Rv/Dcmp/allyC-H"; }
          }
          lst.push_back(rxd);
        }
      }
    }
  }
  outLstClear();
}

// ========================================================= [rxnDecmpRad]
//   Decomposition reactions of radicals

void reaction::rxnDecmpRad(molecule &m) {
  int idxRCnt, anRCnt, toIdx, bo, toIdxNxt, boNxt, prior, maxPri = 0;
  int sumBO, toAN, toAA, numH, priThres, numPrds, anNxt, priAdd;
  int ib, ibn, ip, nSubDB, nH1st, nHrc;
  string namPrd[2], namRct = m.molName, ieStr;
  rxnSiteList::iterator p;
  groupAtom ga;
  listBonds dcmpBnds;       // here member 'order' is used for priority
  listBonds::iterator pb;
  molecule prd[2];
  rxnDescr rxd;

  dcmpBnds.clear();
  idxRCnt = m.molec.getIdxRadCent();
  anRCnt = m.molec.atom[idxRCnt].atomicNumber;
  for (ib = 0; ib < m.molec.atom[idxRCnt].numBonds; ib++) {
    toIdx = m.molec.atom[idxRCnt].bond[ib].toAtomIndex;
    bo = m.molec.atom[idxRCnt].bond[ib].order;
    anNxt = m.molec.atom[toIdx].atomicNumber;
    priAdd = 0; if (anNxt == 8) { priAdd = 60; }
    if (bo < 3) {
      sumBO = 0;
      for (ibn = 0; ibn < m.molec.atom[toIdx].numBonds; ibn++) {
        toIdxNxt = m.molec.atom[toIdx].bond[ibn].toAtomIndex;
        boNxt = m.molec.atom[toIdx].bond[ibn].order;
        sumBO += boNxt;
        if ((toIdxNxt != idxRCnt) && (boNxt == 1)) {
          m.molec.getDissocGroup(toIdxNxt, toIdx, ga);
          if ((p = (*rxLb)["Decomp"].find(ga)) != (*rxLb)["Decomp"].end()) {
            prior = p->prior + 10 * (2 - bo) + priAdd;
            // - July 11, 2010: in-ring bond deserves high priority
            if (m.isCyclic) {
              if (m.isAtomInRing(toIdx) && m.isAtomInRing(toIdxNxt))
                { prior += 100; }
            }
            //
            dcmpBnds.addBond(chemBond(prior, toIdx, toIdxNxt));
            if (prior > maxPri) { maxPri = prior; }
          }
        }
      }
      toAN = m.molec.atom[toIdx].atomicNumber;
      toAA = m.molec.atom[toIdx].atomAttrib;
      numH = elmt[toAN].valence - sumBO - toAA;
      if (numH > 0) {
        prior = 70 + 10 * (2 - bo) + priAdd;
        dcmpBnds.addBond(chemBond(prior, toIdx, -1));
        if (prior > maxPri) { maxPri = prior; }
      }
    }
  }
  if (dcmpBnds.size() == 0) { return; }
  priThres = maxPri - (*rxLb)["Decomp"].priorLim;
  for (pb = dcmpBnds.begin(); pb != dcmpBnds.end(); pb++) {
    if (pb->order >= priThres) {
      numPrds =
        molDissoc(m, pb->idxFirst, pb->idxSecond, prd[0], prd[1], true);
      for (ip = 0; ip < numPrds; ip++) {
        procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName;
      }
      if (numPrds > 0) {
        if (lst.procDupProd(numPrds, namPrd) == false) {
          rxd.clear();
          rxd.isRevrs = true;
          rxd.reac.push_back(namRct);
          for (ip = 0; ip < numPrds; ip++)
            { rxd.prod.push_back(namPrd[ip]); }
          if (anRCnt == 8) {                       // === RO. (alkoxy) ===
            rxd.rate = "Dcmp/RO";
          } else if (anRCnt == 6) {                // === R. (alkyl) =====
            if (numPrds == 1) {                // ----- cyc. radical C-C
              rxd.rate = "Dcmp/Cyc.C-C";
            } else if (pb->idxSecond < 0) {    // --- C-H/O-H fission
              if (m.molec.atom[pb->idxFirst].numBonds > 1) {
                rxd.rate = "Rv/Dcmp/.CC(-H)C";
              } else {
                if (m.molec.atom[pb->idxFirst].atomicNumber == 8)
                  { rxd.rate = "Rv/Dcmp/.CO-H"; }
                else { rxd.rate = "Rv/Dcmp/.CC-H"; }
              }
            } else {
              if (m.molec.atom[pb->idxSecond].atomicNumber == 8) {
                                               // --- C-O/O-O fission
                if (m.molec.atom[pb->idxFirst].atomicNumber == 8)
                  { rxd.rate = "Dcmp/.CO-OH"; rxd.isRevrs = false; }
                else if (m.molec.atom[pb->idxFirst].atomicNumber == 6) {
                  // - Mar 19, 2010
                  nH1st = m.molec.atom[pb->idxFirst].getNumHatoms();
                  nHrc = m.molec.atom[idxRCnt].getNumHatoms();
                  if (nH1st < 2) { ieStr = "i"; } else { ieStr = "e"; }
                  nSubDB = 4 - nH1st - nHrc;
                  // rxd.rate = "Rv/Dcmp/.CC-OH";
                  rxd.rate = "Rv/Dcmp/.CCOH/" + ieStr + format(nSubDB);
                }
              } else {                         // --- C-C/O-C fission
                if (m.molec.atom[pb->idxFirst].numBonds > 2) {
                  rxd.rate = "Rv/Dcmp/.CC(-C)C";
                } else {
                  if (m.molec.atom[pb->idxFirst].atomicNumber == 8)
                    { rxd.rate = "Rv/Dcmp/.CO-C"; }
                  else { rxd.rate = "Rv/Dcmp/.CC-C"; }
                }
              }
            }
          } else { rxd.rate = "unknown"; }         // === unknow case ====
          lst.push_back(rxd);
        }
      }
    }
  }
  outLstClear();
}

void reaction::rxnDecmpCycEther(molecule &m) {     // --- rxnDecmpCycEther
  int ic, ib, idxO, idx, nBnd, toIdx, prior, maxPri = 0, mxpIdx, mxpToIdx;
  int numPrds, idxRc, idxCoNonR, ip;
  vector<int> idxCo;
  string namPrd[2], namRct = m.molName;
  rxnSiteList::iterator p;
  groupAtom ga;
  molecule prd[2];
  rxnDescr rxd;

  idxO = m.getIdxAtomGrpMatch("O/C2", *grLb);
  idxCo = m.molec.atom[idxO].getBondToIdcs();
  for (ic = 0; ic < 2; ic++) {
    idx = idxCo[ic]; nBnd = m.molec.atom[idx].numBonds;
    for (ib = 0; ib < nBnd; ib++) {
      toIdx = m.molec.atom[idx].bond[ib].toAtomIndex;
      if (!m.isAtomInRing(toIdx)) {
        m.molec.getDissocGroup(toIdx, idx, ga);
        if ((p = (*rxLb)["Decomp"].find(ga)) != (*rxLb)["Decomp"].end())
          { prior = p->prior; } else { prior = 10; }
        if (prior > maxPri) { maxPri = prior; mxpIdx = idx, mxpToIdx = toIdx; }
      }
    }
  }
  if (maxPri == 0) { return; }
  numPrds = molDissoc(m, mxpToIdx, mxpIdx, prd[0], prd[1], false);
  idxO = prd[1].getIdxAtomGrpMatch("O/C2", *grLb);
  idxCo = prd[1].molec.atom[idxO].getBondToIdcs();
  idxRc = prd[1].molec.getIdxRadCent();
  idxCoNonR = idxCo[0];
  if (idxCoNonR == idxRc) { idxCoNonR = idxCo[1]; }
  prd[1].molec.removeBond(idxO, idxCoNonR);
  prd[1].molec.incBondOrder(idxRc, idxO);
  prd[1].isCyclic = false;
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct);
  for (ip = 0; ip < numPrds; ip++)
    { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Rv/Dcmp/C-C";
  rxd.isRevrs = true;
  lst.push_back(rxd);
  outLstClear();
}

/* --- combined with decompKetOOH (Sept. 28, 2010)
                                                       // --- rxnDecmpSpec
void reaction::rxnDecmpSpec(molecule &m, string bspc) {
  int ia, ib, ip, idxO1 = -1, idxO2 = -1, found = false, numPrds;
  int nHAtoms = m.molec.numAtoms;
  // - original dislit model for RO2
// double dA0 = 76.385, dA1 = -279.25;
  // temporary reduction (ROOH-LPL) - 0711d
// double dA0 = 77.,    dA1 = -310.;
  // 0927b model
   double dA0 = 76.,    dA1 = -310.;
  double db0 = -19.6735, db1 = 92.158;
  double dE0 = -12769., dE1 = 93550.;
  // original
// double shNHA = 3., factNHA = 1.;
  // 0927b model
   double shNHA = 1.5, factNHA = 0.81;
  string namPrd[2], namRct = m.molName;
  molecule prd[2];
  rxnDescr rxd;

  if (bspc == "O-O") {
    for (ia = 0; ia < m.molec.numAtoms; ia++) {
      if (m.molec.atom[ia].atomicNumber == 8) {
        for (ib = 0; ib < m.molec.atom[ia].numBonds; ib++) {
          if (m.molec.atom[ia].bond[ib].toAtomicNbr == 8) {
            idxO1 = ia; idxO2 = m.molec.atom[ia].bond[ib].toAtomIndex;
            found = true; break;
          }
        }
        if (found) { break; }
      }
    }
    if (!found) { return; }
    numPrds = molDissoc(m, idxO1, idxO2, prd[0], prd[1], false);
    if (numPrds < 2) { return; }
    for (ip = 0; ip < numPrds; ip++) {
      procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName;
    }
    rxd.clear();
    // register reverse way
    rxd.prod.push_back(namRct);
    for (ip = 0; ip < numPrds; ip++)
      { rxd.reac.push_back(namPrd[ip]); }
    rxd.rate = "Troe/Rcmb/RO-OH";
    rxd.isRevrs = true;
    rxd.modTroe = true;
    rxd.modAL = pow(10., dA0 + dA1 / (nHAtoms + shNHA) * factNHA);
    rxd.modbL = db0 + db1 / (nHAtoms + shNHA) * factNHA;
    rxd.modEaL = dE0 + dE1 / (nHAtoms + shNHA) * factNHA;
    rxd.modT3Fc = 50.; rxd.modT1Fc = 50000.;
    switch (nHAtoms) {
      case 4: rxd.modaFc = 0.32; rxd.modT2Fc = 10000.; break;
      case 5: rxd.modaFc = 0.17; rxd.modT2Fc = 4000.; break;
      case 6: rxd.modaFc = 0.15; rxd.modT2Fc = 3000.; break;
      case 7: rxd.modaFc = 0.; rxd.modT2Fc = 1200.; break;
      default: rxd.modaFc = 0.; rxd.modT2Fc = 1000. - (nHAtoms - 8.) * 100.;
        if (rxd.modT2Fc < 200.) { rxd.modT2Fc = 200.; }
        break;
    }
    lst.push_back(rxd);
  } else { return; }
  outLstClear();
}
*/

void reaction::rxnDecmpQaOOH(molecule &m) {           // --- rxnDecmpQaOOH
  vector<molecInfo> rstr;
  int numResStr = m.molec.getResonStr(rstr), ir;
  int cNRxnGen = numRxnGen;
  string namRct = m.molName;
  molecule im;
  for (ir = 0; ir < numResStr; ir++) {
    im.setFrom(rstr[ir], namRct); rxnDecmpQOOH(im);
    if (numRxnGen > cNRxnGen) { return; }
  }
}

// ======================================================== [rxnDecmpQOOH]
//   Decomposition of QOOH

void reaction::rxnDecmpQOOH(molecule &m) {
  static const long bitCyc = 1L << 27, mskCyc = ~bitCyc;
  long mtyp;
  string typCm, rateBeta;
  int ia, ib, ip, ir, dst, idCO, idOc, idOh, idRC, imIdRC, imIdOc, imIdOh;
  int idOr, idCor, idCr, idAlt, dstc, idOoh;
  int n1Prds, n2Prds, numPrds = 0;
//  int nNQCO, nNQRC, nNQtot;
  int nQCring, nTCring, jc;
  string inRingCs;
  string namRct = m.molName, namPrd[3], snRC, gnCO, snCO;
  vector<int> rt, iRC, rtc;
  groupAtom gaCO, gat;
  rxnSiteList::iterator p;
  molecule prd1[2], prd2[2], imPrd, prd[3];
  rxnDescr rxd;

  mtyp = m.getType(*grLb);
  typCm = tyMol->typName(mtyp & mskCyc);
  idOoh = m.molec.getIdxAtomTermMatch(8, 8, 1);
  idOc = m.molec.getIdxNextAtom(idOoh, 8, 1);
  idCO = m.molec.getIdxNextAtom(idOc, 6, 1);
  idRC = m.molec.getIdxRadCent();
  dst = m.molec.getDstRoute(idOc, idRC, rt);
  m.getGroupAtom(idCO, gaCO); gnCO = grLb->main.findName(gaCO);
  if ((gnCO == "C/C3/O") || (gnCO == "C/CD/C2/O")) { snCO = "t"; }
  else if ((gnCO == "C/C2/H/O") || (gnCO == "C/CD/C/H/O")) { snCO = "s"; }
  else { snCO = "p"; }
  m.getGroupAtom(idRC, gat);
  if ((p = (*rxLb)["Isom"].find(gat)) != (*rxLb)["Isom"].end()) {
    snRC = p->snam.substr(0, 1);
    if ((snRC == "A") || (snRC == "O")) { snRC = p->snam.substr(0, 2); }
  } else { snRC = "p"; }
  if (typCm == "QaOOH") { rateBeta = "Dcmp/QaOOH/=/C=O/OH"; }
  else { rateBeta = "Dcmp/QOOH/=/=O/OH" + snCO + snRC; }
  // --------------------------------------------------- bete scission
  if (dst == 3) {
    n1Prds = molDissoc(m, rt[2], rt[1], prd1[0], prd1[1], true);
    if (n1Prds == 2) { imPrd = prd1[1]; } else { imPrd = prd1[0]; }
    imIdRC = imPrd.molec.getIdxRadCent();
    for (ib = 0; ib < imPrd.molec.atom[imIdRC].numBonds; ib++) {
      if (imPrd.molec.atom[imIdRC].bond[ib].toAtomicNbr == 8)
        { imIdOc = imPrd.molec.atom[imIdRC].bond[ib].toAtomIndex; break; }
    }
    for (ib = 0; ib < imPrd.molec.atom[imIdOc].numBonds; ib++) {
      if (imPrd.molec.atom[imIdOc].bond[ib].toAtomicNbr == 8)
        { imIdOh = imPrd.molec.atom[imIdOc].bond[ib].toAtomIndex; break; }
    }
    n2Prds = molDissoc(imPrd, imIdOc, imIdOh, prd2[0], prd2[1], true);
    numPrds = 0;
    for (ip = 0; ip < n1Prds - 1; ip++)
      { prd[numPrds] = prd1[ip]; numPrds++; }
    for (ip = 0; ip < n2Prds; ip++)
      { prd[numPrds] = prd2[ip]; numPrds++; }
    for (ip = 0; ip < numPrds; ip++) {
      procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName;
    }
    rxd.clear();
    rxd.reac.push_back(namRct);
    for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
    rxd.rate = rateBeta;
    rxd.isRevrs = false;
    // --- Apr. 29, 2010 - disabled Sept. 22, 2010
    // rxd.shiftB = -(*rxTp)["gamQOOH/DbAdj"].shiftE;
    // rxd.factA = pow(1. / 500., rxd.shiftB);
    // ---
    lst.push_back(rxd);
  }
  // ----------------------------------------------- cyclic ether + OH
  // (ketone + OH for cyclic QOOH)
  for (ib = 0; ib < m.molec.atom[idOc].numBonds; ib++) {
    if (m.molec.atom[idOc].bond[ib].toAtomicNbr == 8)
      { idOh = m.molec.atom[idOc].bond[ib].toAtomIndex; break; }
  }
  // - count only in-ring quaternary carbons bonded to idRC and idCO
  // nNQCO = 0; nNQRC = 0;
  // if (m.molec.isQuatC(rt[2])) { nNQCO = 1; }
  // if (m.molec.isQuatC(rt[dst - 1])) { nNQRC = 1; }
  // nNQtot = nNQCO + nNQRC;
  // -
  // - count number of in-ring quaternary and tertiary carbons
  nQCring = 0; nTCring = 0; inRingCs = "";
  for (jc = 2; jc < dst; jc++) {
    if (m.molec.isQuatCincdO(rt[jc])) { nQCring++; inRingCs += "Q"; }
    else if (m.molec.isTertCincdO(rt[jc])) { nTCring++; inRingCs += "T"; }
    else { inRingCs += "S"; }
  }
  // -
  if ((dst >= 2) && (dst <= 5)) {
    numPrds = molDissoc(m, idOc, idOh, prd[0], prd[1], false);
    prd[0].beUnique();     // added 22-Mar-2010
    prd[0].molec.getIdcsRadCents(iRC);
    if (prd[0].isCyclic) {        // ----- avoid generating cyclic ethers
      for (ir = 0; ir < 2; ir++) {   // -- generate carbonyls instead
        if (prd[0].molec.atom[iRC[ir]].atomicNumber == 8) { idOr = iRC[ir]; }
        else { idCr = iRC[ir]; }
      }
      dstc = prd[0].molec.getDstRoute(idOr, idCr, rtc);
      idCor = rtc[1];
      if (prd[0].molec.atom[idCor].getNumHatoms() > 0) {
        prd[0].molec.atom[idCr].atomAttrib--;    // - ketone
        prd[0].molec.atom[idCor].atomAttrib++;
        prd[0].molec.incBondOrder(idOr, idCor);
      } else if (prd[0].molec.atom[idCr].getNumHatoms() > 0) {
        prd[0].molec.makeNewBond(idOr, idCr);
        prd[0].molec.removeBond(idOr, idCor);
        prd[0].molec.atom[idCor].atomAttrib--;
        prd[0].molec.atom[idCr].atomAttrib++;
        prd[0].molec.incBondOrder(idOr, idCr);
      } else {
        idAlt = -1;
        for (ia = 0; ia < prd[0].molec.numAtoms; ia++) {
          if ( (prd[0].molec.atom[ia].atomicNumber == 6) &&
               (prd[0].molec.atom[ia].getNumHatoms() > 1) ) {
            idAlt = ia; break;
          }
        }
        if (idAlt > -1) {
          prd[0].molec.atom[idAlt].atomAttrib++;
          prd[0].molec.atom[idCr].atomAttrib--;
          prd[0].molec.makeNewBond(idOr, idAlt);
          prd[0].molec.removeBond(idOr, idCor);
          prd[0].molec.atom[idCor].atomAttrib--;
          prd[0].molec.atom[idAlt].atomAttrib++;
          prd[0].molec.incBondOrder(idOr, idAlt);
        }
      }
    } else {
      prd[0].molec.makeNewBond(iRC[0], iRC[1]);
    }
    prd[0].beUnique();
    for (ip = 0; ip < numPrds; ip++)
      { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
    rxd.clear();
    rxd.reac.push_back(namRct);
    for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
    if (snRC[0] == 'A') {
      rxd.rate = "Dcmp/QOOH/cyc" + format(dst + 1) + "QO" + snRC;
    } else if ((dst == 3) && (nQCring > 0)) {
      rxd.rate = "Dcmp/QOOH/cyc" + format(dst + 1) + "QO" + snCO + "Q" + snRC;
    } else if ((dst == 3) && (nTCring > 0)) {
      rxd.rate = "Dcmp/QOOH/cyc" + format(dst + 1) + "QO" + snCO + "T" + snRC;
    } else if ((dst == 4) && (nQCring > 0)) {
      for (jc = 0; jc < 2; jc++) {
        if (inRingCs[jc] == 'T') { inRingCs[jc] = 'S'; }
      }
      if (inRingCs == "QQ") { inRingCs = "SQ"; }
      rxd.rate = "Dcmp/QOOH/cyc" + format(dst + 1) + "QO" + snCO + inRingCs + snRC;
    } else {
      rxd.rate = "Dcmp/QOOH/cyc" + format(dst + 1) + "QO" + snCO + snRC;
    }
    rxd.isRevrs = false;
    // --- Apr. 29, 2010 - disable Sept. 22, 2010
    // switch (dst) {
    //   case 2: rxd.shiftB = -(*rxTp)["betQOOH/RbAdj"].shiftE; break;
    //   case 3: rxd.shiftB = -(*rxTp)["gamQOOH/RbAdj"].shiftE; break;
    //   default: rxd.shiftB = -(*rxTp)["delQOOH/RbAdj"].shiftE; break;
    // }
    // rxd.factA = pow(1. / 500., rxd.shiftB);
    // ---
    // - Sept. 22, 2010
    // rxd.shiftE = (*rxTp)["quatNeighRC"].shiftE * nNQtot;
    // rxd.factA = pow((*rxTp)["quatNeighRC"].factA, nNQtot);
    // -
    // = Oct. 20, 2010 - correction for 5-membered ring oxolane formation
    // if ((dst == 4) && (nQCring > 0)) {
    //   rxd.shiftE = (*rxTp)["quat5QOring"].shiftE;
    //   rxd.factA = (*rxTp)["quat5QOring"].factA;
    // }
    // =
    lst.push_back(rxd);
  }
  outLstClear();
}

                                                       // --- rxnDecmpROOH
void reaction::rxnDecmpROOH(molecule &m, double ofsNHA) {
  //   Decomposition reactions of ROOH incld. hydroperoxy carbonyl compounds
  int ip, idOh, idOoh, numPrds, idOr, idCor, idOc, idCo, idCcor;
  int nHAtoms = m.molec.numAtoms;
  // - original dislit model for RO2
// double dA0 = 76.385, dA1 = -279.25;
  // - 0711d, 1018b, 1208b-
// double dA0 = 77.,    dA1 = -310.;   // 20110107m2
   double dA0 = (*rxTp)["ROOH/dA0/dA1"].factA, dA1 = (*rxTp)["ROOH/dA0/dA1"].shiftE;
  // original
   double db0 = -19.6735, db1 = 92.158;
   double dE0 = -12769., dE1 = 93550.;
   double shNHA = 3., factNHA = 1.;
  //
  string namRct = m.molName, namPrd[3];
  vector<int> rt;
  molecule prd[3], im;
  rxnDescr rxd;

  idOh = m.molec.getIdxAtomTermMatch(8, 8, 1);
  idOoh = m.molec.getIdxNextAtom(idOh, 8);
// May 7, 2010 - changed to step-wise decomposition
/*
  molDissoc(m, idOh, idOoh, prd[0], im, false);   // --- dissociate OH
  if (m.isCyclic) { im.beUnique(); }
  idOr = im.molec.getIdxRadCent();
  idCor = im.molec.getIdxNextAtom(idOr, 6);
  idOc = im.molec.getIdxAtomTermMatch(8, 6, 2);
  idCo = im.molec.getIdxNextAtom(idOc, 6);
  im.molec.getDstRoute(idCor, idCo, rt); idCcor = rt[1];
  numPrds = molDissoc(im, idCor, idCcor, prd[1], prd[2], true);
  numPrds++;
*/
// new
  numPrds = molDissoc(m, idOh, idOoh, prd[0], prd[1], false);
//
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  // register reverse way
  rxd.prod.push_back(namRct);
  for (ip = 0; ip < numPrds; ip++) { rxd.reac.push_back(namPrd[ip]); }
//  rxd.rate = "Dcmp/ketOOH";
//  rxd.rate = "Rv/Dcmp/ketOOH";

  rxd.isRevrs = true;

  if ((*rxTp)["ROOH/FallOff"].factA > 0.) {
    rxd.rate = "Troe/Rcmb/RO-OH";
    rxd.modTroe = true;
    rxd.modAL = pow(10., dA0 + dA1 / (nHAtoms + ofsNHA + shNHA) * factNHA);
    rxd.modbL = db0 + db1 / (nHAtoms + ofsNHA + shNHA) * factNHA;
    rxd.modEaL = dE0 + dE1 / (nHAtoms + ofsNHA + shNHA) * factNHA;
    rxd.modT3Fc = 50.; rxd.modT1Fc = 50000.;
    if ((*rxTp)["Fc/model/20110107"].factA > 0.) {
      // - 20110107m2 (carate Fc)
      switch (nHAtoms) {
        case 4: rxd.modaFc = 0.32; rxd.modT2Fc = 10000.; break;
        case 5: rxd.modaFc = 0.17; rxd.modT2Fc = 4000.; break;
        case 6: rxd.modaFc = 0.15; rxd.modT2Fc = 3000.; break;
        case 7: rxd.modaFc = 0.; rxd.modT2Fc = 1200.; break;
        default: rxd.modaFc = 0.; rxd.modT2Fc = 1000. - (nHAtoms - 8.) * 100.;
          if (rxd.modT2Fc < 200.) { rxd.modT2Fc = 200.; }
          break;
      }
    } else {
      // - 20110423 (dislit Fc)
      rxd.modaFc = 0.1; rxd.modT2Fc = 2200.;
      if (nHAtoms == 4) { rxd.modaFc = 0.24; }
    }
  } else {
    rxd.rate = "Rcmb/RO-OH";
  }
  lst.push_back(rxd);
  outLstClear();
}

// ====================================================== [rxnDecmpKetAld]
//   Decomposition of ketones and aldehydes (only R...C=O)

void reaction::rxnDecmpKetAld(molecule &m) {
  int ib, iCh, ip, idxCo, idxCnx[2], numCnx = 0, numPrds, isOCO, idCo, idOh;
  string namRct = m.molName, namPrd[3];
  molecule prd[3], im;
  rxnDescr rxd;

  idxCo = m.molec.getIdxAtomMatch(6, 8, 2);
  if (m.isCyclic) {
    if (m.uniqueRepr.chain[0].isIdxInChain(idxCo)) { return; }
  }
  for (ib = 0; ib < m.molec.atom[idxCo].numBonds; ib++) {
    if (m.molec.atom[idxCo].bond[ib].order == 1) {
      idxCnx[numCnx] = m.molec.atom[idxCo].bond[ib].toAtomIndex; numCnx++;
    }
  }
  for (iCh = 0; iCh < numCnx; iCh++) {
    numPrds = molDissoc(m, idxCnx[iCh], idxCo, prd[0], prd[1], false);
    if (numPrds == 2) {
      isOCO = false;
      im = prd[1];
      idCo = im.molec.getIdxAtomMatch(6, 8, 2);
      for (ib = 0; ib < im.molec.atom[idCo].numBonds; ib++) {
        if (im.molec.atom[idCo].bond[ib].order == 1) {
          if (im.molec.atom[idCo].bond[ib].toAtomicNbr == 8) {
            isOCO = true; idOh = im.molec.atom[idCo].bond[ib].toAtomIndex;
            break;
          }
        }
      }
      if (isOCO) {
        molDissoc(im, idCo, idOh, prd[1], prd[2], false);
        numPrds = 3;
      }
    }
    for (ip = 0; ip < numPrds; ip++)
      { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
    if (lst.procDupProd(numPrds, namPrd) == false) {
      rxd.clear();
      rxd.reac.push_back(namRct);
      for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
      rxd.rate = "Rv/Dcmp/C-C";
      if (numPrds > 1) {
        if ((namPrd[1] == "CH3") || (namPrd[0] == "HCO"))
          { rxd.rate = "Rv/Dcmp/C-Me"; }
      }
      rxd.isRevrs = false; lst.push_back(rxd);
    }
  }
  outLstClear();
}

// ========================================================= [rxnDecmpRCO]
//   Decomposition of acyl radical (RCO -> R + CO)

void reaction::rxnDecmpRCO(molecule &m) {
  int ip, idxCo, idxCco, nSubs, numPrds;
  string namRct = m.molName, namPrd[2], snam;
  molecule prd[2];
  rxnDescr rxd;

  idxCo = m.molec.getIdxRadCent();
  if ((idxCco = m.molec.getIdxNextAtom(idxCo, 6)) < 0) { return; }
  nSubs = m.molec.atom[idxCco].numBonds - 1;
  numPrds = molDissoc(m, idxCco, idxCo, prd[0], prd[1], false);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct);
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  switch (nSubs) {
   case 3: snam = "t"; break;
   case 2: snam = "s"; break;
   default: snam = "p"; break;
  }
  rxd.rate = "Rv/Dcmp/RCO" + snam;
  rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

// ======================================================== [rxnDecmpOCCV]
//   Decomposition of O=C.-C=C (O=C=C-C.)

void reaction::rxnDecmpOCCV(molecule &m) {
  int idRC, idCo, idCd, ip, numPrds;
  string namRct = m.molName, namPrd[2];
  molecule prd[2], im;
  rxnDescr rxd;

  im = m;
  idRC = im.molec.getIdxRadCent();
  idCo = im.molec.getIdxAtomMatch(6, 8, 2);
  idCd = im.molec.getIdxNextAtom(idCo, 6);
  im.molec.decBondOrder(idCo, idCd);
  im.molec.incBondOrder(idCd, idRC);  // --- now O=C.-C=C form

  numPrds = molDissoc(im, idCd, idCo, prd[0], prd[1], false);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct);
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Rv/Dcmp/OCCV";
  rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

// ====================================================== [rxnDecmpVinoxy]
//   Decomposition reactions of Vinoxy radicals

void reaction::rxnDecmpVinoxy(molecule &m) {
  int ib, ip, idxRC, idxCo, idxOc, cidx, nBnds, idxTo, numPrds, cidxAn = 0;
  string namRct = m.molName, namPrd[2], bndBr;
  molecule prd[2];
  rxnDescr rxd;

  idxRC = m.molec.getIdxRadCent();
  idxCo = m.molec.getIdxNextAtom(idxRC, 6, 2, 8);
  idxOc = m.molec.getIdxNextAtom(idxCo, 8, 2);
  nBnds = m.molec.atom[idxCo].numBonds;
  if (nBnds < 3) { cidx = -1; }   // C-H fission
  else {                          // C-C fission
    for (ib = 0; ib < nBnds; ib++) {
      idxTo = m.molec.atom[idxCo].bond[ib].toAtomIndex;
      if ((idxTo != idxOc) && (idxTo != idxRC)) { cidx = idxTo; break; }
    }
  }
  if (cidx >= 0) { cidxAn = m.molec.atom[cidx].atomicNumber; }
  numPrds = molDissoc(m, idxCo, cidx, prd[0], prd[1], true);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct);
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  if (cidx < 0) { bndBr = "-H"; } else { bndBr = "-C"; }
  rxd.rate = "Rv/Dcmp/.CC(" + bndBr + ")C";
  if (cidxAn == 8) { rxd.rate = "Rv/Dcmp/.CC-OH"; }
  rxd.isRevrs = true;
  lst.push_back(rxd);
  outLstClear();
}

// ======================================================== [rxnDecmpKetR]
//   Decomposition reactions of RCOQ/HCOQ (keto-alkyl)

void reaction::rxnDecmpKetR(molecule &m) {
  int ib, ip, idxRC, idxCo, idxCb0, idxCb1, dstRCCo, nBndCb0, numPrds;
  int numCyMTS, numHCo;
  string namRct = m.molName, namPrd[2], sn;
  vector<int> rt;
  rxnSiteList::iterator p;
  groupAtom ga;
  molecule prd[2], im;
  rxnDescr rxd;

  idxRC = m.molec.getIdxRadCent();
  idxCo = m.molec.getIdxAtomMatch(6, 8, 2);
  dstRCCo = m.molec.getDstRoute(idxRC, idxCo, rt);
  numHCo = m.molec.atom[idxCo].getNumHatoms();
    // ----- beta scission
  idxCb0 = rt[1]; idxCb1 = rt[2];
  nBndCb0 = m.molec.atom[idxCb0].numBonds;
  numPrds = molDissoc(m, idxCb0, idxCb1, prd[0], prd[1], true);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct);
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  if (nBndCb0 < 3) { rxd.rate = "Rv/Dcmp/.CC-C"; }
  else { rxd.rate = "Rv/Dcmp/.CC(-C)C"; }
  rxd.isRevrs = false;
  lst.push_back(rxd);
    // ----- isomerization followed by CO elimination
  numCyMTS = dstRCCo + 2;
  if ((numCyMTS > 4) && (numCyMTS < 7) && (numHCo > 0)) {
    m.getGroupAtom(idxRC, ga);
    if ((p = (*rxLb)["Isom"].find(ga)) != (*rxLb)["Isom"].end()) {
      sn = p->snam.substr(0, 1);
      im = m;
      im.molec.atom[idxRC].atomAttrib--;
      im.molec.atom[idxCo].atomAttrib++;
      im.beUnique();
      idxCo = im.molec.getIdxRadCent();
      for (ib = 0; ib < im.molec.atom[idxCo].numBonds; ib++) {
        if (im.molec.atom[idxCo].bond[ib].toAtomicNbr == 6)
          { idxCb0 = im.molec.atom[idxCo].bond[ib].toAtomIndex; break; }
      }
      numPrds = molDissoc(im, idxCo, idxCb0, prd[0], prd[1], false);
      for (ip = 0; ip < numPrds; ip++)
        { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
      rxd.clear();
      rxd.reac.push_back(namRct);
      for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
      rxd.rate = "Isom/R/" + format(numCyMTS) + sn + "t";
      rxd.isRevrs = false;
      lst.push_back(rxd);
    }
  }
  outLstClear();
}

// ========================================================== [rxnIsomRad]
//   Isomerization reactions of radicals (R or ROO)

void reaction::rxnIsomRad(molecule &m) {
  int idRC, priRgt, ia, dst, priTgt, prior, maxPri = 0, priThres, idTg;
  int isRO2 = false, isCycRO2 = false, idCO, inRingIsom, nRingMbr, idCoo;
  int sizeTSring, dstInRing, dstCC, semiInRngIsm, posUS;
  // double shiftEallyl = 0.;
  string snamRgt, inRgt, snamTgt, inTgt, namPrd[1], namRct = m.molName;
  string snCO, gnCO;
  groupAtom gar, gat, gaCO;
  rxnSiteList::iterator pRsl;
  vector<int> rt;
  map<int, int> tgtPri;
  map<int, int>::iterator pm;
  molecule prd;
  rxnDescr rxd;

  idRC = m.molec.getIdxRadCent();
  m.getGroupAtom(idRC, gar);
  if ((pRsl = (*rxLb)["Isom"].find(gar)) == (*rxLb)["Isom"].end())
    { return; }
  priRgt = pRsl->prior; snamRgt = pRsl->snam;
  inRgt = snamRgt.substr(0, 1);
  if (inRgt == "o") {
    isRO2 = true; idCO = m.molec.getIdxAtomMatch(6, 8, 1);
    m.getGroupAtom(idCO, gaCO); gnCO = grLb->main.findName(gaCO);
    if ((gnCO == "C/C3/O") || (gnCO == "C/CD/C2/O")) { snCO = "t"; }
    else if ((gnCO == "C/C2/H/O") || (gnCO == "C/CD/C/H/O")) { snCO = "s"; }
    else { snCO = "p"; }
    // if ((gnCO == "C/CD/C2/O") || (gnCO == "C/CD/C/H/O") || (gnCO == "C/CD/H2/O"))
    //   { shiftEallyl = (*rxTp)["AllylOO/Isom"].shiftE; }
    if (m.isCyclic) { isCycRO2 = true; }
  }
  tgtPri.clear();
  for (ia = 0; ia < m.molec.numAtoms; ia++) {
    if (ia != idRC) {
      m.getGroupAtom(ia, gat);
      if ((pRsl = (*rxLb)["Isom"].find(gat)) != (*rxLb)["Isom"].end()) {
        sizeTSring = m.molec.getDistance(idRC, ia) + 2;
        // --- RO2(5-7membRing) R(5-6membRing)
        if (sizeTSring >= 5) {
          // --- changed to 8
          // if ( (isRO2 && (sizeTSring <= 7)) ||
          if ( (isRO2 && (sizeTSring <= 8)) ||
          // ---
               ((!isRO2) && (sizeTSring <= 6))) {
            priTgt = pRsl->prior;
            if (isRO2 || ((!isRO2) && (priTgt >= priRgt))) {
              prior = priTgt + (2 - abs(sizeTSring - 6)) * 30;
              if (prior > maxPri) { maxPri = prior; }
              tgtPri.insert(make_pair(ia, prior));
            }
          }
        }
      }
    }
  }
  if (tgtPri.size() == 0) { return; }
  priThres = maxPri - (*rxLb)["Isom"].priorLim;
  for (pm = tgtPri.begin(); pm != tgtPri.end(); pm++) {
    if (pm->second >= priThres) {
      idTg = pm->first;
      m.getGroupAtom(idTg, gat);
      pRsl = (*rxLb)["Isom"].find(gat); snamTgt = pRsl->snam;
      posUS = snamTgt.find('_'); inTgt = snamTgt.substr(0, posUS);
      dst = m.getDstInRingRoute(idRC, idTg, dstInRing, rt);
      sizeTSring = dst + 2;
      if (isRO2) { dstCC = dst - 2; } else { dstCC = dst; }
      inRingIsom = false; semiInRngIsm = false;
      if (m.isCyclic) {
        nRingMbr = m.uniqueRepr.chain[0].numAtoms;
        if (dstInRing == dstCC) { inRingIsom = true; }
        else if ((dstInRing > 0) && (dstInRing == (dstCC - 1)))
          { semiInRngIsm = true; }
      }
      prd = m; prd.molName = "";
      prd.molec.atom[idRC].atomAttrib--;
      prd.molec.atom[idTg].atomAttrib++;
      prd.beUnique();
      if ( (isRO2 || ((!isRO2) && (!inRingIsom) && (!semiInRngIsm))) &&
           (prd.molec != m.molec) ) {
        procProdAutoBase(prd); namPrd[0] = prd.molName;
        if (rsvLst.findDupRevs(namRct, namPrd[0]) == false) {
          if (lst.procDupProd(1, namPrd) == false) {
            rxd.clear();
            rxd.reac.push_back(namRct);
            rxd.prod.push_back(namPrd[0]);
            if (isRO2) {
              rxd.rate = "Isom/RO2/" + format(sizeTSring) + snCO + inTgt;
            } else {
              rxd.rate = "Isom/R/" + format(sizeTSring) + inRgt + inTgt;
            }
            rxd.isRevrs = true;
            // rxd.shiftE = shiftEallyl;
            if (inRingIsom) {    // --- in-ring ROO isomerization
              switch (nRingMbr) {
               case 5:
                if (sizeTSring == 5) {
                  rxd.factA = (*rxTp)["InRing/cPent/5"].factA;
                  rxd.shiftE = (*rxTp)["InRing/cPent/5"].shiftE;
                }
                else if (sizeTSring == 6) {
                  rxd.factA = (*rxTp)["InRing/cPent/6"].factA;
                  rxd.shiftE = (*rxTp)["InRing/cPent/6"].shiftE;
                }
                break;
               case 6:
                if (sizeTSring == 5) {
                  rxd.factA = (*rxTp)["InRing/cHex/5"].factA;
                  rxd.shiftE = (*rxTp)["InRing/cHex/5"].shiftE;
                }
                else if (sizeTSring == 6) {
                  rxd.factA = (*rxTp)["InRing/cHex/6"].factA;
                  rxd.shiftE = (*rxTp)["InRing/cHex/6"].shiftE;
                }
                else if (sizeTSring == 7) {
                  rxd.factA = (*rxTp)["InRing/cHex/7"].factA;
                  rxd.shiftE = (*rxTp)["InRing/cHex/7"].shiftE;
                }
                break;
               case 7:
                if (sizeTSring == 5) {
                  rxd.factA = (*rxTp)["InRing/cHep/5"].factA;
                  rxd.shiftE = (*rxTp)["InRing/cHep/5"].shiftE;
                }
                else if (sizeTSring == 6) {
                  rxd.factA = (*rxTp)["InRing/cHep/6"].factA;
                  rxd.shiftE = (*rxTp)["InRing/cHep/6"].shiftE;
                }
                else if (sizeTSring == 7) {
                  rxd.factA = (*rxTp)["InRing/cHep/7"].factA;
                  rxd.shiftE = (*rxTp)["InRing/cHep/7"].shiftE;
                }
                break;
               default:
                if (sizeTSring == 5) {
                  rxd.factA = (*rxTp)["InRing/Other/5"].factA;
                  rxd.shiftE = (*rxTp)["InRing/Other/5"].shiftE;
                }
                else if (sizeTSring == 6) {
                  rxd.factA = (*rxTp)["InRing/Other/6"].factA;
                  rxd.shiftE = (*rxTp)["InRing/Other/6"].shiftE;
                }
                else if (sizeTSring == 7) {
                  rxd.factA = (*rxTp)["InRing/Other/7"].factA;
                  rxd.shiftE = (*rxTp)["InRing/Other/7"].shiftE;
                }
                break;
              }
            } else if (semiInRngIsm) {  // --- semi-in-ring
              if (isCycRO2) {
                switch (nRingMbr) {
                 case 5:
                  if (sizeTSring == 5) {
                    rxd.factA = (*rxTp)["SemiIR/cPent/5"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cPent/5"].shiftE;
                  }
                  else if (sizeTSring == 6) {
                    rxd.factA = (*rxTp)["SemiIR/cPent/6"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cPent/6"].shiftE;
                  }
                  else if (sizeTSring == 7) {
                    rxd.factA = (*rxTp)["SemiIR/cPent/7"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cPent/7"].shiftE;
                  }
                  break;
                 case 6:
                  if (sizeTSring == 5) {
                    rxd.factA = (*rxTp)["SemiIR/cHex/5"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cHex/5"].shiftE;
                  }
                  else if (sizeTSring == 6) {
                    rxd.factA = (*rxTp)["SemiIR/cHex/6"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cHex/6"].shiftE;
                  }
                  else if (sizeTSring == 7) {
                    rxd.factA = (*rxTp)["SemiIR/cHex/7"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cHex/7"].shiftE;
                  }
                  break;
                 case 7:
                  if (sizeTSring == 5) {
                    rxd.factA = (*rxTp)["SemiIR/cHep/5"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cHep/5"].shiftE;
                  }
                  else if (sizeTSring == 6) {
                    rxd.factA = (*rxTp)["SemiIR/cHep/6"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cHep/6"].shiftE;
                  }
                  else if (sizeTSring == 7) {
                    rxd.factA = (*rxTp)["SemiIR/cHep/7"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/cHep/7"].shiftE;
                  }
                  break;
                 default:
                  if (sizeTSring == 5) {
                    rxd.factA = (*rxTp)["SemiIR/Other/5"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/Other/5"].shiftE;
                  }
                  else if (sizeTSring == 6) {
                    rxd.factA = (*rxTp)["SemiIR/Other/6"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/Other/6"].shiftE;
                  }
                  else if (sizeTSring == 7) {
                    rxd.factA = (*rxTp)["SemiIR/Other/7"].factA;
                    rxd.shiftE = (*rxTp)["SemiIR/Other/7"].shiftE;
                  }
                  break;
                }
              }
            }
            lst.push_back(rxd);
          }
        }
      }
    }
  }
  outLstResrv();
}

// ======================================================= [rxnIsomOOQOOH]
//   Isomerization reactions of OOQOOH

void reaction::rxnIsomOOQOOH(molecule &m) {
  int ip, idRC, idOh, idOor, idOoh, idCoor, idCooh, idMoveOOH, id2nd, idT;
  int dst, ib, idnx, sizeTSring, pri, maxPri = 0, idxMx, posUS, nH;
  int idCcHO2, numPrds, idHabs;
  vector<int> rt;
  // double shiftEallyl = 0.;
  string namRct = m.molName, namPrd[2], gn, namSite, gnCoor, snCoor, sn, snT;
  string snCcHO2, sn3rd, sn4th, snHabs;
  groupAtom ga, gaCoor, gaCcHO2, ga3rd, ga4th;
  rxnSiteList::iterator pRsl, pCcHO2;
  molecule im, prd[2];
  rxnDescr rxd;

  idRC = m.molec.getIdxRadCent();
  idOor = m.molec.getIdxNextAtom(idRC, 8);
  idCoor = m.molec.getIdxNextAtom(idOor, 6);
  idOh = m.molec.getIdxAtomTermMatch(8, 0, 8, 1);
  idOoh = m.molec.getIdxNextAtom(idOh, 8);
  idCooh = m.molec.getIdxNextAtom(idOoh, 6);
  dst = m.molec.getDstRoute(idCoor, idCooh, rt);
  if ((dst < 1) || (dst > 5)) { return; }
  m.getGroupAtom(idCooh, ga); gn = grLb->main.findName(ga);
  if (gn == "C/C/H2/O") { namSite = "p"; }
  else if (gn == "C/C2/H/O") { namSite = "s"; }
  else if (gn == "C/C3/O") { namSite = "t"; }
  else if (gn == "C/CD/H2/O") { namSite = "Ap"; }
  else if (gn == "C/CD/C/H/O") { namSite = "As"; }
  else if (gn == "C/CD/C2/O") { namSite = "At"; }
  else { namSite = ""; }
  m.getGroupAtom(idCoor, gaCoor); gnCoor = grLb->main.findName(gaCoor);
  if ((gnCoor == "C/C3/O") || (gnCoor == "C/CD/C2/O")) { snCoor = "t"; }
  else if ((gnCoor == "C/C2/H/O") || (gnCoor == "C/CD/C/H/O")) { snCoor = "s"; }
  else { snCoor = "p"; }
  // if ((gnCoor == "C/CD/C2/O") || (gnCoor == "C/CD/C/H/O") || (gnCoor == "C/CD/H2/O"))
  //   { shiftEallyl = (*rxTp)["AllylOO/Isom"].shiftE; }
  if (dst == 3) {
    m.getGroupAtom(rt[2], ga3rd); gn = grLb->main.findName(ga3rd);
    if (gn =="C/C2/H2") { sn3rd = "s"; }
    else if (gn =="C/C3/H") { sn3rd = "t"; }
    else { sn3rd = "q"; }
  } else if (dst == 4) {
    m.getGroupAtom(rt[3], ga4th); gn = grLb->main.findName(ga4th);
    if (gn =="C/C2/H2") { sn4th = "s"; }
    else if (gn =="C/C3/H") { sn4th = "t"; }
    else { sn4th = "q"; }
  }
  if ((namSite == "t") || (namSite == "At")) {
    if ((snCoor != "t") && (dst < 5)) {
      // --- H-transfer from OOH
      prd[0] = m;
      prd[0].molec.atom[idRC].atomAttrib = 0;
      prd[0].molec.atom[idOh].atomAttrib = 1;
      prd[0].beUnique();
      procProdAutoBase(prd[0]); namPrd[0] = prd[0].molName;
      rxd.clear();
      rxd.reac.push_back(namRct);
      rxd.prod.push_back(namPrd[0]);
      // rxd.rate = "Isom/O2QOOH/OHO" + format(dst + 6);
      rxd.rate = "Isom/O2QOOH/OHO" + format(dst + 6) + snCoor + "t";
      rxd.isRevrs = true;
      lst.push_back(rxd);
    }
    if (dst == 4) {
      if (sn4th != "q") {
        if (sn4th == "t") { idMoveOOH = m.nearestCH2noO(idCooh); }
        else { idMoveOOH = rt[3]; }
        im = m;
        im.moveOOH(idCooh, idMoveOOH, idOoh);
        im.molec.atom[idRC].atomAttrib--;
        im.molec.atom[idMoveOOH].atomAttrib++;
        molDissoc(im, idOoh, idOh, prd[0], prd[1], true);
        prd[0].beUnique();
        for (ip = 0; ip < 2; ip++)
          { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
        rxd.clear();
        rxd.reac.push_back(namRct);
        for (ip = 0; ip < 2; ip++) { rxd.prod.push_back(namPrd[ip]); }
        rxd.rate = "[Isom/RO2/7" + snCoor + sn4th + "]";
        rxd.isRevrs = false;
        lst.push_back(rxd);
      }
    } else if (dst == 3) {
      if (sn3rd != "q") {
        if (sn3rd == "t") { idMoveOOH = m.nearestCH2noO(idCooh); }
        else { idMoveOOH = rt[2]; }
        im = m;
        im.moveOOH(idCooh, idMoveOOH, idOoh);
        im.molec.atom[idRC].atomAttrib--;
        im.molec.atom[idMoveOOH].atomAttrib++;
        molDissoc(im, idOoh, idOh, prd[0], prd[1], true);
        prd[0].beUnique();
        for (ip = 0; ip < 2; ip++)
          { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
        rxd.clear();
        rxd.reac.push_back(namRct);
        for (ip = 0; ip < 2; ip++) { rxd.prod.push_back(namPrd[ip]); }
        rxd.rate = "[Isom/RO2/6" + snCoor + sn3rd + "]";
        rxd.isRevrs = false;
        lst.push_back(rxd);
      }
    } else if (dst == 2) {
      id2nd = rt[1]; maxPri = 0;
      // look for possible delta position
      for (ib = 0; ib < m.molec.atom[idCooh].numBonds; ib++) {
        if (((idT = m.molec.atom[idCooh].bond[ib].toAtomIndex) != id2nd) &&
            (m.molec.atom[idCooh].bond[ib].toAtomicNbr == 6)) {
          if ((nH = m.molec.atom[idT].getNumHatoms()) > 0) {
            pri = 4 - nH;
            if (pri > maxPri) { idHabs = idT; maxPri = pri; }
          }
        }
      }
      if (maxPri > 0) {
        if (maxPri == 3) { snHabs = "t"; }
        else if (maxPri == 2) { snHabs = "s"; }
        else { snHabs = "p"; }
        if (snHabs == "t") { idMoveOOH = m.nearestCH2noO(idHabs); }
        else { idMoveOOH = idHabs; }
        im = m;
        im.moveOOH(idCooh, idMoveOOH, idOoh);
        im.molec.atom[idRC].atomAttrib--;
        im.molec.atom[idMoveOOH].atomAttrib++;
        molDissoc(im, idOoh, idOh, prd[0], prd[1], true);
        prd[0].beUnique();
        for (ip = 0; ip < 2; ip++)
          { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
        rxd.clear();
        rxd.reac.push_back(namRct);
        for (ip = 0; ip < 2; ip++) { rxd.prod.push_back(namPrd[ip]); }
        rxd.rate = "[Isom/RO2/7" + snCoor + snHabs + "]";
        rxd.isRevrs = false;
        lst.push_back(rxd);
      }
    }
  } else if (namSite != "") {
    // --- transfer H
    im = m;
    im.molec.atom[idRC].atomAttrib--;
    im.molec.atom[idCooh].atomAttrib++;
    // --- eliminate OH
    molDissoc(im, idOoh, idOh, prd[0], prd[1], true);
    prd[0].beUnique();
    for (ip = 0; ip < 2; ip++)
      { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
    rxd.clear();
    rxd.reac.push_back(namRct);
    for (ip = 0; ip < 2; ip++) { rxd.prod.push_back(namPrd[ip]); }
    if (((dst + 4) < 9) && (namSite[0] != 'A')) {
      // --- *h4-delta + *h3 (Nov. 16, 2010)
      if ((dst == 4) && (sn4th != "q")) {
        rxd.rate = "[Isom/O2QOOH/" + format(dst + 4) + snCoor + namSite + "]+[Isom/RO2/7" + snCoor + sn4th + "]";
      } else if ((dst == 3) && (sn3rd != "q")) {
        rxd.rate = "[Isom/O2QOOH/" + format(dst + 4) + snCoor + namSite + "]+[Isom/RO2/6" + snCoor + sn3rd + "]";
      } else {
        rxd.rate = "Isom/O2QOOH/" + format(dst + 4) + snCoor + namSite;
      }
    } else {
      rxd.rate = "Isom/O2QOOH/" + format(dst + 4) + namSite;
    }
    rxd.isRevrs = false;
    // rxd.shiftE = shiftEallyl;
    lst.push_back(rxd);
  }

/*
  // concerted HO2 elimination
  idCcHO2 = rt[1];
  if (m.molec.atom[idCcHO2].getNumHatoms() > 0) {
    m.getGroupAtom(idCcHO2, gaCcHO2);
    if ((pCcHO2 = (*rxLb)["Isom"].find(gaCcHO2)) != (*rxLb)["Isom"].end()) {
      posUS = pCcHO2->snam.find('_');
      snCcHO2 = pCcHO2->snam.substr(0, posUS);
      im = m;
      im.molec.atom[idCcHO2].atomAttrib++;
      im.molec.atom[idRC].atomAttrib--;
      numPrds = molDissoc(im, idCoor, idOor, prd[0], prd[1], true);
      for (ip = 0; ip < numPrds; ip++)
        { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
      rxd.clear();
      rxd.reac.push_back(namRct);
      for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
      rxd.rate = "Elim/HO2/RO2/" + snCoor + snCcHO2;
      rxd.isRevrs = false;
      lst.push_back(rxd);
    }
  }
*/

  if (lst.size() > 0) { outLstClear(); return; }

/*
  // - for the case no reaction was generated above
  for (ib = 0; ib < m.molec.atom[idCooh].numBonds; ib++) {
    idnx = m.molec.atom[idCooh].bond[ib].toAtomIndex;
    if (idnx != idOoh) {
      if (m.molec.atom[idnx].getNumHatoms() > 0) {
        dst = m.molec.getDistance(idCoor, idnx);
        if ((dst > 0) && (dst < 5)) {
          m.getGroupAtom(idnx, ga);
          if ((pRsl = (*rxLb)["Isom"].find(ga)) != (*rxLb)["Isom"].end()) {
            pri = pRsl->prior;
            if (pri > maxPri) {
              idxMx = idnx; maxPri = pri; snT = pRsl->snam;
              posUS = snT.find('_'); sn = snT.substr(0, posUS);
              sizeTSring = dst + 4;
            }
          }
        }
      }
    }
  }
  if (maxPri > 0) {
    im = m; im.isUnique = false; im.molName = "";
    im.molec.atom[idRC].atomAttrib--;
    im.molec.atom[idxMx].atomAttrib++;
    molDissoc(im, idCooh, idOoh, prd[0], prd[1], true);
    for (ip = 0; ip < 2; ip++)
      { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
    rxd.clear();
    rxd.reac.push_back(namRct);
    for (ip = 0; ip < 2; ip++) { rxd.prod.push_back(namPrd[ip]); }
    rxd.rate = "Isom/RO2/" + format(sizeTSring) + snCoor + sn;
    rxd.isRevrs = false;
    // rxd.shiftE = shiftEallyl;
    lst.push_back(rxd);
  }
  if (lst.size() > 0) { outLstClear(); return; }
*/

}

// ====================================================== [rxnRecmbQOOHO2]
//   recombination reactions of QOOH with O2

void reaction::rxnRecmbQOOHO2(molecule &m) {
//
//  - The OOQOOH will be generated even though there may not be exit,
//    since they will eventually consumed by back decomposition.
//
//  int idCO;
//  groupAtom ga;
//  idCO = m.molec.getIdxAtomMatch(6, 8, 1);
//  m.getGroupAtom(idCO, ga);
//  if (grLb->main.findName(ga) == "C/C3/O") { return; }
  rxnRecmbRadO2(m);
}

// ======================================================= [rxnRecmbRadO2]
//   recombination reactions of radicals with O2

void reaction::rxnRecmbRadO2(molecule &m) {
  static const long bitCyc = 1L << 27, mskCyc = ~bitCyc;
  long mtyp;
  string typ, typCm, prfNam = "", basNam;
  int pos, idRC, nHAtoms = m.molec.numAtoms + 2;
  int isR_falloff = false, isQOOH_falloff = false;
  double shNHA = 0.;
  // - dislit model
// - original
// double dA0 = 76.385, dA1 = -279.25;
// - 1208b, 1210a
  double dA0 = 68.885, dA1 = -227.25; // now these values are changed below!
// - 1209a
// double dA0 = 69.2,   dA1 = -229.77;
//
  double db0 = -19.6735, db1 = 92.158;
  double dE0 = -12769., dE1 = 93550.;
  //
  // - carate-85 model
  // double dA0 = 567.901, dA1 = -12406.25, db0 = -179.7653, db1 = 4053.621;
  // double dE0 = 89114., dE1 = -2010020.;
  //
  string namSite, namRct = m.molName, namPrd;
  groupAtom ga;
  rxnSiteList::iterator pRsl;
  molecule prd;
  rxnDescr rxd;
  // -- for allyl
  int nRs, iRs;
  vector<molecInfo> rstr;
  molecule rsm;

  mtyp = m.getType(*grLb);
  typ = tyMol->typName(mtyp); typCm = tyMol->typName(mtyp & mskCyc);
  if ((pos = m.molName.find(m.getTypeFormula(*grLb, *tyMol))) != string::npos) {
    if (pos > 0) { prfNam = m.molName.substr(0, pos); }
  }
  if (m.numResStruct > 1) {
    // - allyl radicals
    nRs = m.molec.getResonStr(rstr);
    for (iRs = 0; iRs < nRs; iRs++) {
      rsm.setFrom(rstr[iRs], "tmp");
      idRC = rsm.molec.getIdxRadCent();
      rsm.getGroupAtom(idRC, ga);
      if ((pRsl = (*rxLb)["Recomb"].find(ga)) == (*rxLb)["Recomb"].end())
        { return; }
      namSite = pRsl->snam;
      molAddRcO2(rsm, prd);
//    procProdAutoBase(prd); namPrd = prd.molName;
      basNam = prfNam + prd.getTypeFormula(*grLb, *tyMol);
      procProd(prd, basNam); namPrd = prd.molName;
//
      if (lst.procDupProd(namPrd) == false) {
        if (radFuel.find(namRct) != radFuel.end()) {
          if (ro2Fuel.find(namPrd) == ro2Fuel.end())
            { ro2Fuel.push_back(molecRef(prd, *grLb)); }
        }
        rxd.clear();
        rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
        rxd.prod.push_back(namPrd);
        rxd.rate = "Rcmb/R/O2/" + namSite;
        rxd.isRevrs = true;
        lst.push_back(rxd);
      }
    }
  } else {
    // - alkyl radicals
    idRC = m.molec.getIdxRadCent();
    m.getGroupAtom(idRC, ga);
    if ((pRsl = (*rxLb)["Recomb"].find(ga)) == (*rxLb)["Recomb"].end())
      { return; }
    namSite = pRsl->snam;
    molAddRcO2(m, prd);
//  procProdAutoBase(prd); namPrd = prd.molName;
    basNam = prfNam + prd.getTypeFormula(*grLb, *tyMol);
    procProd(prd, basNam); namPrd = prd.molName;
//
    if (radFuel.find(namRct) != radFuel.end()) {
      if (ro2Fuel.find(namPrd) == ro2Fuel.end())
        { ro2Fuel.push_back(molecRef(prd, *grLb)); }
    }
    if ( ((typCm == "R") || (typCm == "hoR") || (typCm == "Alkenyl")) &&
         ((*rxTp)["R+O2/FallOff"].factA > 0.) ) {
      isR_falloff = true;
      dA0 = (*rxTp)["R+O2/dA0/dA1"].factA;
      dA1 = (*rxTp)["R+O2/dA0/dA1"].shiftE;
    }
    if ( ((typCm == "QOOH") || (typCm == "QdOOH") || (typCm == "hoQOOH")) &&
         ((*rxTp)["QOOH+O2/FallOff"].factA > 0.) ) {
      isQOOH_falloff = true;
      dA0 = (*rxTp)["QOOH+O2/dA0/dA1"].factA;
      dA1 = (*rxTp)["QOOH+O2/dA0/dA1"].shiftE;
      shNHA = (*rxTp)["QOOH+O2/shNA"].factA;
    }
    rxd.clear();
    rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
    rxd.prod.push_back(namPrd);
    rxd.isRevrs = true;
    if (isR_falloff || isQOOH_falloff) {
      rxd.rate = "Troe/Rcmb/R/O2/" + namSite;
      rxd.modTroe = true;
      rxd.modAL = pow(10., dA0 + dA1 / (nHAtoms + 3. + shNHA));
      rxd.modbL = db0 + db1 / (nHAtoms + 3. + shNHA);
      rxd.modEaL = dE0 + dE1 / (nHAtoms + 3. + shNHA);
      rxd.modT3Fc = 50.; rxd.modT1Fc = 50000.;
      // rxd.modAL = pow(10., dA0 + dA1 / (nHAtoms + 20.));
      // rxd.modbL = db0 + db1 / (nHAtoms + 20.);
      // rxd.modEaL = dE0 + dE1 / (nHAtoms + 20.);
      // rxd.modT3Fc = 50.; rxd.modT1Fc = 100000.;
      if ((*rxTp)["Fc/model/20110107"].factA > 0.) {
        // - 20110107m2 (carate Fc)
        switch (nHAtoms) {
          case 4: rxd.modaFc = 0.32; rxd.modT2Fc = 10000.; break;
          case 5: rxd.modaFc = 0.17; rxd.modT2Fc = 4000.; break;
          case 6: rxd.modaFc = 0.15; rxd.modT2Fc = 3000.; break;
          case 7: rxd.modaFc = 0.; rxd.modT2Fc = 1200.; break;
          default: rxd.modaFc = 0.; rxd.modT2Fc = 1000. - (nHAtoms - 8.) * 100.;
            if (rxd.modT2Fc < 200.) { rxd.modT2Fc = 200.; }
            break;
          /* - carate-85 model
          case 4: rxd.modaFc = 0.25; rxd.modT2Fc = 3200.; break;
          case 5: rxd.modaFc = 0.15; rxd.modT2Fc = 2400.; break;
          case 6: rxd.modaFc = 0.10; rxd.modT2Fc = 1200.; break;
          default: rxd.modaFc = 0.; rxd.modT2Fc = 100.; break;
          */
        }
      } else {
        // - 20110423 (dislit Fc)
        rxd.modaFc = 0.1; rxd.modT2Fc = 2200.;
        if (nHAtoms == 4) { rxd.modaFc = 0.24; }
      }
    } else {
      rxd.rate = "Rcmb/R/O2/" + namSite;
    }
    lst.push_back(rxd);
  }
  outLstClear();
}

// ======================================================= [rxnElimHO2RO2]
//   5-center concerted HO2 elimination from RO2

void reaction::rxnElimHO2RO2(molecule &m) {
  int ia, it, ip, idmOr, posUS;
  int idRC, idCO, idTg, idOH, idOC, numPrds, nMainRingMbr = 0;
  int inRingCO, inRingTG;
  int idmOor, idmCO;
  int dst, idxSite, isit;
  double fsit[3] = {0.7, 1.0, 1.2};
  int n6mrs[3] = {0, 0, 0}, ncy6mrs[3] = {0, 0, 0}, nsc6mrs[3] = {0, 0, 0};
  int n7mrs[3] = {0, 0, 0}, ncy7mrs[3] = {0, 0, 0}, nsc7mrs[3] = {0, 0, 0};
  double d6mr = 0., dcy6mr = 0., dsc6mr = 0.;
  double d7mr = 0., dcy7mr = 0., dsc7mr = 0.;
  int sizeTSring, dstInRing, dstCC, inRingIsom, semiInRngIsm;
  double dcnum, dcsite;
  string gnCO, snCO, snTg, namRct = m.molName, namPrd[2];
  vector<int> idTgt, rt;
  groupAtom gaCO, gat;
  rxnSiteList::iterator p;
  molecule im, prd[2];
  rxnDescr rxd;

  if (m.isCyclic) { nMainRingMbr = m.uniqueRepr.chain[0].numAtoms; }
  idmOr = m.molec.getIdxRadCent();
  idmOor = m.molec.getIdxNextAtom(idmOr, 8);
  idmCO = m.molec.getIdxNextAtom(idmOor, 6);
  idTgt.clear();
  for (ia = 0; ia < m.molec.numAtoms; ia++) {
    if (m.molec.atom[ia].atomicNumber == 6) {
      if (m.molec.atom[ia].getNumHatoms() > 0) {
        idxSite = m.molec.atom[ia].numBonds - 1;
        inRingIsom = false; semiInRngIsm = false;
        dst = m.getDstInRing(idmOr, ia, dstInRing);
        sizeTSring = dst + 2; dstCC = dst - 2;
        if (m.isCyclic) {
          if (dstInRing == dstCC) { inRingIsom = true; }
          else if ((dstInRing > 0) && (dstInRing == (dstCC - 1)))
            { semiInRngIsm = true; }
        }
        if (sizeTSring == 5) {
          idTgt.push_back(ia);
        } else if (sizeTSring == 6) {
          if (inRingIsom) { ncy6mrs[idxSite]++; }
          else if (semiInRngIsm) { nsc6mrs[idxSite]++; }
          else { n6mrs[idxSite]++; }
        } else if (sizeTSring == 7) {
          if (inRingIsom) { ncy7mrs[idxSite]++; }
          else if (semiInRngIsm) { nsc7mrs[idxSite]++; }
          else { n7mrs[idxSite]++; }
        }
      }
    }
  }
  if (idTgt.size() == 0) { return; }
  m.molec.getDstRoute(idmOr, idTgt[0], rt); idCO = rt[2];
  inRingCO = false;
  if (m.isCyclic) {
    if (m.isAtomInRing(idCO)) { inRingCO = true; }
  }
  m.getGroupAtom(idCO, gaCO);
  gnCO = grLb->main.findName(gaCO);
  if (gnCO == "C/C/H2/O") { snCO = "p"; dcsite = 1.2; }
  else if (gnCO == "C/CD/H2/O") { snCO = "Ap"; dcsite = 2.2; }
  else if (gnCO == "C/C2/H/O") { snCO = "s"; dcsite = 1.; }
  else if (gnCO == "C/CD/C/H/O") { snCO = "As"; dcsite = 2.; }
  else if (gnCO == "C/C3/O") { snCO = "t"; dcsite = 0.2; }
  else if (gnCO == "C/CD/C2/O") { snCO = "At"; dcsite = 1.2; }
  else { return; }
  // - Apr. 14, 2010
  for (isit = 0; isit < 3; isit++) {
    d6mr += fsit[isit] * n6mrs[isit]; dcy6mr += fsit[isit] * ncy6mrs[isit];
    dsc6mr += fsit[isit] * nsc6mrs[isit];
    d7mr += fsit[isit] * n7mrs[isit]; dcy7mr += fsit[isit] * ncy7mrs[isit];
    dsc7mr += fsit[isit] * nsc7mrs[isit];
  }
  dcnum = 1.0 * (d6mr + 0.5 * dsc6mr + 0.3 * dcy6mr)
        + 0.3 * (d7mr + 0.45 * dsc7mr + 0.25 * dcy7mr) + 1.0 * dcsite;
  //
  for (it = 0; it < idTgt.size(); it++) {
    idTg = idTgt[it]; m.getGroupAtom(idTg, gat);
    inRingTG = false;
    if (m.isCyclic) {
      if (m.isAtomInRing(idTg)) { inRingTG = true; }
    }
    if ((p = (*rxLb)["Isom"].find(gat)) != (*rxLb)["Isom"].end()) {
      posUS = p->snam.find('_');
      snTg = p->snam.substr(0, posUS);
      im = m; im.molec.atom[idTg].atomAttrib++;
      im.molec.atom[idmOr].atomAttrib--;
      idRC = idTg;
      idOH = idmOr;
      im.molec.getDstRoute(idOH, idRC, rt);
      idOC = rt[1]; idCO = rt[2];
      numPrds = molDissoc(im, idCO, idOC, prd[0], prd[1], true);
      for (ip = 0; ip < numPrds; ip++)
        { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
      if (lst.procDupProd(numPrds, namPrd) == false) {
        rxd.clear();
        rxd.reac.push_back(namRct);
        for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
        rxd.rate = "Elim/HO2/RO2/" + snCO + snTg;
        // -- re-activated 29-Mar-2010
        // - Apr. 14, 2010
        // rxd.shiftE = (*rxTp)["RO2/HO2ElimDF"].shiftE * pow((double)dcnum, (*rxTp)["RO2/HO2ElimDF"].factA);
        // shEval = (*rxTp)["RO2/HO2ElimDF"].shiftE * pow((double)dcnum, (*rxTp)["RO2/HO2ElimDF"].factA);
        // --- disabled Sept. 22, 2010
        // rxd.shiftB = -(*rxTp)["RO2/HO2EbAdj"].shiftE * pow(dcnum, (*rxTp)["RO2/HO2EbAdj"].factA);
        // rxd.factA = pow(1. / 500., rxd.shiftB);
        // ---
        if (m.isCyclic && inRingCO && inRingTG) {
          switch (nMainRingMbr) {
           case 5: rxd.factA *= (*rxTp)["InR/cHO2/cPent"].factA;
                   rxd.shiftE += (*rxTp)["InR/cHO2/cPent"].shiftE; break;
           case 6: rxd.factA *= (*rxTp)["InR/cHO2/cHex"].factA;
                   rxd.shiftE += (*rxTp)["InR/cHO2/cHex"].shiftE; break;
           case 7: rxd.factA *= (*rxTp)["InR/cHO2/cHep"].factA;
                   rxd.shiftE += (*rxTp)["InR/cHO2/cHep"].shiftE; break;
           default: rxd.factA *= (*rxTp)["InR/cHO2/Other"].factA;
                   rxd.shiftE += (*rxTp)["InR/cHO2/Other"].shiftE; break;
          }
        }
        rxd.isRevrs = false;
        lst.push_back(rxd);
      }
    }
  }
  outLstClear();
}

// =========================================================== [rxnRO2RO2]
//   RO2 reaction with CH3OO and ro2Fuel

void reaction::rxnRO2RO2(molecule &ro2) {
  string namRO2 = ro2.molName, namRO, namRtO2, namRtO;
  molecule prdRO, rctRtO2, prdRtO;
  molecList::iterator pm;
  rxnDescr rxd;

  molRmvRcO(ro2, prdRO);
  procProdAutoBase(prdRO); namRO = prdRO.molName;
  if (checkIsRo2Fuel(ro2)) {
    for (pm = ro2Fuel.begin(); pm != ro2Fuel.end(); pm++) {
      namRtO2 = pm->name; rctRtO2.setFrom(pm->molec, pm->name);
      molRmvRcO(rctRtO2, prdRtO);
      procProdAutoBase(prdRtO); namRtO = prdRtO.molName;
      rxd.clear();
      rxd.reac.push_back(namRO2); rxd.reac.push_back(namRtO2);
      rxd.prod.push_back(namRO); rxd.prod.push_back(namRtO);
      rxd.prod.push_back("O2");
      rxd.rate = "Cplx/RO2/RO2"; rxd.isRevrs = false;
      lst.push_back(rxd);
    }
    if ((pm = ro2Fuel.find(namRO2)) != ro2Fuel.end()) { ro2Fuel.erase(pm); }
  }
  rxd.clear();
  rxd.reac.push_back(namRO2); rxd.reac.push_back("CH3OO");
  rxd.prod.push_back(namRO); rxd.prod.push_back("CH3O");
  rxd.prod.push_back("O2");
  rxd.rate = "Cplx/RO2/RO2"; rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

// ===================================================== [rxnAddRadAlkene]
//   OH and HO2 addition to alkene

void reaction::rxnAddRadAlkene(molecule &m) {
  int ip, ia, nSubs;
  string subn, namRct = m.molName, namPrd;
  vector<vector<int> > dblBnds;
  molecule im, prd;
  rxnDescr rxd;

  dblBnds = m.molec.getAllBondsMatch(6, 6, 2);
  if (dblBnds.empty()) { return; }
  for (ip = 0; ip < dblBnds.size(); ip++) {
    nSubs = 0;
    for (ia = 0; ia < 2; ia++)
      { nSubs += m.molec.atom[ dblBnds[ip][ia] ].numBonds - 1; }
    switch (nSubs) {
     case 0: subn = "nosb"; break;
     case 1: subn = "R"; break;
     case 2: subn = "2R"; break;
     default: subn = "more"; break;
    }
    // im is an intermediated with broken double bond
    im = m; im.molec.decBondOrder(dblBnds[ip][0], dblBnds[ip][1]);
    for (ia = 0; ia < 2; ia++) {
      molAddOH(im, prd, dblBnds[ip][ia]);       // - OH addition
      procProdAutoBase(prd); namPrd = prd.molName;
      // - register OH adduct as radFuel
      if (radFuel.find(namPrd) == radFuel.end())
        { radFuel.push_back(molecRef(prd, *grLb)); }
      // - Here only product is registered since reverse reaction is
      //   generated as the subsequent decomposition of adduct.
      molAddHO2(im, prd, dblBnds[ip][ia]);      // - HO2 addition
      procProdAutoBase(prd); namPrd = prd.molName;
      // - will NOT register HO2 adduct as radFuel
      // if (radFuel.find(namPrd) == radFuel.end())
      //   { radFuel.push_back(molecRef(prd, *grLb)); }
      if (lst.procDupProd(namPrd) == false) {
        rxd.clear();
        rxd.reac.push_back(namRct); rxd.reac.push_back("HO2");
        rxd.prod.push_back(namPrd); rxd.rate = "Adtn/HO2/DB/" + subn;
        rxd.isRevrs = true;
        lst.push_back(rxd);
      }
    }
  }
  outLstClear();
}

// ===================================================== [rxnAddHelAlkene]
//   H addition/elimination to alkene

void reaction::rxnAddHelAlkene(molecule &m) {
  int ip, ia, ia2, nb, ib, ibt, idbr1, idbr2, numPrds, iprd;
  string namRct = m.molName, namPrd[2];
  vector<vector<int> > dblBnds;
  molecule im, iim, prd[2];
  rxnDescr rxd;

  dblBnds = m.molec.getAllBondsMatch(6, 6, 2);
  if (dblBnds.empty()) { return; }
  for (ip = 0; ip < dblBnds.size(); ip++) {
    // im is an intermediated with broken double bond
    im = m; im.molec.decBondOrder(dblBnds[ip][0], dblBnds[ip][1]);
    for (ia = 0; ia < 2; ia++) {
      nb = m.molec.atom[ dblBnds[ip][ia] ].numBonds;
      if (nb > 1) {
        ia2 = ia + 1; if (ia2 == 2) { ia2 = 0; }
        ibt = -1;
        for (ib = 0; ib < nb; ib++) {
          if ( (m.molec.atom[ dblBnds[ip][ia] ].bond[ib].toAtomIndex != dblBnds[ip][ia2]) &&
               (m.molec.atom[ dblBnds[ip][ia] ].bond[ib].order == 1) ) {
            ibt = ib;
            idbr1 = dblBnds[ip][ia];
            idbr2 = m.molec.atom[idbr1].bond[ib].toAtomIndex;
            break;
          }
        }
        if (ibt >= 0) {
          iim = im;
            // add - H
          iim.molec.atom[idbr1].atomAttrib--;
            // break bond
          numPrds = molDissoc(iim, idbr1, idbr2, prd[0], prd[1], true);
          for (iprd = 0; iprd < numPrds; iprd++) {
            procProdAutoBase(prd[iprd]); namPrd[iprd] = prd[iprd].molName;
          }
          if (numPrds > 0) {
            if (lst.procDupProd(numPrds, namPrd) == false) {
              rxd.clear();
              rxd.isRevrs = true;
              rxd.reac.push_back(namRct); rxd.reac.push_back("H");
              for (iprd = 0; iprd < numPrds; iprd++) {
                rxd.prod.push_back(namPrd[iprd]);
              }
              rxd.rate = "AdEl/H/DB";
              lst.push_back(rxd);
            }
          }
        }
      }
    }
  }
  outLstClear();
}

// ======================================================= [rxnOaddAlkene]
//   O atom addition to alkene (including dienes) and allene

void reaction::rxnOaddAlkene(molecule &m) {
  static const long bitCyc = 1L << 27, mskCyc = ~bitCyc;
  long mtyp;
  string typCm;
  int ib, ic, ip, minNumBnds = 4, numBnds, idxAdd, idxNoAdd, idxO;
  int numSubs, numPrds, idxTo, prior, maxPri = 0, idxMxP;
  int idxAlleneC, idxToA[2];
  string namRct = m.molName, namPrd[2], ssubs;
  vector<int> dblBnd, idRC;
  rxnSiteList::iterator p;
  groupAtom ga;
  molecule im, prd[2];
  rxnDescr rxd;

  mtyp = m.getType(*grLb);
  typCm = tyMol->typName(mtyp & mskCyc);
  if (typCm == "Allene") {
    idxAlleneC = m.molec.getAlleneCent(idxToA[0], idxToA[1]);
    for (ic = 0; ic < 2; ic++) {     // --- search for =C with min. subst.
      numBnds = m.molec.atom[idxToA[ic]].numBonds;
      if (numBnds < minNumBnds) {
        minNumBnds = numBnds; idxAdd = idxToA[ic];
      }
    }
    idxNoAdd = idxAlleneC;
  } else {
    dblBnd = m.molec.getBondMatch(6, 6, 2);
    if (dblBnd[0] < 0) { return; }
    for (ic = 0; ic < 2; ic++) {     // --- search for =C with min. subst.
      numBnds = m.molec.atom[dblBnd[ic]].numBonds;
      if (numBnds < minNumBnds) {
        minNumBnds = numBnds; idxAdd = dblBnd[ic];
        idxNoAdd = dblBnd[(ic + 1) % 2];
      }
    }
  }
  numSubs = m.molec.atom[idxAdd].numBonds - 1;
  im = m; im.molec.decBondOrder(idxAdd, idxNoAdd);
  idxO = rfmAddOat(im, idxAdd);           // ... O-alkene adduct biradical
  if (numSubs < 1) {              // --- case for addition to =CH2 site
    im.molec.atom[idxNoAdd].atomAttrib--;
    im.molec.atom[idxAdd].atomAttrib++;     //  H shift and form
    im.molec.incBondOrder(idxAdd, idxO);    //   an aldehyde
    im.beUnique();                      // ... numbering is changed here
    idxAdd = im.molec.getIdxAtomMatch(6, 8, 2);
    for (ib = 0; ib < im.molec.atom[idxAdd].numBonds; ib++) {
      if ((im.molec.atom[idxAdd].bond[ib].toAtomicNbr == 6) &&
          (im.molec.atom[idxAdd].bond[ib].order == 1)) {
        idxNoAdd = im.molec.atom[idxAdd].bond[ib].toAtomIndex; break;
      }
    }
    numPrds = molDissoc(im, idxNoAdd, idxAdd, prd[0], prd[1], false);
  } else {                        // --- =CH-R site
    im.beUnique();
    im.molec.getIdcsRadCents(idRC);
    for (ic = 0; ic < 2; ic++) {
      if (im.molec.atom[idRC[ic]].atomicNumber == 8)
        { idxO = idRC[ic]; }
      else if (im.molec.atom[idRC[ic]].atomicNumber == 6)
        { idxNoAdd = idRC[ic]; }
    }
    idxAdd = im.molec.atom[idxO].bond[0].toAtomIndex;
    for (ib = 0; ib < im.molec.atom[idxAdd].numBonds; ib++) {
      idxTo = im.molec.atom[idxAdd].bond[ib].toAtomIndex;
      if ((idxTo != idxNoAdd) && (idxTo != idxO)) {
        im.molec.getDissocGroup(idxTo, idxAdd, ga);
        if ((p = (*rxLb)["Decomp"].find(ga)) != (*rxLb)["Decomp"].end()) {
          prior = p->prior;
          if (im.isCyclic && (!im.uniqueRepr.chain[0].isIdxInChain(idxTo)))
            { prior += 100; }
          if (prior > maxPri) { maxPri = prior; idxMxP = idxTo; }
        }
      }
    }
    numPrds = molDissoc(im, idxAdd, idxMxP, prd[0], prd[1], true);
    if (numPrds < 2) { return; }   // ----- avoid biradical product
  }
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("O");
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  switch (numSubs) {
    case 0: ssubs = "nosb"; break;
    case 1: ssubs = "R"; break;
    default: ssubs = "2Ror1V"; break;
  }
  rxd.rate = "Adtn/O/DB/" + ssubs;
  rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

// ======================================================= [rxnO2addAllyl]
//   Allyl-like radical + O2

void reaction::rxnO2addAllyl(molecule &m) {
  int ip, idxOr, idxOc, idxCo, idxCdb, cidxCdb;
  int numPrds;
  string namRct = m.molName, namPrd[2];
  molecule im, prd[2];
  rxnDescr rxd;

  molAddRcO2(m, im);
  idxOr = im.molec.getIdxRadCent();
  idxOc = im.molec.getIdxNextAtom(idxOr, 8);
  idxCo = im.molec.getIdxNextAtom(idxOc, 6);
  idxCdb = im.molec.getIdxNextAtom(idxCo, 6, 2, 6);
  cidxCdb = im.molec.getIdxNextAtom(idxCdb, 6, 2);
  im.molec.decBondOrder(idxCdb, cidxCdb);
  im.molec.makeNewBond(idxOr, idxCdb);
  im.molec.removeBond(idxCdb, idxCo);
  im.molec.incBondOrder(idxCdb, cidxCdb);
  if (m.isCyclic) { im.beUnique(); }       // - Mar. 22, 2010
  idxCo = im.molec.getIdxRadCent();
  idxOc = im.molec.getIdxNextAtom(idxCo, 8, 1, 8);
  idxOr = im.molec.getIdxNextAtom(idxOc, 8);
  numPrds = molDissoc(im, idxOc, idxOr, prd[0], prd[1], true);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Adtn/O2/Allyl";
  rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

// ======================================================= [rxnO2addVinyl]
//   Vinyl-like radical + O2

void reaction::rxnO2addVinyl(molecule &m) {
  int ip, idxOr, idxOor, idxCoor, idxCd, numPrds;
  string namRct = m.molName, namPrd[2];
  molecule im, prd[2];
  rxnDescr rxd;

  molAddRcO2(m, im);
  idxOr = im.molec.getIdxRadCent();
  idxOor = im.molec.getIdxNextAtom(idxOr, 8);
  idxCoor = im.molec.getIdxNextAtom(idxOor, 6);
  idxCd = im.molec.getIdxNextAtom(idxCoor, 6, 2);
  im.molec.decBondOrder(idxCoor, idxCd);
  im.molec.makeNewBond(idxOr, idxCd);
  im.molec.removeBond(idxOor, idxOr);
  im.molec.incBondOrder(idxCoor, idxOor);
  numPrds = molDissoc(im, idxCd, idxCoor, prd[0], prd[1], true);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Adtn/O2/Vinyl";
  rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

// -------------------------------------------------------- rxnO2addHORalp
//   alpha-hydroxyalkyl + O2 -> HO2 + carbonyl

void reaction::rxnO2addHORalp(molecule &m) {
  int idxRCnt, idxO, nbRC;
  string namRct = m.molName, namPrd, sn;
  molecule prd;
  rxnDescr rxd;

  idxRCnt = m.molec.getIdxRadCent();
  idxO = m.molec.getIdxNextAtom(idxRCnt, 8, 1);
  if (m.molec.atom[idxO].numBonds > 1) { return; }
  nbRC = m.molec.atom[idxRCnt].numBonds;
  if (nbRC == 3) { sn = "T"; } else { sn = "S"; }
  prd = m;
  prd.molec.removeHatom(idxO);
  prd.molec.incBondOrder(idxRCnt, idxO);
  procProdAutoBase(prd); namPrd = prd.molName;
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
  rxd.prod.push_back(namPrd); rxd.prod.push_back("HO2");
  rxd.rate = "Rcmb/R/O2/Oh" + sn;
  rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

// ====================================================== [rxnO2addVinoxy]
//   Vinoxy + O2 reactions

void reaction::rxnO2addVinoxy(molecule &m) {
  int idxOe, idxOoc, idxCoo, idxCo, idxCbeta, nBndsCo, nHCbeta, alphaH;
  int ib, ip, numPrds, nPstep1;
  string namRct = m.molName, namPrd[3];
  vector<int> rt;
  molecule im, imd, prd[3];
  rxnDescr rxd;

  //
  // return;
  //
  molAddRcO2(m, im);
  idxOe = im.molec.getIdxRadCent();
  idxOoc = im.molec.getIdxNextAtom(idxOe, 8);
  idxCoo = im.molec.getIdxNextAtom(idxOoc, 6);
  idxCo = im.molec.getIdxNextAtom(idxCoo, 6, 2, 8);
  nBndsCo = im.molec.atom[idxCo].numBonds;
  if (nBndsCo < 3) { alphaH = true; } else { alphaH = false; }
  if (alphaH) {              // --------------- alpha-H vinoxy radical
     // ... H transfer
    im.molec.atom[idxOe].atomAttrib--;
    im.molec.atom[idxCo].atomAttrib++;
    // im.beUnique();
    idxOe = im.molec.getIdxAtomTermMatch(8, 8, 1);
    idxCo = im.molec.getIdxRadCent();
    im.molec.getDstRoute(idxOe, idxCo, rt);
    idxOoc = rt[1]; idxCoo = rt[2];
                             // ----- ketene + HO2
    numPrds = molDissoc(im, idxCoo, idxOoc, prd[0], prd[1], true);
    for (ip = 0; ip < numPrds; ip++)
      { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
    rxd.clear();
    rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
    for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
    rxd.rate = "Adtn/O2/aHvin/HO2";
    rxd.isRevrs = false;
    lst.push_back(rxd);
                             // ----- CO + aldehyde + OH
    molDissoc(im, idxCo, idxCoo, prd[0], imd, false);
    idxOe = imd.molec.getIdxAtomTermMatch(8, 8, 1);
    idxOoc = imd.molec.getIdxNextAtom(idxOe, 8);
    numPrds = molDissoc(imd, idxOoc, idxOe, prd[1], prd[2], true) + 1;
    for (ip = 0; ip < numPrds; ip++)
      { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
    rxd.clear();
    rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
    for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
    rxd.rate = "Adtn/O2/aHvin/OH";
    rxd.isRevrs = false;
    lst.push_back(rxd);
  } else {                   // --------------- alpha-R vinoxy radical

    for (ib = 0; ib < im.molec.atom[idxCo].numBonds; ib++) {
      if ((im.molec.atom[idxCo].bond[ib].order == 1) &&
          (im.molec.atom[idxCo].bond[ib].toAtomIndex != idxCoo))
        { idxCbeta = im.molec.atom[idxCo].bond[ib].toAtomIndex; break; }
    }
    if (im.molec.atom[idxCbeta].atomicNumber != 6) { return; }
    nHCbeta = im.molec.atom[idxCbeta].getNumHatoms();
    if (nHCbeta > 0) {
      im.molec.atom[idxOe].atomAttrib--;
      im.molec.atom[idxCbeta].atomAttrib++;  // ... H transfer
      // im.beUnique();               // - May 8, 2010
      // idxCbeta = im.molec.getIdxRadCent();
      // idxCoo = im.molec.getIdxAtomMatch(6, 8, 1);
      // im.molec.getDstRoute(idxCbeta, idxCoo, rt);
      // idxCo = rt[1];
                             // ----- ketene + aldehyde + OH
      nPstep1 = molDissoc(im, idxCo, idxCoo, prd[0], prd[1], true);
      imd = prd[nPstep1 - 1];
      idxOe = imd.molec.getIdxAtomTermMatch(8, 8, 1);
      idxOoc = imd.molec.getIdxNextAtom(idxOe, 8);
      // idxCoo = imd.molec.getIdxRadCent();
      idxCoo = imd.molec.getIdxNextAtom(idxOoc, 6);
      numPrds = molDissoc(imd, idxOoc, idxOe, prd[nPstep1 - 1],
                          prd[nPstep1], true) + nPstep1 - 1;
      for (ip = 0; ip < numPrds; ip++)
        { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
      rxd.clear();
      rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
      for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
      rxd.rate = "Adtn/O2/aRvin/OH";
      rxd.isRevrs = false;
      lst.push_back(rxd);
    }

  }
  outLstClear();
}

// ======================================================= [rxnO2addKetyl]
//   Ketyl-like radical + O2

void reaction::rxnO2addKetyl(molecule &m) {
  int ip, idxOr, idxOor, idxCoor, idxCd, numPrds;
  string namRct = m.molName, namPrd[2];
  molecule im, prd[2];
  rxnDescr rxd;

  molAddRcO2(m, im);
  idxOr = im.molec.getIdxRadCent();
  idxOor = im.molec.getIdxNextAtom(idxOr, 8);
  idxCoor = im.molec.getIdxNextAtom(idxOor, 6);
  idxCd = im.molec.getIdxNextAtom(idxCoor, 6, 2);
  im.molec.decBondOrder(idxCoor, idxCd);
  im.molec.makeNewBond(idxOr, idxCd);
  im.molec.removeBond(idxOor, idxOr);
  im.molec.incBondOrder(idxCoor, idxOor);
  numPrds = molDissoc(im, idxCd, idxCoor, prd[0], prd[1], true);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Adtn/O2/Vinyl";
  rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

// ======================================================== [rxnO2addKetR]
//   Reactions of keto-alkyl + O2

void reaction::rxnO2addKetR(molecule &m) {
  int ib, ip, idxOe, idxCo, idxCnn, idxCnf;
  int dstOeCo, numHCo, numHCnf, numPrds;
  string namRct = m.molName, namPrd[2];
  vector<int> rt;
  molecule im, prd[2];
  rxnDescr rxd;

  molAddRcO2(m, im);
  idxOe = im.molec.getIdxRadCent();
  idxCo = im.molec.getIdxAtomMatch(6, 8, 2);
  dstOeCo = im.molec.getDstRoute(idxOe, idxCo, rt);
  numHCo = im.molec.atom[idxCo].getNumHatoms();
  if (numHCo > 0) {          // ---------- HCO-Q
    if ((dstOeCo < 4) || (dstOeCo > 5)) { return; }
    im.molec.atom[idxOe].atomAttrib--;
    im.molec.atom[idxCo].atomAttrib++;
    im.beUnique();           // --- H-shift
    idxCo = im.molec.getIdxRadCent();
    for (ib = 0; ib < im.molec.atom[idxCo].numBonds; ib++) {
      if (im.molec.atom[idxCo].bond[ib].order == 1)
        { idxCnn = im.molec.atom[idxCo].bond[ib].toAtomIndex; break; }
    }
    numPrds = molDissoc(im, idxCnn, idxCo, prd[0], prd[1], false);
    for (ip = 0; ip < numPrds; ip++)
      { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
    rxd.clear();
    rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
    for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
    rxd.rate = "Adtn/O2/HCOQ/ts" + format(dstOeCo + 2);
    rxd.isRevrs = false;
    lst.push_back(rxd);
  } else {                   // ---------- RCO-Q
    if (dstOeCo != 4) { return; }
    idxCnn = rt[dstOeCo - 1];
    for (ib = 0; ib < im.molec.atom[idxCo].numBonds; ib++) {
      if ((im.molec.atom[idxCo].bond[ib].order == 1) &&
          (im.molec.atom[idxCo].bond[ib].toAtomIndex != idxCnn))
        { idxCnf = im.molec.atom[idxCo].bond[ib].toAtomIndex; break; }
    }
    numHCnf = im.molec.atom[idxCnf].getNumHatoms();
    if (numHCnf < 1) { return; }
    im.molec.atom[idxOe].atomAttrib--;
    im.molec.atom[idxCnf].atomAttrib++;  // --- H-shift
    numPrds = molDissoc(im, idxCo, idxCnn, prd[0], prd[1], true);
    for (ip = 0; ip < numPrds; ip++)
      { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
    rxd.clear();
    rxd.reac.push_back(namRct); rxd.reac.push_back("O2");
    for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
    rxd.rate = "Adtn/O2/RCOQ/ts7";
    rxd.isRevrs = false;
    lst.push_back(rxd);
  }
  outLstClear();
}

// =========================================================== [rxnKetene]
//   reactions of ketenes

void reaction::rxnKetene(molecule &m) {
  int ib, ip, idcO, idcCk, idcCd, numPrds;
  string namRct = m.molName, namPrd[2];
  molecule prc, im, prd[2];
  rxnDescr rxd;

  prc = m;
  idcO = prc.molec.getIdxAtomTermMatch(8, 2, 6, 2, 6);
  idcCk = prc.molec.getIdxNextAtom(idcO, 6, 2);
  idcCd = prc.molec.getIdxNextAtom(idcCk, 6, 2);
  prc.molec.decBondOrder(idcCd, idcCk);
                              // ----- H-addition --------------------
  im = prc; im.molec.atom[idcCd].atomAttrib--;
  numPrds = molDissoc(im, idcCd, idcCk, prd[0], prd[1], false);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("H");
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Adtn/H/C=C=O"; rxd.isRevrs = false;
  lst.push_back(rxd);
                              // ----- O-addition --------------------
  im = prc; rfmAddOat(im, idcCd);
  numPrds = molDissoc(im, idcCd, idcCk, prd[0], prd[1], true);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("O");
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Adtn/O/C=C=O"; rxd.isRevrs = false;
  lst.push_back(rxd);
                              // ----- OH-addition -------------------
  im = prc; rfmAddOat(im, idcCk); im.molec.atom[idcCd].atomAttrib--;
  numPrds = molDissoc(im, idcCk, idcCd, prd[0], prd[1], true);
  for (ip = 0; ip < numPrds; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.reac.push_back("OH");
  for (ip = 0; ip < numPrds; ip++) { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Adtn/OH/C=C=O"; rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

void reaction::rxnIntAddRdO2(molecule &m) {           // --- rxnIntAddRdO2
  //   internal addition of RO2 with double bond
  int idRC, idDBnear, idDBfar, ringSnear, ringSfar, idBridgeO, dstRcOc;
  int numProd, ip;
  string namPrd[3], namRct = m.molName;
  vector<int> rt;
  vector<vector<int> > dblBnds;
  molecule RdO, prd[3], imp;
  rxnDescr rxd;

  if (m.isCyclic) { return; }   // - will not process cyclic alkenylperoxy
  molRmvRcO(m, RdO);
  dblBnds = RdO.molec.getAllBondsMatch(6, 6, 2);
  if (dblBnds.size() != 1) { return; }  // - one double bond only
  idRC = RdO.molec.getIdxRadCent();
  idDBnear = dblBnds[0][0]; idDBfar = dblBnds[0][1];
  ringSnear = RdO.molec.getDistance(idRC, idDBnear) + 2;
  ringSfar = RdO.molec.getDistance(idRC, idDBfar) + 2;
  if (ringSnear > ringSfar)
    { swap(idDBnear, idDBfar); swap(ringSnear, ringSfar); }
  if ((ringSnear < 5) || (ringSnear > 7)) { return; }
  rfmBridgeOatDB(RdO, idDBnear, idDBfar);
  RdO.molec.getDstRoute(idRC, idDBnear, rt);
  molDissoc(RdO, rt[1], rt[2], prd[0], prd[1], true);
  imp = prd[1]; numProd = 2;
  idBridgeO = imp.getIdxAtomGrpMatch("O/C2", *grLb);
  dblBnds[0] = imp.molec.atom[idBridgeO].getBondToIdcs();
  idRC = imp.molec.getIdxRadCent();
  idDBnear = dblBnds[0][0]; idDBfar = dblBnds[0][1];
  if (imp.molec.getDistance(idRC, idDBnear) >
    imp.molec.getDistance(idRC, idDBfar)) { swap(idDBnear, idDBfar); }
  dstRcOc = imp.molec.getDstRoute(idRC, idDBnear, rt);
  if (dstRcOc > 1) {
    molDissoc(imp, rt[1], rt[2], prd[1], prd[2], true);
    imp = prd[2]; numProd = 3;
    idBridgeO = imp.getIdxAtomGrpMatch("O/C2", *grLb);
    dblBnds[0] = imp.molec.atom[idBridgeO].getBondToIdcs();
    idRC = imp.molec.getIdxRadCent();
    idDBnear = dblBnds[0][0]; idDBfar = dblBnds[0][1];
    if (imp.molec.getDistance(idRC, idDBnear) >
      imp.molec.getDistance(idRC, idDBfar)) { swap(idDBnear, idDBfar); }
    dstRcOc = imp.molec.getDstRoute(idRC, idDBnear, rt);
  }
  if (dstRcOc == 0) {
    imp.molec.removeBond(idBridgeO, idDBfar);
    imp.molec.incBondOrder(idDBnear, idBridgeO);
  } else {
    imp.molec.removeBond(idBridgeO, idDBnear);
    imp.molec.incBondOrder(idRC, idDBnear);
  }
  prd[numProd - 1] = imp;
  for (ip = 0; ip < numProd; ip++)
    { procProdAutoBase(prd[ip]); namPrd[ip] = prd[ip].molName; }
  rxd.clear();
  rxd.reac.push_back(namRct);
  for (ip = 0; ip < numProd; ip++) { rxd.prod.push_back(namPrd[ip]); }
  rxd.rate = "Iadd/RdO2/near" + format(ringSnear);
  rxd.isRevrs = false;
  lst.push_back(rxd);
  outLstClear();
}

void reaction::rxnEnol(molecule &m) {                       // --- rxnEnol
  int idCdOH, idCd, idOH;
  string namRct = m.molName, namPrd;
  molecule prd;
  rxnDescr rxd;
  if (!m.isEnol(idCdOH, idCd, idOH)) { return; }
  prd = m;
  prd.isUnique = false;
  prd.molec.decBondOrder(idCdOH, idCd);
  prd.molec.atom[idCd].atomAttrib--;
  prd.molec.atom[idOH].atomAttrib++;
  prd.molec.incBondOrder(idCdOH, idOH);
  procProdAutoBase(prd); namPrd = prd.molName;
  rxd.clear();
  rxd.reac.push_back(namRct); rxd.prod.push_back(namPrd);
  rxd.rate = "Isom/Enol";
  rxd.isRevrs = false;
  lst.push_back(rxd);
  rxd.reac.push_back("H"); rxd.prod.push_back("H");
  rxd.rate = "Adtn/H/Enol";
  lst.push_back(rxd);
  outLstClear();
}


/****** molecular manipulation functions ************************************/

// ----------------------------------------------------------- [molDissoc]
//   get dissociated structures (if isRad, radical center is in id1 side).

int reaction::molDissoc(molecule &m, int id1, int id2, molecule &p1,
 molecule &p2, int isRad) {
  int inc1, inc2, idRC, ib;
  molecList::iterator pml;

  if (id2 < 0) {     // --- C-H bond fission
    pml = spAd->molAd.find("H"); p2.setFrom(pml->molec, pml->name);
    p2.beUnique(); p1 = m; p1.molName = "";
    if (isRad) {
      idRC = p1.molec.getIdxRadCent();
      for (ib = 0; ib < p1.molec.atom[id1].numBonds; ib++) {
        if (p1.molec.atom[id1].bond[ib].toAtomIndex == idRC)
          { p1.molec.atom[id1].bond[ib].order++; }
      }
      for (ib = 0; ib < p1.molec.atom[idRC].numBonds; ib++) {
        if (p1.molec.atom[idRC].bond[ib].toAtomIndex == id1)
          { p1.molec.atom[idRC].bond[ib].order++; }
      }
      p1.molec.atom[idRC].atomAttrib--;
      p1.beUnique();
    } else { p1.beRadicalAt(id1); }
    return 2;
  }
  if (m.isCyclic) {
    inc1 = m.uniqueRepr.chain[0].isIdxInChain(id1);
    inc2 = m.uniqueRepr.chain[0].isIdxInChain(id2);
    if (inc1 && inc2) {
      if (isRad) { return molCycRadDis(m, id1, id2, p1); }
      else { return molCycDissoc(m, id1, id2, p1); }
    }
  }
  molDissocHalf(m, id2, id1, p1, isRad);
  molDissocHalf(m, id1, id2, p2, false);
  return 2;
}

// -------------------------------------------------------- [molCycDissoc]
int reaction::molCycDissoc(molecule &m, int id1, int id2, molecule &prd) {
  int idxdb[2], ic, jc, nBnds, ib, toIdx, ibRm;
  int selIc = -1, selIdx = -1, maxPri = 0, idxmb[2];
  groupAtom ga;
  rxnSiteList::iterator p;

  idxdb[0] = id1; idxdb[1] = id2;
  for (ic = 0; ic < 2; ic++) {
    jc = ic + 1; if (jc > 1) { jc -= 2; }
    nBnds = m.molec.atom[idxdb[ic]].numBonds;
    for (ib = 0; ib < nBnds; ib++) {
      toIdx = m.molec.atom[idxdb[ic]].bond[ib].toAtomIndex;
      if (toIdx != idxdb[jc]) {
        m.getGroupAtom(toIdx, ga);
        if ((p = (*rxLb)["Habs"].find(ga)) != (*rxLb)["Habs"].end()) {
          if (p->prior > maxPri)
            { maxPri = p->prior; selIc = ic; selIdx = toIdx; }
        }
      }
    }
  }
  if (selIc < 0) { return 0; }
  prd = m; prd.molName = "";
  for (ic = 0; ic < 2; ic++) {
    jc = ic + 1; if (jc > 1) { jc -= 2; }
    nBnds = prd.molec.atom[idxdb[ic]].numBonds;
    for (ib = 0; ib < nBnds; ib++) {
      toIdx = prd.molec.atom[idxdb[ic]].bond[ib].toAtomIndex;
      if (toIdx == idxdb[jc]) { ibRm = ib; break; }
    }
    for (ib = ibRm; ib < (nBnds - 1); ib++) {
      prd.molec.atom[idxdb[ic]].bond[ib]
       = prd.molec.atom[idxdb[ic]].bond[ib + 1];
    }
    prd.molec.atom[idxdb[ic]].numBonds--;
  }
  idxmb[0] = idxdb[selIc]; idxmb[1] = selIdx;
  for (ic = 0; ic < 2; ic++) {
    jc = ic + 1; if (jc > 1) { jc -= 2; }
    nBnds = prd.molec.atom[idxmb[ic]].numBonds;
    for (ib = 0; ib < nBnds; ib++) {
      toIdx = prd.molec.atom[idxmb[ic]].bond[ib].toAtomIndex;
      if (toIdx == idxmb[jc])
        { prd.molec.atom[idxmb[ic]].bond[ib].order++; break; }
    }
  }
  prd.beUnique();
  return 1;
}

// -------------------------------------------------------- [molCycRadDis]
int reaction::molCycRadDis(molecule &m, int id1, int id2, molecule &prd) {
  int idxRCnt, nBnds, ib, toIdx, ibRm;

  idxRCnt = m.molec.getIdxRadCent();
  prd = m; prd.molName = "";
  nBnds = prd.molec.atom[id1].numBonds;
  for (ib = 0; ib < nBnds; ib++) {
    toIdx = prd.molec.atom[id1].bond[ib].toAtomIndex;
    if (toIdx == id2) { ibRm = ib; }
    if (toIdx == idxRCnt) { prd.molec.atom[id1].bond[ib].order++; }
  }
  for (ib = ibRm; ib < (nBnds - 1); ib++)
    { prd.molec.atom[id1].bond[ib] = prd.molec.atom[id1].bond[ib + 1]; }
  prd.molec.atom[id1].numBonds--;
  nBnds = prd.molec.atom[id2].numBonds;
  for (ib = 0; ib < nBnds; ib++) {
    toIdx = prd.molec.atom[id2].bond[ib].toAtomIndex;
    if (toIdx == id1) { ibRm = ib; }
  }
  for (ib = ibRm; ib < (nBnds - 1); ib++)
    { prd.molec.atom[id2].bond[ib] = prd.molec.atom[id2].bond[ib + 1]; }
  prd.molec.atom[id2].numBonds--;
  prd.molec.atom[id2].atomAttrib++;
  nBnds = prd.molec.atom[idxRCnt].numBonds;
  for (ib = 0; ib < nBnds; ib++) {
    toIdx = prd.molec.atom[idxRCnt].bond[ib].toAtomIndex;
    if (toIdx == id1) { prd.molec.atom[idxRCnt].bond[ib].order++; }
  }
  prd.molec.atom[idxRCnt].atomAttrib--;
  prd.beUnique();
  return 1;
}

// ------------------------------------------------------- [molDissocHalf]
void reaction::molDissocHalf(molecule &m, int cid, int idx,
 molecule &prd, int isRad) {
  int ir, idxRCnt = -1, nid, oid, nBnds, ib, toIdx, bo, ibn;
  vector<int> idcsRCnts;
  map<int, int> mapIdx;
  map<int, int>::iterator p, f;
  atomInfo ai;

  if (isRad) {
    m.molec.getIdcsRadCents(idcsRCnts);
    for (ib = 0; ib < m.molec.atom[idx].numBonds; ib++) {
      toIdx = m.molec.atom[idx].bond[ib].toAtomIndex;
      for (ir = 0; ir < idcsRCnts.size(); ir++) {
        if ((idcsRCnts[ir] == toIdx) && (idcsRCnts[ir] != cid))
          { idxRCnt = idcsRCnts[ir]; break; }
      }
      if (idxRCnt >= 0) { break; }
    }
  }
  mapIdx.clear(); prd.clear();
  followConnxn(m.molec, cid, idx, mapIdx);
  for (p = mapIdx.begin(), nid = 0; p != mapIdx.end(); p++, nid++)
    { p->second = nid; }
  prd.molec.numAtoms = mapIdx.size();
  for (p = mapIdx.begin(); p != mapIdx.end(); p++) {
    oid = p->first;
    ai.atomicNumber = m.molec.atom[oid].atomicNumber;
    ai.atomAttrib = m.molec.atom[oid].atomAttrib;
    ai.numBonds = (nBnds = m.molec.atom[oid].numBonds);
    for (ib = 0, ibn = 0; ib < nBnds; ib++) {
      toIdx = m.molec.atom[oid].bond[ib].toAtomIndex;
      bo = m.molec.atom[oid].bond[ib].order;
      if ((f = mapIdx.find(toIdx)) == mapIdx.end()) {
        ai.atomAttrib += bo; ai.numBonds--;
      } else {
        ai.bond[ibn].order = bo; ai.bond[ibn].toAtomIndex = f->second;
        ai.bond[ibn].toAtomicNbr = m.molec.atom[oid].bond[ib].toAtomicNbr;
        if (isRad) {
          if ( ((oid == idx) && (toIdx == idxRCnt)) ||
               ((oid == idxRCnt) && (toIdx == idx)) ) {
            ai.bond[ibn].order++;
          }
        }
        ibn++;
      }
    }
    if (isRad) {
      if ((oid == idx) || (oid == idxRCnt)) { ai.atomAttrib = 0; }
    }
    prd.molec.atom.push_back(ai);
  }
  // prd.beUnique();
}

// -------------------------------------------------------- [followConnxn]
void reaction::followConnxn(molecInfo &mi, int cid, int idx,
 map<int, int> &l) {
  int newIdx = 0, ib, toIdx;
  if (l.find(idx) != l.end()) { return; }
  l.insert(make_pair(idx, newIdx));
  for (ib = 0; ib < mi.atom[idx].numBonds; ib++) {
    toIdx = mi.atom[idx].bond[ib].toAtomIndex;
    if (toIdx != cid) { followConnxn(mi, idx, toIdx, l); }
  }
}

// ---------------------------------------------------------- [molDisKetR]
//   Dissociation of carbonyl alkyl radicals

int reaction::molDisKetR(molecule &m, molecule &p1, molecule &p2) {
  int ib, idxCO, idxRC, cidx, distRCCO;
  vector<int> rt;
  if ((idxCO = m.molec.getIdxAtomMatch(6, 8, 2)) < 0) { return 0; }
  if ((idxRC = m.molec.getIdxRadCent()) < 0) { return 0; }
  distRCCO = m.molec.getDstRoute(idxRC, idxCO, rt);
  if (distRCCO == 1) {                    // === ketene formation
    if (m.molec.atom[idxCO].numBonds > 2) {   // --- break C-C=O
      for (ib = 0; ib < m.molec.atom[idxCO].numBonds; ib++) {
        if ((m.molec.atom[idxCO].bond[ib].toAtomicNbr == 6) &&
            (m.molec.atom[idxCO].bond[ib].toAtomIndex != idxRC))
          { cidx = m.molec.atom[idxCO].bond[ib].toAtomIndex; break; }
      }
      return molDissoc(m, idxCO, cidx, p1, p2, true);
    } else {                                  // --- break H-C=O
      return molDissoc(m, idxCO, -1, p1, p2, true);
    }
  } else {
    if (m.molec.getBondOrder(rt[1], rt[2]) > 1) {
      p1 = m; return 1;
    } else {                          // === R + C=C- formation
      return molDissoc(m, rt[1], rt[2], p1, p2, true);
    }
  }
}

// ---------------------------------------------------------- [molAddRcO2]
//   Add OO to the radical center

void reaction::molAddRcO2(molecule &m, molecule &prd) {
  int idr, nAtoms, nBonds;
  atomInfo ai;

  idr = m.molec.getIdxRadCent();
  prd = m; prd.molName = "";
  nAtoms = prd.molec.numAtoms;
  prd.molec.atom[idr].atomAttrib--;
  nBonds = prd.molec.atom[idr].numBonds;
  prd.molec.atom[idr].bond[nBonds].order = 1;
  prd.molec.atom[idr].bond[nBonds].toAtomicNbr = 8;
  prd.molec.atom[idr].bond[nBonds].toAtomIndex = nAtoms;
  prd.molec.atom[idr].numBonds++;
  ai.atomicNumber = 8; ai.atomAttrib = 0; ai.numBonds = 2;
  ai.bond[0].order = 1; ai.bond[0].toAtomicNbr = 6;
  ai.bond[0].toAtomIndex = idr;
  ai.bond[1].order = 1; ai.bond[1].toAtomicNbr = 8;
  ai.bond[1].toAtomIndex = nAtoms + 1;
  prd.molec.atom.push_back(ai);
  prd.molec.numAtoms++;
  nAtoms = prd.molec.numAtoms;
  ai.atomicNumber = 8; ai.atomAttrib = 1; ai.numBonds = 1;
  ai.bond[0].order = 1; ai.bond[0].toAtomicNbr = 8;
  ai.bond[0].toAtomIndex = nAtoms - 1;
  prd.molec.atom.push_back(ai);
  prd.molec.numAtoms++;
  prd.beUnique();
}

// ----------------------------------------------------------- [molAddRcO]
//   Add O to the radical center

void reaction::molAddRcO(molecule &m, molecule &prd) {
  int idr = m.molec.getIdxRadCent();
  prd = m; rfmAddOat(prd, idr); prd.beUnique();
}

// ----------------------------------------------------------- [molRmvRcO]
//   remove O from the radical center

void reaction::molRmvRcO(molecule &m, molecule &prd) {
  int idr, idn;

  idr = m.molec.getIdxRadCent();
  idn = m.molec.atom[idr].bond[0].toAtomIndex;
  molDissocHalf(m, idr, idn, prd, false);
}

                                                           // --- molAddOH
void reaction::molAddOH(molecule &m, molecule &prd, int id) {
  int na, nb;
  atomInfo ai;
  prd = m; prd.molName = "";
  na = prd.molec.numAtoms;
  nb = prd.molec.atom[id].numBonds;
  prd.molec.atom[id].atomAttrib--;
  prd.molec.atom[id].bond[nb].order = 1;
  prd.molec.atom[id].bond[nb].toAtomicNbr = 8;
  prd.molec.atom[id].bond[nb].toAtomIndex = na;
  prd.molec.atom[id].numBonds++;
  ai.atomicNumber = 8; ai.atomAttrib = 0; ai.numBonds = 1;
  ai.bond[0].order = 1; ai.bond[0].toAtomicNbr = 6;
  ai.bond[0].toAtomIndex = id;
  prd.molec.atom.push_back(ai);
  prd.molec.numAtoms++;
  prd.beUnique();
}

                                                          // --- molAddHO2
void reaction::molAddHO2(molecule &m, molecule &prd, int id) {
  int na, nb;
  atomInfo ai;
  prd = m; prd.molName = "";
  na = prd.molec.numAtoms;
  nb = prd.molec.atom[id].numBonds;
  prd.molec.atom[id].atomAttrib--;
  prd.molec.atom[id].bond[nb].order = 1;
  prd.molec.atom[id].bond[nb].toAtomicNbr = 8;
  prd.molec.atom[id].bond[nb].toAtomIndex = na;
  prd.molec.atom[id].numBonds++;
  ai.atomicNumber = 8; ai.atomAttrib = 0; ai.numBonds = 2;
  ai.bond[0].order = 1; ai.bond[0].toAtomicNbr = 6;
  ai.bond[0].toAtomIndex = id;
  ai.bond[1].order = 1; ai.bond[1].toAtomicNbr = 8;
  ai.bond[1].toAtomIndex = na + 1;
  prd.molec.atom.push_back(ai);
  prd.molec.numAtoms++;
  na = prd.molec.numAtoms;
  ai.atomicNumber = 8; ai.atomAttrib = 0; ai.numBonds = 1;
  ai.bond[0].order = 1; ai.bond[0].toAtomicNbr = 8;
  ai.bond[0].toAtomIndex = na - 1;
  prd.molec.atom.push_back(ai);
  prd.molec.numAtoms++;
  prd.beUnique();
}

// ----------------------------------------------------------- [rfmAddOat]
//   Add O to the specified atom

int reaction::rfmAddOat(molecule &tm, int idr) {
  int nAtoms, nBonds;
  atomInfo ai;

  nAtoms = tm.molec.numAtoms;
  tm.molec.atom[idr].atomAttrib--;
  nBonds = tm.molec.atom[idr].numBonds;
  tm.molec.atom[idr].bond[nBonds].order = 1;
  tm.molec.atom[idr].bond[nBonds].toAtomicNbr = 8;
  tm.molec.atom[idr].bond[nBonds].toAtomIndex = nAtoms;
  tm.molec.atom[idr].numBonds++;
  ai.atomicNumber = 8; ai.atomAttrib = 1; ai.numBonds = 1;
  ai.bond[0].order = 1; ai.bond[0].toAtomicNbr = 6;
  ai.bond[0].toAtomIndex = idr;
  tm.molec.atom.push_back(ai);
  tm.molec.numAtoms++;
  return nAtoms;  // --- returns index of added O
}

                                                     // --- rfmBridgeOatDB
int reaction::rfmBridgeOatDB(molecule &tm, int id1, int id2) {
  // reform tm by newly making bridged O atom (epoxy ring) at double bond
  // returns index of new O atom
  int na = tm.molec.numAtoms;
  bondInfo bi;
  atomInfo ai;

  tm.molec.decBondOrder(id1, id2);
  bi.order = 1; bi.toAtomicNbr = 8; bi.toAtomIndex = na;
  tm.molec.atom[id1].bond[tm.molec.atom[id1].numBonds++] = bi;
  tm.molec.atom[id1].atomAttrib--;
  tm.molec.atom[id2].bond[tm.molec.atom[id2].numBonds++] = bi;
  tm.molec.atom[id2].atomAttrib--;
  ai.atomicNumber = 8; ai.atomAttrib = 0; ai.numBonds = 2;
  bi.toAtomicNbr = 6;
  bi.toAtomIndex = id1; ai.bond[0] = bi;
  bi.toAtomIndex = id2; ai.bond[1] = bi;
  tm.molec.atom.push_back(ai);
  tm.molec.numAtoms++;
  return na;
}

